package com.polytechnik.utils;
/*
 * (C) Vladislav Malyshkin 2011
 * This file is under GPL version 3.
 *
 */

/** Chebyshev polynomials basis,
 * using  T_k=alfa_{k-1} x*T_{k-1}-T_{k-2}, where alfa_{k}=1 for k=0 & 2 otherwise.
 *  @version $Id: Chebyshev.java,v 1.65 2017/09/08 10:13:21 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
public class Chebyshev extends BasisPolynomials {


/** Calculates n-th Chebyshev polynomial
 * using  T_k=alfa_{k-1} x*T_{k-1}-T_{k-2}, where alfa_{k}=1 for k=0 & 2 otherwise.
 * @param n T index.
 * @param res Output vector.
 * @param Tnm1 Input T_{n-1}
 * @param Tnm2 Input T_{n-2}
 */
public void setNext(final int nd,final double [] res,
		    final double [] Tnm1,final double [] Tnm2,
		    final boolean flag_type2)
{
    if(nd<0) throw new IllegalArgumentException();
    else if(nd==0)
    {
	res[0]=1;
    }
    else if(nd==1)
    {
	res[0]=0;
	res[1]=!flag_type2?1:2;
    }
    else
    {
	res[nd]=2*Tnm1[nd-1];
	res[nd-1]=2*Tnm1[nd-2];
	for(int k=nd-1;--k>=0;)
	{
	    if(k>0)
		res[k]=2*Tnm1[k-1]-Tnm2[k];
	    else
		res[k]=-Tnm2[k];
	}
    }
}


/**
 * Calculation of the value of the Chebyshev polynomial.
 * @param n T index.
 * @param x Argument.
 * @return The value of the Chebyshev polynomial Tn at x.
 */
public double calculate(final int n,final double x)
{
    if(n<0) throw new RuntimeException("n="+n);
    double b=1,s=x;
    if(n==0) return b;
    if(n==1) return s;
    for(int i=2;i<=n;++i)
    {
	s=-b+2*x*(b=s);
    }
    return s;
}

public double [] calculateMomentsFromSample(final int n,final double [] x,final double [] w,final int nelsinsample)
{
    final double [] m=new double [n+1];
    for(int k=0;k<nelsinsample;k++)
    {
	final double xx=x[k],ww=w!=null?w[k]:1;
	double b=1,s=xx;
	for(int i=0;i<=n;++i)
	{
	    if(i<=0) m[i]+=b*ww;
	    else if(i<=1) m[i]+=s*ww;
	    else {s=-b+2*(xx)*(b=s);m[i]+=s*ww;}
	}
    }
    return m;
}

/** Calculates sum_{k=0}^{k=n}T_k(x) Tcoefs_k.
 *  Two versions: original & the one copied from http://www.alglib.net/.
 * @param n T index.
 * @param Tcoefs Coefficients of T_n.
 * @param x Point to estimate the sum.
 * @return The sum.
 */
public double calculateSumD0(final int n,final double [] Tcoefs,final double x)
{
    /*
    if(n<0) throw new RuntimeException("n="+n);
    double b=1,s=x,sum=0;
    if(n>=0) sum+=b*Tcoefs[0];
    if(n>=1) sum+=s*Tcoefs[1];
    for(int i=2;i<=n;++i)
    {
	sum+=(s=-b+2*x*(b=s))*Tcoefs[i];

    }
    return sum;
    */

    if(n<0) throw new RuntimeException("n="+n);
    double result = 0,b1 = 0,b2 = 0;

    for(int i=n; i>=0; i--)
    {
	final double ai=i<=0?1:2;
	result = ai*x*b1-b2+Tcoefs[i];
	b2 = b1;
	b1 = result;
    }

    return result;
}

/** Calculates d/dx(sum_{k=0}^{k=n}T_k(x) Tcoefs_k).
 * @param n T index, elements=n+1
 * @param Tcoefs Coefficients of T_n.
 * @param x Pont to estimate the sum.
 * @return The derivative.
 */
public double calculateSumD1(final int n,final double [] Tcoefs,final double x)
{
    if(n<0) throw new RuntimeException("n="+n);
    /*
    double b=1,s=2*x,sum=0;
    if(n>=1) sum+=b*Tcoefs[1];
    if(n>=2) sum+=2*s*Tcoefs[2];
    for(int i=3;i<=n;++i)
    {
	sum+=(s=-b+2*x*(b=s))*Tcoefs[i]*i;

    }
    return sum;
    */

    double result = 0,b1 = 0,b2 = 0;
    double se=0,so=0;
    for(int k=n;--k>=0;)
    {
	final double c=(2*(k+1))*Tcoefs[k+1],dk;
	if(k%2==0)
	{
	    se+=c;
	    dk=(k==0?se/2:se);
	}
	else
	{
	    so+=c;
	    dk=so;
	}
	final double ak=k<=0?1:2;
	result = ak*x*b1-b2+dk;
	b2 = b1;
	b1 = result;
    }
    return result;
}


/** Calculates d^2/dx^2(sum_{k=0}^{k=n}T_k(x) Tcoefs_k).
 * @param n T index, elements=n+1
 * @param Tcoefs Coefficients of T_n.
 * @param x Pont to estimate the sum.
 * @return The second derivative.
 */
public double calculateSumD2(final int n,final double [] Tcoefs,final double x)
{
    if(n<0) throw new RuntimeException("n="+n);
    double se=0,so=0,ase=0,aso=0;
    double result = 0,b1 = 0,b2 = 0;

    for(int i=n-2; i>=0; i--)
    {
	final double c=(i+2)*Tcoefs[i+2],di;
	if(!((i%2)==0))
	{
	    se+=2*c;
	    final double ac=(i+1)*se;
	    aso+=2*ac;
	    di=aso;
	}
	else
	{
	    so+=2*c;
	    final double ac=(i+1)*so;
	    ase+=2*ac;
	    di=(i<=0?ase/2:ase);
	}

	final double ai=i<=0?1:2;
	result = ai*x*b1-b2+di;
	b2 = b1;
	b1 = result;
    }

    return result;
}




/** Differentiate a function in Chebyshev basis,
 * F(x) - a polynomial of n-th order.
 * То calculate use the formulae
 * Prudnikov, Brychkov, Marichev, volume 2, 2003, page 651 T_k(x)'=kU_{k-1}(x)
 * and then express U_{k-1}(x) through T_k(x) using
 * http://en.wikipedia.org/wiki/Chebyshev_polynomials#Relation_between_Chebyshev_polynomials_of_the_first_and_second_kinds
 * (recurrently apply formulae 22.5.8 Abramowitz and Stegun 1972, p. 778).
 * @param n Max index.
 * @param Tcoefs Input: F(x)=T_k(x)*Tcoefs[k].
 * @return  dF/dx=T'_k(x)*Tcoefs[k].
 */
public double [] getDifferentiated(final int n,final double [] Tcoefs)
{
    // use formulae from I.Gohberg, V.Olshevsky
    // Fast Inversion of Chebyshev-Vandermonde matrices.
    // http://www.math.uconn.edu/~olshevsky/papers/chvan.ps
    // I.Gohberg and V. Olshevsky. Fast inversion of Chebyshev-Vandermonde matrices. Numerische Mathematik,   67 (1994), 71-92.
    final double [] d=new double [n];
    double se=0,so=0;
    for(int k=n;--k>=0;)
    {
	final double c=(k+1)*Tcoefs[k+1];
	if(k%2==0)
	{
	    se+=2*c;
	    d[k]=(k==0?se/2:se);
	}
	else
	{
	    so+=2*c;
	    d[k]=so;
	}
    }
    return d;
}


/** Integrate a function in Chebyshev basis,
 * То calculate use the formulae
 * Prudnikov, Brychkov, Marichev, volume 2, 2003, formulae 1.14.2.1 page 46.
 * @param n Max index.
 * @param Tcoefs Input: F(x)=T_k(x)*Tcoefs[k].
 * @return  \int_0^x F(x)dx=\int_0^x T_k(x)dx*Tcoefs[k].
 */
public double [] getIntegrated(final int n,final double [] Tcoefs)
{
    if(n<0) throw new RuntimeException("n="+n);
    final double [] i=new double [n+2];
    for(int k=n;k>=0;--k)
    {
	final double c=0.5*Tcoefs[k];
	if(k<=0)
	{
	    i[k+1]+=2*c;
	}
	else
	{
	    i[k+1]+=c/(k+1);
	    if(k>1){i[k-1]-=c/(k-1);}
	    else {i[k-1]+=0.5*c;}
	}
    }
    i[0]-=calculateSumD0(n+1,i,0); // normalize integral to f(0)=0
    return i;
}



/** Expand Tl(z)Tm(z)=sum_k=0^{l+m} s_k Tk(z),
 *  formulae 22.7.24 Abramowitz and Stegun 1972, p. 872.
 *  @param l l.
 *  @param m m.
 *  @param s coefs.
 */
public void setQlQmExpansionCoefs(final int l,final int m,final double [] s)
{
    java.util.Arrays.fill(s,0,l+m+1,0);
    s[l+m]=0.5;
    s[l>m?l-m:m-l]+=0.5;
}


/** Multiply p by ax+b.
 * Calculations in Chebyshev.
 * @param n The max dimension of p, elements=n+1.
 * @param res The result.
 * @param p Input polynomial.
 */
@Override
public void setaxbP(final int n,final double [] res,final double [] p,final double a,final double b)
{
    double r=p[n];
    res[n]=res[n+1]=0;
    for(int k=n;k>=0;k--)
    {
	final double ak=(k<=0?1:2),ar=a*r;
	res[k+1]+=ar/ak;
	res[k]+=b*r;
	if(k>0)
	{
	    final double t=p[k-1];
	    res[k-1]=ar/ak;
	    r=t;
	}
    }
}


/** Divide  p by (x-x0), returns the remainder.
 * Calculations in Chebyshev basis.
 * @param n The max dimension of p, n_elements=n+1.
 * @param res The result.
 * @param p Input polynomial.
 * @param x0 To divide on (x-x0)
 * @return The remainder.
 */
@Override
public double sdiv1(final int n,final double [] res,final double [] p,final double x0)
{
    double r=p[n];
    //System.err.println("pi="+java.util.Arrays.toString(p));
    if(n>0) res[n-1]=p[n-1];
    for(int k=n;--k>=0;)
    {
	final double al=(k<=0?1:2);
	final double c=r*al,b=(res[k]+=c*(x0));
	if(k>0)res[k-1]=p[k-1]-r;
	res[k]=c;
	r=b;
    }
    //System.err.println("p="+java.util.Arrays.toString(p));
    //System.err.println("res="+java.util.Arrays.toString(res));
    return r;
}


/** Calculates T_n(a(x+x0)) using 3 term recurrence.
 * Calculations in Chebyshev basis.
 * The b_k^{n} are defined as:
 * T_n(a(x+x0))=sum_{k=0}^{k=n}b_k^{n} T_k(x)
 * The method should be applied consequently to get b_k^{n} for n=0,1,2,3,...
 * @param n The max dimension of b.
 * @param b On input: b_k^{n-1}, on output b_k^{n}.
 * @param q On input: b_k^{n-2}, on output: b_k^{n-1}.
 */
public void setNewtonBinomialLikeCoefs(final int n,final double [] b,final double [] q,final double a,final double x0)
{
    if(n<=0)
    {
	b[n]=1;
	return;
    }
    final int n1=n-1;
    double r=b[n1];
    b[n1]=b[n]=0;
    final double an=n<=1?1:2;
    for(int k=n;--k>=0;)
    {
	final double ak=k<=0?1:2,r_a_anak=r*a*(an/ak);
	b[k+1]+=r_a_anak;
	b[k]+=(x0*an*a)*r;
	if(k<n1) b[k]-=q[k];
	q[k]=r;
	if(k>0)
	{
	    final double bb=b[k-1];
	    b[k-1]=r_a_anak;
	    r=bb;
	}
    }
}


/** Calculate <F|xTk> moments for k=0..n-1 from <F|Tk>, k=0..n
 *  @param n Max index.
 *  @param d Output, d[k]=<F|xTk>, k=0..n-1;
 *  @param Pcoefs Input, Pcoefs[k]=<F|Tk>, k=0..n;
 */
@Override
public void setXQnMomentsFromQnMoments(final int n,final double [] d,final double [] moments)
{
    for(int k=n;--k>=0;)
    {
	final double ak=k<=0?1:2;
	d[k]=(moments[k+1]+(k>0?moments[k-1]:0))/ak;
    }
}


/** Calculates confederate matrix to solve
 * polynomial equations in Chebyshev basis directly.
 * Eigenvalues of confederate matrix are the roots.
 * See:
 * Tom Bella, Y.Eidelman, I.Gohberg, V.Olshevsky, E.Tyrtyshnikov,
 * "Fast inversion of Hessenberg-quasiseparable-Vandermonde matrices
 * and resulting recurrence relations and characterizations", March 2007.
 * http://www.math.uconn.edu/~olshevsky/papers/traubqs2_tb36.pdf
 * We use the three term recurrence:
 * T_k=alfa_{k-1} x*T_{k-1}-T_{k-2}, where alfa_{k}=1 for k=0, and 2 otherwise.
 * Use the class PolynomialRootsConfederateMatrixABasis.ABASISCHEBYSHEV
 * to calculate roots.
 */
public double [] getConfederateMatrix(final double [] coefs)
{
    final int n=coefs.length-1;
    final double [] matr=new double [n*n];
    for(int m=0;m<n;m++)
    {
	// alfa_{m} for T, for U alfa_{m} always =2.
	final double am=m<=0?1:2;
	matr[m*n+m]=0/am;                   // delta/alfa
	if(m>0) matr[(m-1)*n+m]=1/am;   // gamma/alfa
	if(m<n-1) matr[(m+1)*n+m]=1/am; // 1/alfa
	else
	{
	    final double f=coefs[n]*am; // p[n]*alfa
	    for(int j=0;j<n;j++) matr[j*n+m]-=coefs[j]/f;
	}
    }
    return matr;
}


public static void main(String [] args)
{
    final double eps=1e-6;
    new TestBasisPolynomials("Chebyshev",new Chebyshev(),new java.util.Random(-79099),eps){
	double [] getMonomialsMoments(final int n)
	{
	    final double [] pmom=new double [n];
	    for(int k=0;k<pmom.length;k++){  // monomials moments
		// Gradstein Ryzik, 1963, 3.248.{2,3}, page 308.
		if(k%2!=0) {pmom[k]=0;}
		else {pmom[k]=Math.PI*Polynomials.fact2(k-1)/Polynomials.fact2(k);}
	    }
	    return pmom;
	}
	OrthogonalPolynomialsABasis getOrtPol()
	{
	    return new OrthogonalPolynomialsChebyshevBasis();
	}
	double [][] getExact()
	{
	    return new double [][]{
		{1},
		{0,1},
		{-1,0,2},
		{0,-3,0,4},
		{1,0,-8,0,8},
		{0,5,0,-20,0,16},
		{-1,0,18,0,-48,0,32},
		{0,-7,0,56,0,-112,0,64},
		{1,0,-32,0,160,0,-256,0,128},
		{0,9,0,-120,0,432,0,-576,0,256},
		{-1,0,50,0,-400,0,1120,0,-1280,0,512},
		{0,-11,0,220,0,-1232,0,2816,0,-2816,0,1024},
		{1,0,-72,0,840,0,-3584,0,6912,0,-6144,0,2048}
	    };
	}
    }.runTests("$Id: Chebyshev.java,v 1.65 2017/09/08 10:13:21 mal Exp $");
}

} // END of class Chebyshev
