package com.polytechnik.utils;
/*
 * (C) Vladislav Malyshkin 2008
 * This file is under GPL version 3.
 *
 */
import java.util.ConcurrentModificationException;
/** A circular buffer to store data.
 *
 *  @version $Id: CircularList.java,v 1.36 2012/01/13 16:51:04 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
public class CircularList<T extends com.polytechnik.utils.CircularList.TimeAccess> {

    public static interface TimeAccess {
	long getTime();
    }


    private T [] buffer=null;
    /* A pointer to where to add new element. */
    private int ptr=0;
    private int n_elements=0;
    private long last_cleared_before=0L;

    public CircularList(final int initialcapacity)
    {
	@SuppressWarnings("unchecked")
	final T [] t=(T [])new TimeAccess [initialcapacity];
	this.buffer=t;
    }

    public CircularList()
    {
	this(20);
    }


    public void ensureCapacity(final int newcapacity)
    {
	final int n0=this.n_elements;
	final T [] buf=this.buffer;
	if(newcapacity<=buf.length)
	{
	    //return;
	    throw new IllegalArgumentException("Try to catch wrong usage");
	}
	@SuppressWarnings("unchecked")
	final T [] newbuffer=(T [])new TimeAccess [newcapacity];
	final int n_before=this.ptr-this.n_elements;
	if(n_before>=0)
	{
	    System.arraycopy(buf,n_before,newbuffer,0,this.n_elements);
	}
	else
	{
	    System.arraycopy(buf,buf.length+n_before,newbuffer,0,-n_before);
	    System.arraycopy(buf,0,newbuffer,-n_before,this.ptr);
	}
	if(n0!=this.n_elements)
	{
	    throw new ConcurrentModificationException();
	}
	this.ptr=this.n_elements;
	this.buffer=newbuffer;

    }

    public void add(T t)
    {
	final int n0=this.n_elements;
	// check for ordered list
	if(this.n_elements>0)
	{
	    int pprev=this.ptr-1;
	    if(pprev<0)
	    {
		pprev+=this.buffer.length;
	    }
	    if(t.getTime()<this.buffer[pprev].getTime())
	    {
		throw new RuntimeException("An attempt to add an element with time="+
					   t.getTime()+
					   " before last added element time="+
					   this.buffer[pprev].getTime());
	    }
	}

	if(this.n_elements>=this.buffer.length)
	{
	    ensureCapacity(1+2*this.buffer.length);
	}
	if(this.buffer[this.ptr]!=null)
	{
	    throw new RuntimeException("Something is terrible wrong. The buffer["+this.ptr+"]="+this.buffer[this.ptr]+". It should be null.");
	}
	this.buffer[this.ptr++]=t;
	if(this.ptr>=this.buffer.length)
	{
	    this.ptr=0;
	}
	this.n_elements++;
	if(n0+1!=this.n_elements)
	{
	    throw new ConcurrentModificationException();
	}
    }


    public int clearBefore(final long t0,final TObjectProcedure<T> callback)
    {
	return clearBefore(t0,callback,0);
    }

    /** Clear all elements with time <t0.
     *  We assume trades are ordered,
     *  so we only clean the tail.
     *  Also make sure the are at least min_elements_to_store
     *  elements left in buffer.
     */
    public int clearBefore(final long t0,final TObjectProcedure<T> callback,final int min_elements_to_keep)
    {
	if(t0>=this.last_cleared_before)
	{
	    this.last_cleared_before=t0;
	}
	else
	{
	    throw new RuntimeException(
		"Previous clearBefore was at t="+this.last_cleared_before+
		". New attempt to clear before t="+t0+
		". Cannot perform because "+t0+" < "+this.last_cleared_before);
	}

	final int n0=this.n_elements,ptr0=this.ptr;
	final T [] buf=this.buffer;
	int p0=this.ptr-this.n_elements;
	if(p0<0)
	{
	    p0+=buf.length;
	}
	int n=0;
	for(int p=p0,ne=this.n_elements;ne-- >0;p++)
	{
	    if(p>=buf.length)
	    {
		p=0;
	    }
	    if(buf[p].getTime()<t0 && this.n_elements>min_elements_to_keep)
	    {
		if(callback!=null)
		{
		    callback.execute(buf[p]);
		}
		buf[p]=null;
		this.n_elements--;
		n++;
	    }
	    else
	    {
		break; // we asume the list is ordered.
	    }
	}
	if(n0-n!=this.n_elements || buf!=this.buffer || ptr0!=this.ptr)
	{
	    throw new ConcurrentModificationException();
	}
	return n;
    }

    /** Remove element.
     *  Slow, because we scan all elements (from end to the beginning).
     *  The method removes a sigle element when checker.execute() return false.
     */
    public int removeBrokenElement(final TObjectProcedure<T> checker)
    {
	if(false)
	{
	    System.err.println("-------- REMOVE ELEMENT (seldom operation)");
	}
	final int n0=this.n_elements;
	final T [] buf=this.buffer;
	int pxlast=this.ptr-this.n_elements;
	if(pxlast<0)
	{
	    pxlast+=buf.length;
	}

	int n=0;
	// scan from the end to the beginning
	for(int p=this.ptr-1,ne=this.n_elements;ne-- >0;p--)
	{
	    if(p<0)
	    {
		p+=buf.length;
	    }
	    if(!checker.execute(buf[p]))
	    {
		if(p<this.ptr)
		{ // shift head
		    System.arraycopy(buf,p+1,buf,p,this.ptr-p-1);
		    buf[--this.ptr]=null;
		}
		else if(p>=pxlast)
		{ // shift tail
		    System.arraycopy(buf,pxlast,buf,pxlast+1,p-pxlast);
		    buf[pxlast]=null;
		}
		else
		{
		    throw new RuntimeException("This should never happen p="+p+" pxlast="+pxlast);
		}
		this.n_elements--;
		n++;
		break;
	    }
	}
	if(n0-n!=this.n_elements || buf!=this.buffer)
	{
	    throw new ConcurrentModificationException();
	}
	return n;
    }


    public int copyData(final T [] items)
    {
	final int n0=this.n_elements,ptr0=this.ptr;
	final T [] buf=this.buffer;
	final int n_before=this.ptr-this.n_elements;
	if(n_before>=0)
	{
	    System.arraycopy(buf,n_before,items,0,this.n_elements);
	}
	else
	{
	    System.arraycopy(buf,buf.length+n_before,items,0,-n_before);
	    System.arraycopy(buf,0,items,-n_before,this.ptr);
	}
	// clear the rest
	for(int i=items.length,ne=this.n_elements;i-- > ne;)
	{
	    items[i]=null;
	}
	if(n0!=this.n_elements || buf!=this.buffer || ptr0!=this.ptr)
	{
	    throw new ConcurrentModificationException();
	}
	return this.n_elements;
    }



    public long getLastClearedBefore()
    {
	return this.last_cleared_before;
    }


    public int getSize()
    {
	return this.n_elements;
    }


    public T getLastAddedElement()
    {
	if(this.n_elements<=0)
	{
	    return null;
	}
	int pprev=this.ptr-1;
	if(pprev<0)
	{
	    pprev+=this.buffer.length;
	}
	return this.buffer[pprev];
    }


    /** The method loops backward over the elements of the list,
     *   set the n_skip_from_the_end=0 to loop over all elements.
     */
    public boolean backwardForEachEntry(final com.polytechnik.utils.TObjectProcedure<T> procedure,
					final int n_skip_from_the_end)
    {
	final int n0=this.n_elements,ptr0=this.ptr;
	if(n_skip_from_the_end<0||n_skip_from_the_end>n0)
	{
	    throw new IllegalArgumentException("Invalid offset="+n_skip_from_the_end+
					       " n_elements="+n0+" ptr="+ptr0);
	}
	final T [] buf=this.buffer;

	// scan back, from newest to oldest trades
	for(int p=ptr0-1-n_skip_from_the_end,ne=n0-n_skip_from_the_end;ne-- >0;p--)
	{
	    if(p<0)
	    {
		p+=buf.length;
	    }
	    if(!procedure.execute(buf[p]))
	    {
		if(n0!=this.n_elements || buf!=this.buffer || ptr0!=this.ptr)
		{
		    throw new ConcurrentModificationException();
		}
		return false;
	    }
	}

	if(n0!=this.n_elements || buf!=this.buffer || ptr0!=this.ptr)
	{
	    throw new ConcurrentModificationException();
	}
	return true;
    }



    /** The method loops forward over the elements of the list,
     *   set the n_elements_to_scan=getSize(), to loop over all elements.
     */
    public boolean forwardForEachEntry(final com.polytechnik.utils.TObjectProcedure<T> procedure,
				       final int n_elements_to_scan)
    {
	final int n0=this.n_elements,ptr0=this.ptr;
	if(n_elements_to_scan<0||n_elements_to_scan>n0)
	{
	    throw new IllegalArgumentException("Invalid offset="+n_elements_to_scan+
					       " n_elements="+n0+" ptr="+ptr0);
	}
	final T [] buf=this.buffer;
	int p=ptr0-n_elements_to_scan;
	if(p<0)
	{
	    p+=buf.length;
	}

	// scan forward, from oldest to newest
	for(int ne=n_elements_to_scan;ne-- >0;p++)
	{
	    if(p>=buf.length)
	    {
		p=0;
	    }
	    if(!procedure.execute(buf[p]))
	    {
		if(n0!=this.n_elements || buf!=this.buffer || ptr0!=this.ptr)
		{
		    throw new ConcurrentModificationException();
		}
		return false;
	    }
	}

	if(n0!=this.n_elements || buf!=this.buffer || ptr0!=this.ptr)
	{
	    throw new ConcurrentModificationException();
	}
	return true;
    }






    private static class TEST implements TimeAccess {
	final long t;
	TEST(final long t)
	{
	    this.t=t;
	}
	public long getTime()
	{
	    return this.t;
	}
	static void test1()
	{
	    for(int i=0;i<100;i++)
	    {
		final CircularList<TEST> tl=new CircularList<TEST>(10);
		//tl.ptr=(int)(10*Math.random());
		tl.ptr=7;
		for(int j=0;j<7;j++)
		{
		    tl.add(new TEST(j+1000));
		}
		final int [] n_loop=new int []{0};
		final com.polytechnik.utils.TObjectProcedure<TEST> foreach_proc=
		    new com.polytechnik.utils.TObjectProcedure<TEST>(){
		    public boolean execute(final TEST b)
		    {
			n_loop[0]++;
			System.err.println("b.time="+b.getTime());
			return true;
		    }
		};
		int n_skip_from_the_end=-1;
		int n_end=-1;

		System.err.println("=========ORIG BUFFER PTR="+tl.ptr+" NELs="+tl.n_elements);
		printPrintmatchid(tl);
		final int t0=1000+(int)(10*Math.random());
		final int n_cleared=tl.clearBefore(t0,null);
		System.err.println("clearBefore: PTR="+tl.ptr+" t0="+t0+" n_cleared="+
				   n_cleared+" n_elements="+tl.getSize());
		printPrintmatchid(tl);
		n_skip_from_the_end=tl.getSize()/2;
		n_loop[0]=0;
		tl.backwardForEachEntry(foreach_proc,n_skip_from_the_end);
		System.err.println("skip="+n_skip_from_the_end+" backwardForEachEntry n_loop="+n_loop[0]+" should be "+(tl.getSize()-n_skip_from_the_end));
		if(n_loop[0]!=(tl.getSize()-n_skip_from_the_end))
		{
		    throw new RuntimeException();
		}

		n_end=tl.getSize()/2;
		n_loop[0]=0;
		tl.forwardForEachEntry(foreach_proc,n_end);
		System.err.println("skip="+n_end+" forwardForEachEntry n_loop="+n_loop[0]+" should be "+n_end);
		if(n_loop[0]!=n_end)
		{
		    throw new RuntimeException();
		}

		tl.ensureCapacity(15);
		System.err.println("ensureCapacity: PTR="+tl.ptr+" t0="+t0+" n_elements="+tl.getSize());
		printPrintmatchid(tl);

		n_loop[0]=0;
		n_skip_from_the_end=tl.getSize()/2;
		tl.backwardForEachEntry(foreach_proc,n_skip_from_the_end);
		System.err.println("skip="+n_skip_from_the_end+" backwardForEachEntry n_loop="+n_loop[0]+" should be "+(tl.getSize()-n_skip_from_the_end));
		if(n_loop[0]!=(tl.getSize()-n_skip_from_the_end))
		{
		    throw new RuntimeException();
		}

		n_end=tl.getSize()/2;
		n_loop[0]=0;
		tl.forwardForEachEntry(foreach_proc,n_end);
		System.err.println("skip="+n_end+" forwardForEachEntry n_loop="+n_loop[0]+" should be "+n_end);
		if(n_loop[0]!=n_end)
		{
		    throw new RuntimeException();
		}
	    }
	}
	private static <S extends com.polytechnik.utils.CircularList.TimeAccess> void printPrintmatchid(CircularList<S> tt)
	{
	    for(int i=0;i<tt.buffer.length;i++)
	    {
		final S t=tt.buffer[i];
		System.err.print(String.valueOf(i)+"="+(t!=null?String.valueOf(t.getTime()):"NULL")+";");
	    }
	    System.err.println();
	}
    } // END of class TEST
    public static void main(String [] args)
    {
	TEST.test1();
    }
}
