package com.polytechnik.utils;
/*
 * (C) Vladislav Malyshkin 2017
 * This file is under GPL version 3.
 *
 */

/** Build two--node Gauss quadrature and calculate
 * its characteristics as parameter--dependent values.
 * An example of such such characterists can be
 * distribution skewness and its derivatives.
 * @version $Id: DiffSkewness.java,v 1.40 2017/12/14 13:35:32 mal Exp $
 * @author Vladislav Malyshkin mal@gromco.com
 */
public class DiffSkewness {
    public final OrthogonalPolynomialsABasis M;
    public final Skewness sknodi;
    public final double di;
    final double [] NOM,SDENOM,origmoments,A,B,D,XA,L4NOM,ONE;

    DiffSkewness(final double [] NOM,final double [] SDENOM,
		 final double [] origmoments,
		 final double di,
		 final OrthogonalPolynomialsABasis M,
		 final double [] A,
		 final double [] B,
		 final double [] D,
		 final double [] XA,
		 final double [] L4NOM)
    {
	this.NOM=NOM;
	this.SDENOM=SDENOM;
	this.origmoments=origmoments;
	this.di=di;
	this.M=M;
	this.sknodi=Skewness.getSkewness(origmoments[0],
					 origmoments[1],
					 origmoments[2],
					 origmoments[3],
					 M);
	this.A=A;
	this.B=B;
	this.D=D;
	this.XA=XA;
	this.L4NOM=L4NOM;
	this.ONE=new double []{1};
        ONE[0]/=M.fx(ONE,0);
    }
    /** Calculate Skewness asympthotics.
     */
    public double getGamma1(final double x){
	final double q=M.getX0Moments(1,0)[0];
	double gamma1=(origmoments[0]-di*q)/(origmoments[0]+di*q);
	if(x>sknodi.xa){}
	else if(x<sknodi.xa){gamma1=-gamma1;}
	else gamma1=0;
	return gamma1;
    }

    /** Calculate Skewness & Sekewness derivatives. */
    public DerivativesCalculateable getGamma()
    {
	return new PolynomialsSQRatio(SDENOM,NOM,M);
    }

    /** Calculate skewness for the moments: origmoments[k]+di*Qm(p).
     */
    public Skewness getSkewnessAt(final double p)
    {
	final double [] q=M.getX0Moments(4,p);
	return Skewness.getSkewness(origmoments[0]+di*q[0],
				    origmoments[1]+di*q[1],
				    origmoments[2]+di*q[2],
				    origmoments[3]+di*q[3],
				    M);
    }

    /** Returns second order monic orthogonal polynomial coefficients.
     * The polynomial is built on parameter-dependent measure.
     */
    public DerivativesCalculateable [] getOrthogonalPolynomial2Coefs()
    {
	return new PolynomialsRatio []{
	    new PolynomialsRatio(D,A,M),
	    new PolynomialsRatio(D,B,M)
	};
    }
    /** Returns L^4 volatility (within a constant, assume a_4=0).
     */
    public DerivativesCalculateable getL4Volatility()
    {
	return new PolynomialsRatio(D,L4NOM,M);
    }
    /** Returns L^2 volatility.
     */
    public DerivativesCalculateable getL2Volatility()
    {
	return new PolynomialsRatio(ONE,D,M);
    }
    /** Returns the midpoint (x_1+x_2)/2.
     */
    public DerivativesCalculateable getX1X2midpoint()
    {
	return new PolynomialsRatio(D,TMatr.getVectorScaled(B,-0.5),M);
    }
    /** Returns quadrature nodes difference x_2-x_1.
     * This difference can be successfully used
     * in place of standard deviation. It has proper asymthotic
     * and no divergence, thus provide much better stability in most cases.
     */
    public DerivativesCalculateable getX2minusX1()
    {
	return new PolynomialsSQDRatio(D,SDENOM,M);
    }
    /** Returns x1 & x2 quadrature nodes.
     */
    public DerivativesCalculateable [] getX1X2()
    {
	final DerivativesCalculateable
	    q=new PolynomialsRatio(D,B,M),r=new PolynomialsSQDRatio(D,SDENOM,M);
	return new DerivativesCalculateable []{
	    new Sum2DerivativesCalculateable(-0.5,q,-0.5,r),
	    new Sum2DerivativesCalculateable(-0.5,q,0.5,r)
	  };
    }
    /** Parameter--dependent average (the root of first order orthogonal polynomial).
     */
    public DerivativesCalculateable getXaverage()
    {
	return new PolynomialsRatio(ONE,XA,M);
    }

    public static DiffSkewness getDiffSkewness_FromMonomials(final double m0,final double m1,final double m2,final double m3,final double dm)
    {
	final double [][] t=getNOM_SDENOM_FromMonomials(m0,m1,m2,m3,dm);
	return new DiffSkewness(t[2],t[1],t[4],dm,MB,
				t[6],t[7],t[8],t[9],t[10]);
    }
    /** Internal method! I made it public for unit test purposes only!
     * Calculates arrays for skewness calculation:
     * Skewness=NOM(p)/sqrt(SQDENOM(p)).
     * @return array with [0]: NOM, [1]: SQDENOM, [2]: NOM(calculated directly) [3]: null, [4]: original moments.
     */
    public static double [][] getNOM_SDENOM_FromMonomials(final double m0,final double m1,final double m2,final double m3,final double dm)
    {
	final double m=m0+dm,b=dm/m,
	    a=m1/m,  // aver(p)=a+b*p
	    a2=m2/m, // aver(p^2)=a2+b*p^2
	    a3=m3/m; // aver(p^3)=a3+b*p^3
	final double []
	// <p^3><p>-<p^2>^2
	    A=new double []{a3*a-a2*a2,a3*b,-2*a2*b,a*b},
	    // <p^2><p>-<p^3><1>
	    B=new double []{a2*a-a3,a2*b,a*b,-(1-b)*b},
	    // <p^2><1>-<p>^2
	    D=new double []{a2-a*a,-2*a*b,(1-b)*b},
	    XA=new double []{a,b},
	    NOM=TMatr.getVectorScaled(B,-1),
	    SDENOM=MB.getPQMultiplied(B,B);
        TMatr.sumP1(NOM,MB.getPQMultiplied(D,XA),-2);
	TMatr.sumP1(SDENOM,MB.getPQMultiplied(D,A),-4);

	/*
	// a_4 is unknown, set 0.
	final double [] L4NOM=MB.getPQMultiplied(new double []{0,0,0,0,b},D);
	TMatr.sumP1(L4NOM,MB.getPQMultiplied(new double []{a3,0,0,b},B),1);
	TMatr.sumP1(L4NOM,MB.getPQMultiplied(new double []{a2,0,b},A),1);
	//System.err.println("L4NOM="+java.util.Arrays.toString(L4NOM));
	*/

	final double [] E=new double []{
	    a3*a-a2*a2,
	    a2*a-a3*(1-b),
	    a2*(1-b)-a*a},
	    L4NOM=MB.getPQMultiplied(new double []{0,0,b},E);
	TMatr.sumP1(L4NOM,B,a3);
	TMatr.sumP1(L4NOM,A,a2);
	//System.err.println("L4NOM="+java.util.Arrays.toString(L4NOM));

	return new double [][]{NOM,SDENOM,
			       new double []{
		(a3+a*(-3*a2+2*a*a)),
		3*b*(-a2+2*a*a),
		-3*a*b*(1-2*b),
		b*(1-b)*(1-2*b)},
			       null, //3
			       new double []{
				   m0,m1,m2,m3
			       },
			       new double []{dm}, //5
			       A, //6
			       B, //7
			       D, //8
			       XA, //9
			       L4NOM //10
	};
    }



    public static final OrthogonalPolynomialsMonomialsBasis MB=Skewness.MB;
}
