package com.polytechnik.utils;
/*
 * (C) Vladislav Malyshkin 2010
 * This file is under GPL version 3.
 *
 */


/** N points E-moments calculations.
 *  @version $Id: EMoments.java,v 1.11 2010/04/13 13:32:25 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */

public class EMoments  {

    private static boolean PRINT_DATA_DEBUG=false;

    /** Return moments coefficients.
     */
    public static double [] getCoefs(final int order,final double ksi0)
    {
	if(order<2)
	{
	    throw new IllegalArgumentException("Size should be equalto at least 3, order="+order+
					       " is invalid");
	}

	final double [] coefs=getRightPart(order,ksi0);
	final double [][] matrix=getMatrix(order,ksi0);

	// now solve linear system
	final int [] ipvt=new int [order];
	final int info=Linsystems.dgefa(matrix,order,ipvt);
	if(info!=0)
	{
	    throw new RuntimeException("INFO="+info);
	}
	// rigth part is in coefs
	Linsystems.dgesl(matrix,order,ipvt,coefs,1);
	// now the solution is in coefs
	return coefs;
    }

    static double [] getRightPart(final int n,final double ksi0)
    {
	final double [] y=new double [n]; // all 0 by default
	y[0]=1.0-ksi0;
	y[1]=-1;
	return y;
    }


    static double [][] getMatrix(final int n,final double ksi0)
    {
	final double [][] m=new double [n][n]; // all 0 by default
	final double [] d0=m[0],d1=m[1];
	double kp=ksi0;
	for(int i=0;i<n;i++)
	{
	    d1[i]=(i+2)*kp;
	    kp*=ksi0;
	    d0[i]=kp;
	}
	for(int r=2;r<n;r++)
	{
	    kp=1;
	    final double [] d=m[r];
	    for(int i=r-2;i<n;i++)
	    {
		d[i]=kp*get_Ca_amk_div_Cb_bmk(i+2,r,n+2);
		kp*=ksi0;
	    }
	}
	if(PRINT_DATA_DEBUG)
	{
	    System.err.println("Printing matrix for order="+n+" ksi0="+ksi0);
	    System.err.print("I:");
	    for(int j=0;j<n;j++)
	    {
		System.err.print("\t"+j);
	    }
	    System.err.println();
	    for(int i=0;i<n;i++)
	    {
		System.err.print(String.valueOf(i)+":");
		for(int j=0;j<n;j++)
		{
		    System.err.print("\t"+m[i][j]);
		}
		System.err.println();
	    }
	}
	return m;
    }

    /** Calculate C_a^{a-k}/C_b^{b-k}.
     */
    static double get_Ca_amk_div_Cb_bmk(final int a,final int k,final int b)
    {
	double s=1;
	for(int i=k,l1=a,l2=b;i-->0;l1--,l2--)
	{
	    s*=(l1/((double)l2));
	}
	return s;
    }



    private static double [] getCoefsDirect(final int order,final double ksi0)
    {
	final double [][] m=new double [order][order];
	final double [] coefs=new double [order];
	coefs[0]=1.0-ksi0;
	coefs[1]=-1;

	for(int i=0;i<order;i++)
	{
	    m[0][i]=Math.pow(ksi0,i+2);
	    m[1][i]=(i+2)*Math.pow(ksi0,i+1);
	}
	for(int r=2;r<order;r++)
	{
	    for(int i=r-2,p=0;i<order;i++,p++)
	    {
		double s=1;
		for(int mm=0,f=i+2;mm<r;mm++,f--)
		{
		    s*=f;
		}
		m[r][i]=s*Math.pow(ksi0,p);
	    }
	}

	// now solve linear system
	final int [] ipvt=new int [order];
	final int info=Linsystems.dgefa(m,order,ipvt);
	if(info!=0)
	{
	    throw new RuntimeException("INFO="+info);
	}
	// rigth part is in coefs
	Linsystems.dgesl(m,order,ipvt,coefs,1);
	// now the solution is in this.coefs
	return coefs;
    }



    private static void checkCoefsMatched(final int order,final double ksi0)
    {
	final double [] coefs1=getCoefs(order,ksi0);
	final double [] coefs2=getCoefsDirect(order,ksi0);
	double s=0;
	for(int i=0;i<order;i++)
	{
	    s+=Math.abs(coefs1[i]-coefs2[i]);
	}
	final double eps=1e-5;
	if(!(s<=eps))
	{
	    throw new RuntimeException("Results mismatch diff="+s);
	}
	System.err.println("checkCoefsMatched order="+order+" ksi0="+ksi0+" diff="+s);
    }



    static void printCoefsAndInterpolation(final int order,final double ksi0)
    {
	final double [] coefs=getCoefs(order,ksi0);
	System.err.println("Printing coefs for order="+order+" ksi0="+ksi0);
	for(int i=0;i<coefs.length;i++)
	{
	    System.err.println("coefs\t"+i+"\t"+coefs[i]);
	}
	System.err.println("#ksi(o="+order+" ksi0="+ksi0+")\t1-ksi\tsum bn ksi^{n+2}");
	for(double ksi=0;ksi<=4+0.5e-2;ksi+=1e-2)
	{
	    double s=0,ksicur=ksi*ksi;
	    for(int i=0;i<coefs.length;i++)
	    {
		s+=coefs[i]*ksicur;
		ksicur*=ksi;
	    }
	    System.err.println(ksi+"\t"+(1-ksi)+"\t"+s);
	}
    }






    public static void main(String [] args)
    {
	final int n=args.length>0?Integer.parseInt(args[0]):10;

	for(double ksi0=1;ksi0<5.1;ksi0+=1)
	{
	    for(int i=3;i<10;i++)
	    {
		checkCoefsMatched(i,ksi0);
	    }
	}
	printCoefsAndInterpolation(n,1);
	printCoefsAndInterpolation(n,2);
	printCoefsAndInterpolation(n,5);
    }

}
