package com.polytechnik.utils;
/*
 * (C) Vladislav Malyshkin 2014
 * This file is under GPL version 3.
 *
 */

/** Convert to spectrum.
 *  @version $Id: EVXData.java,v 1.67 2018/06/01 08:16:13 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
public class EVXData {
    public final double sL,sH,wL,wH,s0,o_o,x0;
    public final double [] psiL,psiH,psi0,ev,psik_at_x0;
    public final int neq;

    public EVXData(final int neq,
		   final double wL,final double sL,final double [] psiL,
		   final double wH,final double sH,final double [] psiH,
		   final double [] psi0,final double [] ev,final double [] psik_at_x0,
		   final double s0,final double o_o,final double x0)
    {
	this.neq=neq;
	this.wL=wL;
	this.sL=sL;
	this.psiL=psiL;
	this.wH=wH;
	this.sH=sH;
	this.psiH=psiH;
	this.psi0=psi0;
	this.ev=ev;
	this.psik_at_x0=psik_at_x0;
	this.s0=s0;
	this.o_o=o_o;
	this.x0=x0;
    }
public double [] getPsi(final int ind_lev)
{
    if(ind_lev<0||ind_lev>=neq)
	throw new RuntimeException("ind_lev="+ind_lev+" neq="+neq);
    return getPsiFromBasis(ind_lev,this.ev,neq);
}
/** First order perturbation theory to psi_{ind_lev}.
 *  @return sum_{k=0;k!=ind_lev}^{k=neq-1} |psi_k><psi_k|QQ|psi_{ind_lev}>/(ev[ind_lev]-ev[k])
 */
public double [] getPsiPertr(final int ind_lev,final double [] QQ, final int ldQQ)
{
    if(ind_lev<0||ind_lev>=neq||ldQQ<neq)
	throw new RuntimeException("ind_lev="+ind_lev+" neq="+neq+" ldQQ="+ldQQ);
    final double [] dpsi=new double [neq];
    for(int k=0;k<neq;k++)
    {
	if(k!=ind_lev)
	{
	    final double s=getFPsiPsi(ind_lev,k,QQ,ldQQ)/(this.ev[ind_lev]-this.ev[k]);
	    TMatr.sumP1(dpsi,this.ev,neq+k*neq,neq,s);
	}
    }
    return dpsi;
}

/** Calculates <psi_{ind_lev}|FQQ|psi_{ind_lev}>
 */
public double getFPsi(final int ind_lev,final double [] FQQ, final int ldFQQ)
{
    return getFPsiPsi(ind_lev,ind_lev,FQQ,ldFQQ);
}
/** Calculates <psi_{ind_lev}|FQQ|psi>
 */
public double getFPsiPsi(final int ind_lev,final double [] psi,final double [] FQQ, final int ldFQQ)
{
    if(psi.length!=neq||ind_lev<0||ind_lev>=neq||ldFQQ<neq) throw new IllegalArgumentException();
    return TMatr.getAMatrB(neq,this.ev,neq+ind_lev*neq,neq,psi,0,FQQ,ldFQQ);
}
/** Calculates <psi_i|FQQ|psi_j>
 */
public double getFPsiPsi(final int i,final int j,final double [] FQQ, final int ldFQQ)
{
    if(i<0||i>=neq||j<0||j>=neq||ldFQQ<neq) throw new IllegalArgumentException();
    return TMatr.getAMatrB(neq,this.ev,neq+i*neq,neq,this.ev,neq+j*neq,FQQ,ldFQQ);
}

/** Calculates Perturbation term
\sum_k <psi_{ind_lev}|VQQ|psi_k><psi_{ind_lev}|FQQ|psi_k>/(ev[ind_lev]-ev[k])
corresponding to <psi_{ind_lev}|FQQ|var psi_{ind_lev}> term when the hamiltonian
perturbation is V.
 */
public double getPertr1Term(final int ind_lev,final double [] VQQ,final double [] FQQ, final int ldQQ)
{
    double s=0;
    for(int k=0;k<neq;k++)
    {
	if(k!=ind_lev)
	{
	    s+=getFPsiPsi(ind_lev,k,VQQ,ldQQ)*getFPsiPsi(k,ind_lev,FQQ,ldQQ)/(this.ev[ind_lev]-this.ev[k]);
	}
    }
    return s;
}

public static double [] getMSign(final double [] a,final double sign)
{
    final double [] b=a.clone();
    if(Math.signum(b[b.length-1])==Math.signum(sign)) return b;
    for(int k=b.length;--k>=0;) b[k]=-b[k];
    return b;
}

public static void test_orthogonality(
    final String txt,
    final int n,final double [] basis,final double [] QQi,final double [] QQt,
    final double eps)
{
    test_orthogonality(txt,n,basis,QQi,QQt,false,eps);
}
public static void test_orthogonalitySKIP0I(
    final String txt,
    final int n,final double [] basis,final double [] QQi,final double [] QQt,
    final double eps)
{
    test_orthogonality(txt,n,basis,QQi,QQt,true,eps);
}
private static void test_orthogonality(
    final String txt,
    final int n,final double [] basis,final double [] QQi,final double [] QQt,
    final boolean skip_I0_orthogonality_test,final double eps)
{
    for(int k=0;k<n;k++)
    {
	final double [] psik=EVXData.getPsiFromBasis(k,basis,n);
	for(int l=0;l<n;l++)
	{
	    final double [] psil=EVXData.getPsiFromBasis(l,basis,n);
	    UnitTests.checkDoubleRel(TMatr.getAMatrB(psik,psil,QQt,n),"<psik|psik>",l==k?1:0,l==k?"1":"0",eps);
	    if(k==l)
		UnitTests.checkDoubleRel(TMatr.getAMatrB(psik,psil,QQi,n),"<psik|I|psik>",basis[k],"basis[k]",eps);
	    else if(!skip_I0_orthogonality_test || (k!=0 && l!=0))
		UnitTests.checkDoubleRel(TMatr.getAMatrB(psik,psil,QQi,n),"<psik|I|psil>",0,"0",eps);
	}
    }
    System.err.println(txt+":test_orthogonality"+(skip_I0_orthogonality_test?" With 0-th element skipped":"")+" test OK n="+n+" eps="+eps);
}


public static void matchEVXData(final EVXData a,final EVXData b,final double eps)
{
    matchEVXData(a,b,true,eps);
}
public static void matchEVXData(final EVXData a,final EVXData b,final boolean flag_psicheck,final double eps)
{
    matchEVXData(a,b,flag_psicheck,true,eps);
}
public static void matchEVXData(final EVXData a,final EVXData b,final boolean flag_psicheck,final boolean flag_0check,final double eps)
{
    UnitTests.checkDoubleRel(a.neq,"a.neq",b.neq,"b.neq",0);
    UnitTests.checkDoubleArrayRelFirstindex(a.neq,a.ev,"a.ev",b.ev,"b.ev",eps);
    UnitTests.checkDoubleRel(a.sL,"a.sL",b.sL,"b.sL",eps);
    UnitTests.checkDoubleRel(a.sH,"a.sH",b.sH,"b.sH",eps);
    if(flag_0check)
    {
	UnitTests.checkDoubleRel(a.o_o,"a.o_o",b.o_o,"b.o_o",eps);
	UnitTests.checkDoubleRel(a.s0,"a.s0",b.s0,"b.s0",eps);
	UnitTests.checkDoubleRel(Math.abs(a.wL),"|a.wL|",Math.abs(b.wL),"|b.wL|",eps);
	UnitTests.checkDoubleRel(Math.abs(a.wH),"|a.wH|",Math.abs(b.wH),"|b.wH|",eps);
    }
    if(flag_psicheck)
    {
    UnitTests.checkDoubleArrayRel(a.psiL,"a.psiL",getMSign(b.psiL,a.psiL[a.neq-1]),"b.psiL",eps);
    UnitTests.checkDoubleArrayRel(a.psiH,"a.psiH",getMSign(b.psiH,a.psiH[a.neq-1]),"b.psiH",eps);
    UnitTests.checkDoubleArrayRel(a.psi0,"a.psi0",getMSign(b.psi0,a.psi0[a.neq-1]),"b.psi0",eps);
    }
    System.err.println("EVXData:matchEVXData test OK neq="+a.neq+" eps="+eps);
}

public static EVXData getEVXDataFromQQ(final int n,final double [] QQf,final double [] QQa,final OrthogonalPolynomialsABasis M)
{
    return _getEVXDataFromQQ(n,M.EV.getGEV_Generalized(n,QQf,n,QQa,n,true),M,M.getRadauRoot());
}
public static EVXData getEVXDataFromQQfDiagQQa(final int n,final double [] QQf,final double [] QQa,final OrthogonalPolynomialsABasis M)
{
    return _getEVXDataFromQQ(n,getSymEVDiagT(n,QQf,QQa,M.EV),M,M.getRadauRoot());
}

private static EVXData _getEVXDataFromQQ(final int n,final double [] evS,final OrthogonalPolynomialsABasis M,
					 final double x0)
{

    final double [] psiL=new double [n],psiH=new double [n],psi0=new double [n],psik_at_x0=new double [n];
    final double [] r=psiL; // use as a temporary array
    double s_s_o_o=0,ss0=0,s_wL=0,s_wH=0;
    for(int k=0;k<n;k++)
    {
	System.arraycopy(evS,n+(k)*n,r,0,n);
	final double fkx0=M.fx(r,x0),f2kx0=fkx0*fkx0;
	psik_at_x0[k]=fkx0;
	TMatr.sumP1(psi0,r,fkx0);
	if(k<=0) s_wL=fkx0;
	if(k>=n-1) s_wH=fkx0;
	s_s_o_o+=f2kx0;
	ss0+=f2kx0*evS[k];
    }
    final double sL=evS[0],sH=evS[n-1];
    System.arraycopy(evS,n+(0)*n,psiL,0,n);
    System.arraycopy(evS,n+(n-1)*n,psiH,0,n);
    final double sqrt_s_s_o_o=Math.sqrt(s_s_o_o);
    final double wL=s_wL/sqrt_s_s_o_o,wH=s_wH/sqrt_s_s_o_o,s0=ss0/s_s_o_o;
    for(int k=0;k<n;k++) psi0[k]/=sqrt_s_s_o_o;

    return new EVXData(n,wL,sL,psiL,wH,sH,psiH,psi0,evS,psik_at_x0,s0,s_s_o_o,x0);
}


public static double [] getPsiFromBasis(final int ind,final double [] basis,final int n)
{
    if(ind<0||ind>=n||n*n+n!=basis.length) throw new IllegalArgumentException();
    return getPsiFromBasis(ind,basis,n,n);
}

public static double [] getPsiFromBasis(final int ind,final double [] basis,final int n,final int basis_offset)
{
    final double [] psi=new double [n];
    System.arraycopy(basis,basis_offset+ind*n,psi,0,n);
    return psi;
}


/**
 * Given b_i=\sum_k eVB_{ik} Q_k abd <b_i b_j>=1, <Q_k Q_l>=G_{lk},
 * then Q_k = \sum_l <Q_k Q_l> eVB_lk and given a matrix matrB=<b_i|f|b_j>
 * the matrix <Q_k|f|Q_s> = \sum_{l,i,j,m} G_{kl} eVB_il <b_i|f|b_j> eVB_jm G_{ms}.
 * The method calculates G eV^T B eV G^T.
 */
public static void convertMatrFromBToQ_Basis(final double [] res,final int n,final double [] matrB,final double [] eVB,final double [] G)
{
    final double [] GeVT=new double [n*n],tmp=new double [n*n];
    final int basis_offset=n;
    TMatr.setMA_x_MBTrans(GeVT,n,G,n,0+0,n,eVB,n,n,basis_offset+0*n);
    convertMatrFromQToB_Basis(res,n,matrB,GeVT,n,0+0,tmp);
}


/** Calculate eV matrQ eV^T
 */
public static void convertMatrFromQToB_Basis(final double [] res,final int n,final double [] matrQ,final double [] eVB,final double [] tmp)
{
    //case n=nd, when the number
    convertMatrFromQToB_Basis(res,n,matrQ,eVB,n,n+0,tmp);
}

/** Transorm quadratic form matrix to new basis.
 *  @param res Output, nd x nd matrix, res=basis x matrQ x basis^T.
 *  @param n Original basis dimesion.
 *  @param matrQ Matrix in original basis, n x n matrix lD=n.
 *  @param basis New basis, nd x n matrix, lD=n, possible data offset basis_offset.
 *  @param nd New basis dimension.
 *  @param basis_offset Offset of the baisis.
 *  @param tmp Temporary array, at least nd*n size.
 */
public static void convertMatrFromQToB_Basis(final double [] res,final int n,final double [] matrQ,final double [] basis,final int nd,final int basis_offset,final double [] tmp)
{
    // matrQ: n x n matrix in original Q basis.
    // basis: nd x n matrix original Q basis.
    // res: nd x nd matrix in new basis
    TMatr.setMA_x_MBTrans(tmp,n,matrQ,n,0,n,basis,nd,n,basis_offset+0*n); // tmp: n x nd
    TMatr.setMA_x_MB(res,nd,basis,n,basis_offset+0*n,n,tmp,nd,nd,0); // res: nd x nd
}

/** Calculate basis QQt matrQ basis^T
 */
public static void convertOperatorFromQToB_Basis(final double [] res,final int n,final double [] matrQ,final double [] QQt,final double [] basis,final int basis_offset,final double [] tmp)
{
    // matrQ: n x n matrix in original Q basis.
    TMatr.setMA_x_MBTrans(tmp,n,matrQ,n,0,n,basis,n,n,basis_offset+0*n);
    TMatr.setMA_x_MB(res,n,QQt,n,0,n,tmp,n,n,0);
    TMatr.setMA_x_MB(tmp,n,basis,n,basis_offset+0*n,n,res,n,n,0);
    System.arraycopy(tmp,0,res,0,n*n);
}

/** Calculates eVB (ndxldQQ) x QQT (ldQQxldQQ) x psiQ.
 */
public static double [] convertPsiFromQToB_Basis(final double [] psiQ,final double [] QQt,final int ldQQ,final double [] eVB,final int nd,final int basis_offset)
{
    final double [] tmp=new double [ldQQ];
    TMatr.setMA_x_MB(tmp,ldQQ,QQt,ldQQ,0,psiQ.length,psiQ,1,1,0); // tmp: 1xldQQ
    final double [] psiB=new double [nd];
    TMatr.setMA_x_MB(psiB,nd,eVB,ldQQ,basis_offset,ldQQ,tmp,1,1,0); // psiB: nd x 1
    return psiB;
}
/** Calculate eVB G psiQ
 */
public static double [] convertPsiFromQToB_Basis(final int n,final double [] psiQ,final double [] QQt,final double [] eVB)
{
    return convertPsiFromQToB_Basis(psiQ,QQt,n,eVB,n,n);
}
public static double [] convertPsiFromBToQ_Basis(final int n,final double [] psiB,final double [] eVB)
{
    final double [] psiQ=new double [n];
    TMatr.setMATrans_x_MB(psiQ,n,eVB,n,n+0*n,n,psiB,1,1,0);
    //TMatr.setMA_x_MB(psiQ,1,psiB,n,0,n,eVB,n,n,n+0*n); // same thing
    return psiQ;
}
public static double [] convertQX0FromQToB_Basis(final int n,final double [] QX0,final double [] eVB)
{
    final double [] BX0=new double [n];
    TMatr.setMA_x_MB(BX0,n,eVB,n,n+0*n,n,QX0,1,1,0);
    return BX0;
}

public static void shiftPOffset(final int nmoms,final double [] p0moms,final double [] p1moms,final double [] p2moms,final double dpo)
{
    shiftPOffset(nmoms,p0moms,p1moms,p2moms,null,dpo);
}

public static void shiftPOffset(final int nmoms,final double [] p0moms,final double [] p1moms,final double [] p2moms,final double [] p3moms,final double dpo)
{
    for(int k=0;k<nmoms;k++)
    {
	if(p3moms!=null) p3moms[k]+=dpo*(-3*p2moms[k]+dpo*(3*p1moms[k]-dpo*p0moms[k]));
	if(p2moms!=null) p2moms[k]+=dpo*(-2*p1moms[k]+dpo*p0moms[k]);
	if(p1moms!=null) p1moms[k]+=(-dpo)*p0moms[k];
    }
}

/** Solve generalised eigenvalues proglem,
 *  assuming QQt is diagonal.
 */
public static double [] getSymEVDiagT(final int n,final double [] QQv,final double [] QQt,final EVSolver MEV)
{
    final int oqq=n*n;
    final double [] qq=new double [oqq+n];
    // here we assume the QQt is diagonal.
    for(int k=0;k<n;k++)
    {
	final int ok=k*n;
	final double diagk=QQt[ok+k];
	if(!(diagk>0))
	{
	    java.util.Arrays.fill(qq,Double.NaN);
	    return qq;
	}
	final double s=qq[oqq+k]=Math.sqrt(diagk);
	for(int l=0;l<=k;l++)
	{
	    qq[ok+l]=qq[l*n+k]=QQv[ok+l]/(s*qq[oqq+l]);
	}
    }
    final double [] ev=MEV.getEV(n,qq,n,true);
    // convert back to our basis
    for(int k=0;k<n;k++)
    {
	for(int l=0,o=(k+1)*n;l<n;l++,o++) ev[o]=ev[o]/(qq[oqq+l]);
    }
    return ev;
}


} // END OF class EVXData
