package com.polytechnik.utils;
/*
 * (C) Vladislav Malyshkin 2010
 * This file is under GPL version 3.
 *
 */

/** N points decaying exponenets interpolation.
 *  @version $Id: ExponentialInterpolation.java,v 1.10 2011/11/05 19:31:52 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */

public class ExponentialInterpolation extends VandermondeInterpolation  {

    public ExponentialInterpolation(final int order)
    {
	super(order);
    }
    protected double [][] getMatrix(final int o)
    {
	final double [][] m=new double [o][];
	for(int i=0;i<o;i++)
	{
	    final double [] l=m[i]=new double [o];
	    final double x=((double)(i+1))/(1+o); // equal distance
	    l[0]=1;
	    l[1]=Math.exp(-x);
	    for(int j=2;j<o;j++)
	    {
		// equivalent to Vandermonde with 1,x,x^2,x^4,x^8,... basis and non-uniform mesh
		l[j]=l[j-1]*l[j-1];
	    }
	}
	return m;
    }

    public double getF(final double x)
    {
	double s=this.rightpart[0],e=Math.exp(-x);
	for(int i=1;i<this.order;i++)
	{
	    s+=e*this.rightpart[i];
	    e*=e;
	}
	return s;
    }
    public double getDF(final double x)
    {
	double s=0,e=Math.exp(-x),f=-1;
	for(int i=1;i<this.order;i++)
	{
	    s+=this.rightpart[i]*e*f;
	    e*=e;
	    f*=2;
	}
	return s;
    }
    public double getD2F(final double x)
    {
	double s=0,e=Math.exp(-x),f=-1;
	for(int i=1;i<this.order;i++)
	{
	     s+=this.rightpart[i]*e*f*f;
	     e*=e;
	     f*=2;
	}
	return s;
    }
    public double getD3F(final double x)
    {
	double s=0,e=Math.exp(-x),f=-1;
	for(int i=1;i<this.order;i++)
	{
	     s+=this.rightpart[i]*e*f*f*f;
	     e*=e;
	     f*=2;
	}
	return s;
    }
    public double getFConst()
    {
	return this.rightpart[0];
    }






    private static void checkExactExponents()
    {
	final ExponentialInterpolation v=new ExponentialInterpolation(5);
	final double [] fm=new double [v.order+1],coefs=new double [v.order];
	final java.util.Random r=new java.util.Random(-79091);
	final double eps=1e-6;
	int n=0;
	for(;n<100000;n++)
	{
	    // random exponent coefficinets
	    for(int j=0;j<coefs.length;j++)
	    {
		coefs[j]=r.nextDouble();
	    }
	    for(int j=1;j<fm.length;j++)
	    {
		final double x=((double)j)/(v.order+1);
		double s=coefs[0],f=1;
		for(int k=1;k<v.order;k++)
		{
		    s+=coefs[k]*Math.exp(-f*x);
		    f*=2;
		}
		fm[j]=s;
	    }
	    v.setFMoments(fm);

	    // match exact values of v.order-1 exponents
	    for(double x=-1.1;x<1.1;x+=1e-2)
	    {
		double sf=coefs[0],sd1f=0,sd2f=0,sd3f=0,f=-1;
		for(int k=1;k<v.order;k++)
		{
		    final double ex=Math.exp(f*x);
		    sf+=coefs[k]*ex;
		    sd1f+=coefs[k]*ex*f;
		    sd2f+=coefs[k]*ex*f*f;
		    sd3f+=coefs[k]*ex*f*f*f;
		    f*=2;
		}

		// match values
		UnitTests.checkDouble(sf,"sf",v.getF(x),"v.getF(x)",eps);
		UnitTests.checkDouble(sd1f,"sd1f",v.getDF(x),"v.getDF(x)",eps);
		UnitTests.checkDouble(sd2f,"sd2f",v.getD2F(x),"v.getD2F(x)",eps);
		UnitTests.checkDouble(sd3f,"sd3f",v.getD3F(x),"v.getD3F(x)",eps);
		UnitTests.checkDouble(coefs[0],"a0",v.getFConst(),"v.getFConst()",eps);
	    }
	}
	System.err.println("checkExactPolinomial() "+n+
			   " test OK values matched order="+v.order+".");
    }



    private static double f1(final double x)
    {
	return x*x+2-3*x;
    }
    private static double f2(final double x)
    {
	return Math.sin(4*x);
    }

    private static void printValues()
    {
	System.err.println("#x\tf1\tf1_int\tf2\tf2_int");
	final ExponentialInterpolation v1=new ExponentialInterpolation(5),
	    v2=new ExponentialInterpolation(5);
	final double [] fm1=new double [v1.order+1],fm2=new double [v2.order+1];
	for(int j=1;j<fm1.length;j++)
	{
	    final double x=((double)j)/(v1.order+1);
	    fm1[j]=f1(x);
	    fm2[j]=f2(x);
	}
	v1.setFMoments(fm1);
	v2.setFMoments(fm2);
	for(double x=0;x<1;x+=1e-3)
	{
	    System.err.println(String.valueOf(x)+"\t"+
			       f1(x)+"\t"+
			       v1.getF(x)+"\t"+
			       f2(x)+"\t"+
			       v2.getF(x));
	}
    }



    public static void main(String [] args)
    {
	if(args.length==0)
	{
	    // run tests
	    checkValues(new ExponentialInterpolation(7));
	    checkExactExponents();
	}
	else if(args[0].equals("show values"))
	{
	    printValues();
	}
	else
	{
	    throw new RuntimeException("Unknown argumenet "+args[0]);
	}
    }
}
