package com.polytechnik.utils;
/*
 * (C) Vladislav Malyshkin 2010
 * This file is under GPL version 3.
 *
 */


/** N points F-moments calculations.
 *  @version $Id: FMoments.java,v 1.17 2013/03/08 16:50:47 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */

public class FMoments  {

    /** FMoments, i-th of d fmoment is at [d][i] element, d-norm is at [d][0] element. */
    public final double [][] fmoments;
    FMoments(final double [][] fmoments)
    {
	this.fmoments=fmoments;
    }

    /** Calculate all F-moments from 1 to n.
     *  @param n_from Array index to start from (inclusive).
     *  @param n_to Array index to end to (exclusive).
     *  @param weight Point weights [n_from .. n_to-1] will be used.
     *  @param values Point values [n_from .. n_to-1] will be used.
     *  @param n Maximal F-moment order to calculate.
     *  @return Calculated F-moments.
     *  Given k=0..n-1 n calculate
     * fm(k)=\sum f(ind_k) dv_0 dv_1 .. dv_{n-1}.
     * here we sum over all n-points
     * ind_0<ind_1<..ind_{n-1}.
     * Case n=2 give Kendall-like statistics.
     */
    public static FMoments getFMoments(
	final int n_from,
	final int n_to,
	final double [] weights,
	final double [] values,
	final int n)
    {
	if(weights.length!=values.length)
	{
	    throw new IllegalArgumentException();
	}
	double wscale=0; // crude attempt to reduce chances of overflow
	for(int i=n_to;i-->n_from;)
	{
	    final double w0=weights[i];
	    if(w0<=0)
	    {
		throw new IllegalArgumentException("nagative weight");
	    }
	    wscale+=w0;
	}

	final double [][] sums=new double [n+1][];
	for(int pn=n;pn>=0;pn--)
	{
	    sums[pn]=new double [pn+1]; // all 0 by default
	}
	sums[0][0]=1;

	// recurrent summs
	for(int i=n_to;i-->n_from;)
	{
	    final double w=weights[i]/wscale,wf=w*values[i];
	    for(int pn=n;pn>=1;pn--)
	    {
		final double [] spn=sums[pn],spn1=sums[pn-1];
		for(int pf=pn;pf>=2;pf--)
		{
		    spn[pf]+=w*spn1[pf-1];
		}
		spn[1]+=wf*spn1[0];
		spn[0]+=w*spn1[0];
	    }
	}
	// normalize F-moments
	for(int pn=n;pn>=1;pn--)
	{
	    final double [] spn=sums[pn];
	    final double norm=spn[0]>0?spn[0]:Double.NaN;
	    for(int pf=pn;pf>=1;pf--)
	    {
		spn[pf]/=norm;
		spn[0]*=wscale; // restore original scale, may overflow
	    }
	}
	return new FMoments(sums);
    }



/** Results:
============== CHECK RESULTS. N DATAPOINTS=93
fm.fmoments[1][0]=167.35156657702672 norma1=167.3515665770267
fm.fmoments[1][1]=-0.009100061042279447 f1a1=-0.00910006104227945

fm.fmoments[2][0]=13847.89620897265 norma2=13847.896208972668
fm.fmoments[2][1]=-0.015072509709044437 f1a2=-0.015072509709044458
fm.fmoments[2][2]=-0.0032083324005653243 f2a2=-0.0032083324005653265

fm.fmoments[5][0]=9.772134177286938E8 norma5=9.772134177288375E8
fm.fmoments[5][1]=-0.01475258635796231 f1=-0.014752586357960193
fm.fmoments[5][2]=-0.026893548468181972 f2=-0.026893548468179984
fm.fmoments[5][3]=-0.014779002408428617 f3=-0.014779002408425742
fm.fmoments[5][4]=0.0169419095286922 f4=0.016941909528688372
fm.fmoments[5][5]=-0.006827901970356536 f5=-0.006827901970356409
============== END RESULTS
.
*/
    private static void checkCalcs()
    {
	final int points=93;

	final java.util.Random r=new java.util.Random(-79091);
	final double [] w=new double [points],f=new double [points];

	// init points from random array
	for(int i=0;i<points;i++)
	{
	    w[i]=r.nextDouble()+1.3;
	    f[i]=r.nextDouble()-0.5;
	}
	final FMoments fm=getFMoments(0,w.length,w,f,5);


	double ns=0,f1=0,f2=0,f3=0,f4=0,f5=0;
	double nsa2=0,f1a2=0,f2a2=0,nsa1=0,f1a1=0;

	for(int i1=0;i1<points;i1++)
	{
	    final double wpa1=w[i1];
	    nsa1+=wpa1;
	    f1a1+=f[i1]*wpa1;

	    for(int i2=i1+1;i2<points;i2++)
	    {
		final double wpa=w[i1]*w[i2];
		nsa2+=wpa;
		f1a2+=f[i1]*wpa;
		f2a2+=f[i2]*wpa;

		for(int i3=i2+1;i3<points;i3++)
		{
		    for(int i4=i3+1;i4<points;i4++)
		    {
			for(int i5=i4+1;i5<points;i5++)
			{
			    final double wp=w[i1]*w[i2]*w[i3]*w[i4]*w[i5];
			    ns+=wp;
			    f1+=f[i1]*wp;
			    f2+=f[i2]*wp;
			    f3+=f[i3]*wp;
			    f4+=f[i4]*wp;
			    f5+=f[i5]*wp;
			}
		    }
		}
	    }
	}

	final double norma5=ns>0?ns:Double.NaN,norma2=nsa2>0?nsa2:Double.NaN,norma1=nsa1>0?nsa1:0,eps=1e-9;
	f1/=norma5;
	f2/=norma5;
	f3/=norma5;
	f4/=norma5;
	f5/=norma5;

	f1a2/=norma2;
	f2a2/=norma2;

	f1a1/=norma1;


	System.err.println("============== CHECK RESULTS. N DATAPOINTS="+points);
	System.err.println("fm.fmoments[1][0]="+fm.fmoments[1][0]+" norma1="+norma1);
	System.err.println("fm.fmoments[1][1]="+fm.fmoments[1][1]+" f1a1="+f1a1);
	System.err.println();
	System.err.println("fm.fmoments[2][0]="+fm.fmoments[2][0]+" norma2="+norma2);
	System.err.println("fm.fmoments[2][1]="+fm.fmoments[2][1]+" f1a2="+f1a2);
	System.err.println("fm.fmoments[2][2]="+fm.fmoments[2][2]+" f2a2="+f2a2);
	System.err.println();
	System.err.println("fm.fmoments[5][0]="+fm.fmoments[5][0]+" norma5="+norma5);
	System.err.println("fm.fmoments[5][1]="+fm.fmoments[5][1]+" f1="+f1);
	System.err.println("fm.fmoments[5][2]="+fm.fmoments[5][2]+" f2="+f2);
	System.err.println("fm.fmoments[5][3]="+fm.fmoments[5][3]+" f3="+f3);
	System.err.println("fm.fmoments[5][4]="+fm.fmoments[5][4]+" f4="+f4);
	System.err.println("fm.fmoments[5][5]="+fm.fmoments[5][5]+" f5="+f5);
	System.err.println("============== END RESULTS");

	// check constrains
	UnitTests.checkDouble(fm.fmoments[1][1],"fm.fmoments[1][1]",f1a1,"f1a1",eps);
	UnitTests.checkDouble(fm.fmoments[1][0]/norma1,"fm.fmoments[1][0]/norma1",nsa1/norma1,"nsa1/norma1",eps);

	UnitTests.checkDouble(fm.fmoments[2][1],"fm.fmoments[2][1]",f1a2,"f1a2",eps);
	UnitTests.checkDouble(fm.fmoments[2][2],"fm.fmoments[2][2]",f2a2,"f2a2",eps);
	UnitTests.checkDouble(fm.fmoments[2][0]/norma2,"fm.fmoments[2][0]/norma2",nsa2/norma2,"nsa2/norma2",eps);

	UnitTests.checkDouble(fm.fmoments[5][1],"fm.fmoments[5][1]",f1,"f1",eps);
	UnitTests.checkDouble(fm.fmoments[5][2],"fm.fmoments[5][2]",f2,"f2",eps);
	UnitTests.checkDouble(fm.fmoments[5][3],"fm.fmoments[5][3]",f3,"f3",eps);
	UnitTests.checkDouble(fm.fmoments[5][4],"fm.fmoments[5][4]",f4,"f4",eps);
	UnitTests.checkDouble(fm.fmoments[5][5],"fm.fmoments[5][5]",f5,"f5",eps);
	UnitTests.checkDouble(fm.fmoments[5][0]/norma5,"fm.fmoments[5][0]/norma5",ns/norma5,"ns/norma5",eps);

    }

/** Results:
============== CHECK DIST RESULTS. N DATAPOINTS=1000 DMAX=7
M1:f[1][0]=1000.0 f[1][1]=499.5
DM1 exact=1/2=0.5:
M2:f[2][0]=499500.0 f[2][1]=332.6666666666667 f[2][2]=666.3333333333334
DM2 exact=1/3=0.3333333333333333:df(2-1)=0.33333333333333337
M3:f[3][0]=1.66167E8 f[3][1]=249.25 f[3][2]=499.5 f[3][3]=749.75
DM3 exact=1/4=0.25:df(2-1)=0.25 df(3-2)=0.25
M4:f[4][0]=4.141712475E10 f[4][1]=199.2 f[4][2]=399.4 f[4][3]=599.6 f[4][4]=799.8
DM4 exact=1/5=0.2:df(2-1)=0.19999999999999998 df(3-2)=0.20000000000000004 df(4-3)=0.19999999999999993
M5:f[5][0]=8.2502912502E12 f[5][1]=165.83333333333334 f[5][2]=332.6666666666667 f[5][3]=499.5 f[5][4]=666.3333333333334 f[5][5]=833.1666666666666
DM5 exact=1/6=0.16666666666666666:df(2-1)=0.16666666666666669 df(3-2)=0.16666666666666666 df(4-3)=0.1666666666666667 df(5-4)=0.1666666666666666
M6:f[6][0]=1.3681732989915E15 f[6][1]=142.00000000000017 f[6][2]=285.0 f[6][3]=427.99999999999983 f[6][4]=571.0 f[6][5]=713.9999999999999 f[6][6]=856.9999999999999
DM6 exact=1/7=0.14285714285714285:df(2-1)=0.14285714285714268 df(3-2)=0.14285714285714268 df(4-3)=0.14285714285714302 df(5-4)=0.14285714285714274 df(6-5)=0.14285714285714285
M7:f[7][0]=1.9428060845679296E17 f[7][1]=124.12500000000011 f[7][2]=249.2500000000004 f[7][3]=374.375 f[7][4]=499.50000000000017 f[7][5]=624.6249999999999 f[7][6]=749.7499999999999 f[7][7]=874.8750000000005
DM7 exact=1/8=0.125:df(2-1)=0.12500000000000028 df(3-2)=0.1249999999999996 df(4-3)=0.12500000000000017 df(5-4)=0.12499999999999972 df(6-5)=0.125 df(7-6)=0.12500000000000056
============== END CHECK DIST
.
*/
    private static void checkDist()
    {
	final int points=1000,dmax=7;
	final double [] w=new double [points],f=new double [points];

	// init points
	for(int i=0;i<points;i++)
	{
	    w[i]=1;
	    f[i]=i;
	}
	final FMoments fm=getFMoments(0,w.length,w,f,dmax);

	System.err.println("============== CHECK DIST RESULTS. N DATAPOINTS="+points+" DMAX="+dmax);
	for(int i=1;i<fm.fmoments.length;i++)
	{
	    final double [] fmd=fm.fmoments[i];
	    System.err.print("M"+i+":");
	    for(int j=0;j<fmd.length;j++)
	    {
		System.err.print("f["+i+"]["+j+"]="+fmd[j]);
		System.err.print(" ");
	    }
	    System.err.println();
	    System.err.print("DM"+i+" exact=1/"+(i+1)+"="+(1.0/(1+i))+":");
	    for(int j=2;j<fmd.length;j++)
	    {
		System.err.print("df("+j+"-"+(j-1)+")="+((fmd[j]-fmd[j-1])/(1+points)));
		System.err.print(" ");
	    }
	    System.err.println();
	}
	System.err.println("============== END CHECK DIST");
    }


    static void printValues()
    {
	final int npoints=9999,dmax=7;
	final double scale=1+npoints;
	final double [] w=new double [npoints],f=new double [npoints];
	java.util.Arrays.fill(w,1);
	System.err.print("#x");
	for(int pn=1;pn<=dmax;pn++)
	{
	    System.err.print("\tf"+pn);
	}
	System.err.println();
	for(int i=0;i<npoints;i++)
	{
	    if(i>0)
	    {
		f[i-1]=0;
	    }
	    f[i]=1*scale;
	    final FMoments fm=getFMoments(0,npoints,w,f,dmax);
	    System.err.print(i/scale);
	    for(int pn=1;pn<=dmax;pn++)
	    {
		System.err.print("\t"+fm.fmoments[dmax][pn]);
	    }
	    System.err.println();
	}
    }



    public static void main(String [] args)
    {
	if(args.length==0)
	{
	    // run tests
	    checkCalcs();
	    checkDist();
	}
	else if(args[0].equals("show values"))
	{
	    printValues();
	}
	else
	{
	    throw new RuntimeException("Unknown argumenet "+args[0]);
	}
    }

}
