package com.polytechnik.utils;
/*
 * (C) Vladislav Malyshkin 2011
 * This file is under GPL version 3.
 *
 */

/** Legendre polynomials shifted to [0:1] interval.
 *  @version $Id: LegendreShifted.java,v 1.161 2017/09/08 10:13:21 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
public class LegendreShifted extends BasisPolynomials {

/** Calculates n-th Legendre polynomial shifted to [0,1].
 *  Uses: nP_n=2*(2n-1)xP_{n-1}-(2n-1)P_{n-1}-(n-1)P_{n-2}.
 *        nP_n=(2n-1)P_{n-1}(2x-1)-(n-1)P_{n-2}.
 * @param n P index.
 * @param res Output vector.
 * @param Pnm1 Input P_{n-1}
 * @param Pnm2 Input P_{n-2}
 */
public void setNext(final int nd,final double [] res,
		    final double [] Pnm1,final double [] Pnm2,
		    final boolean flag_type2)
{
    if(nd<0) throw new IllegalArgumentException();
    else if(nd==0)
    {
	res[0]=!flag_type2?1:2;
    }
    else
    {
	final int n=!flag_type2?nd:nd+1;
	final double nn12=2*n-1;
	res[nd]=2*(nn12/n)*Pnm1[nd-1];
	if(nd>1) res[nd-1]=(nn12/n)*(2*Pnm1[nd-2]-Pnm1[nd-1]);
	else res[nd-1]=(nn12/n)*(-Pnm1[nd-1]);
	for(int k=nd-1;--k>=0;)
	{
	    if(k>0)
		res[k]=(nn12*(2*Pnm1[k-1]-Pnm1[k])-(n-1)*Pnm2[k])/n;
	    else
		res[k]=-(nn12*Pnm1[k]+(n-1)*Pnm2[k])/n;
	}
    }
}



/**
 * Calculation of the value of n-th Legendre polynomial shifted to [0,1].
 * Uses: nP_n=(2n-1)P_{n-1}(2x-1)-(n-1)P_{n-2}.
 * @param n P index.
 * @param x Argument.
 * @return The value of the Legendre shifted polynomial Pn at x.
 */
public double calculate(final int n,final double x)
{
    return Legendre.BASIS.calculate(n,2*x-1);
}


public double [] calculateMomentsFromSample(final int n,final double [] x,final double [] w,final int nelsinsample)
{
    final double [] m=new double [n+1];
    for(int k=0;k<nelsinsample;k++)
    {
	final double xx=2*x[k]-1,ww=w!=null?w[k]:1;
	double b=0,s=1;
	for(int i=0;i<=n;++i)
	{
	    if(i<=0) m[i]+=s*ww;
	    else {s=((1-i)*b+(2*i-1)*(b=s)*(xx))/i;m[i]+=s*ww;}
	}
    }
    return m;
}


/** Calculates sum_{k=0}^{k=n}P_k(x) Pcoefs_k
 *  using Clenshaw recurrence formula.
 * @param n P index.
 * @param Pcoefs Coefficients of P_n.
 * @param x Pont to estimate the sum.
 * @return The sum.
 */
public double calculateSumD0(final int n,final double [] Pcoefs,final double x)
{
    return Legendre.BASIS.calculateSumD0(n,Pcoefs,2*x-1);
}

/** Calculates d/dx(sum_{k=0}^{k=n}P_k(x) Pcoefs_k)
 *  using Clenshaw recurrence formula.
 * @param n P index, elements=n+1.
 * @param Pcoefs Coefficients of P_n.
 * @param x Pont to estimate the sum.
 * @return The derivative.
 */
public double calculateSumD1(final int n,final double [] Pcoefs,final double x)
{
    return 2*Legendre.BASIS.calculateSumD1(n,Pcoefs,2*x-1);
}


/** Calculates d^2/dx^2(sum_{k=0}^{k=n}P_k(x) Pcoefs_k)
 *  using Clenshaw recurrence formula.
 * @param n P index, elements=n+1.
 * @param Pcoefs Coefficients of P_n.
 * @param x Pont to estimate the sum.
 * @return The derivative.
 */
public double calculateSumD2(final int n,final double [] Pcoefs,final double x)
{
    return 4*Legendre.BASIS.calculateSumD2(n,Pcoefs,2*x-1);
}

public double [] getDifferentiated(final int n,final double [] Pcoefs)
{
    final double [] d=Legendre.BASIS.getDifferentiated(n,Pcoefs);
    for(int k=0;k<d.length;k++) d[k]*=2;
    return d;
}

/** Integrate shifted Legendre polynomial.
 * Prudnikov, Brychkov, Marichev, volume 2, 2003,
 * formulae 1.14.1.1 page 46.
 */
public double [] getIntegrated(final int n,final double [] Pcoefs)
{
    final double [] i=Legendre.BASIS.getIntegrated(n,Pcoefs);
    for(int k=0;k<i.length;k++) i[k]*=0.5;
    i[0]-=calculateSumD0(i.length-1,i,0); // normalize integral to f(0)=0
    return i;
}



/** Expand Pl(z)Pm(z)=sum_k=0^{l+m} s_k Pk(z),
 *  formulae 8.915.5, A(9036), page 1040, Gradshtein & Ryzhik 1963.
 *  The formulae of Neumann-Adams, see W. N. Bailey,
 *  Generalized hypergeometric series, p.101, Cambridge, 1962.
 *  @param l l.
 *  @param m m.
 *  @param s coefs.
 */
public void setQlQmExpansionCoefs(final int l,final int m,final double [] s)
{
    Legendre.BASIS.setQlQmExpansionCoefs(l,m,s);
}


/** Multiply p by ax+b.
 * Calculations in Shifted Legendre basis.
 * Use Legendre recurrence, see
 * formulae 8.914.1, UVII99, page 1040, Gradshtein & Ryzhik 1963.
 * @param n The max dimension of p, elements=n+1.
 * @param res The result.
 * @param p Input polynomial.
 */
@Override
public void setaxbP(final int n,final double [] res,final double [] p,final double a,final double b)
{
    Legendre.BASIS.setaxbP(n,res,p,0.5*a,b+0.5*a);
}



/** Divide  p by (x-x0), returns the remainder.
 * Calculations in Shifted Legendre basis.
 * Use Legendre recurrence, see
 * formulae 8.914.1, UVII99, page 1040, Gradshtein & Ryzhik 1963.
 * @param n The max dimension of p, n_elements=n+1.
 * @param res The result.
 * @param p Input polynomial.
 * @param x0 To divide on (x-x0)
 * @return The remainder.
 */
@Override
public double sdiv1(final int n,final double [] res,final double [] p,final double x0)
{
    final double r=Legendre.BASIS.sdiv1(n,res,p,2*x0-1);
    for(int k=n;--k>=0;){res[k]*=2;}
    return r;
}



/** Calculates P_n(a(x+x0)) using 3 term recurrence
 * Calculations in Shifted Legendre basis.
 * The b_k are defined as:
 * P_n(a(x+x0))=sum_{k=0}^{k=n}b_k^{n} P_k(x)
 * The method should be applied consequently to get b_k^{n} for n=0,1,2,3,...
 * Uses Legendre recurrence, see
 * formulae 8.914.1, UVII99, page 1040, Gradshtein & Ryzhik 1963.
 * @param n The max dimension of b.
 * @param b On input: b_k^{n-1}, on output b_k^{n}
 * @param q On input: b_k^{n-2}, on output: b_k^{n-1}.
 */
public void setNewtonBinomialLikeCoefs(final int n,final double [] b,final double [] q,final double a,final double x0)
{
    Legendre.BASIS.setNewtonBinomialLikeCoefs(n,b,q,a,1+2*x0-1/a);
}


/** Calculate <F|xPk> moments for k=0..n-1 from <F|Pk>, k=0..n
 *  @param n Max index.
 *  @param d Output, d[k]=<F|xPk>, k=0..n-1;
 *  @param Pcoefs Input, Pcoefs[k]=<F|Pk>, k=0..n;
 */
@Override
public void setXQnMomentsFromQnMoments(final int n,final double [] d,final double [] Pcoefs)
{
    double pk1=0,pk=Pcoefs[0];
    for(int k=0;k<n;++k)
    {
	final double f=2*(2*k+1);
	d[k]=(k/f)*pk1+0.5*(pk1=pk)+((k+1)/f)*(pk=Pcoefs[k+1]);
    }
}


/** Calculates confederate matrix to solve
 * polynomial equations in Legendre basis directly.
 * Eigenvalues of confederate matrix are the roots.
 * See:
 * Tom Bella, Y.Eidelman, I.Gohberg, V.Olshevsky, E.Tyrtyshnikov,
 * "Fast inversion of Hessenberg-quasiseparable-Vandermonde matrices
 * and resulting recurrence relations and characterizations", March 2007.
 * http://www.math.uconn.edu/~olshevsky/papers/traubqs2_tb36.pdf
 * We use the three term recurrence:
 * P_k=(x*2*(2*k-1)/k-(2*k-1)/k)*P_{k-1}-((k-1)/k)*P_{k-2}.
 * Use the PolynomialRootsConfederateMatrixABasis.ABASISLEGENDRESHIFTED
 * to calculate roots.
 */
public double [] getConfederateMatrix(final double [] coefs)
{
    final int n=coefs.length-1;
    final double [] matr=new double [n*n];
    for(int m=0;m<n;m++)
    {
	final int k=m+1;
	final double k42=4*k-2;
	matr[m*n+m]=0.5;                    // delta/alfa
	if(m>0) matr[(m-1)*n+m]=(k-1)/k42;  // gamma/alfa
	if(m<n-1) matr[(m+1)*n+m]=k/k42;    // 1/alfa
	else
	{
	    final double f=k42*coefs[n]/k;  // p[n]*alfa
	    for(int j=0;j<n;j++) matr[j*n+m]-=coefs[j]/f;
	}
    }
    return matr;
}


private static final LegendreShifted BASIS=new LegendreShifted();


public static double get2LegendreAveraged(final int n,final double [] p1,final double [] p2)
{
    double s=0;
    for(int k=0;k<=n;k++)
    {
	s+=p1[k]*p2[k]/(2*k+1);
    }
    return s;
}

/** Generating expansion coefficients
 *  xP'n+Pn=(n+1)Pn+sum_{k=0}^{k=n-1}(2k+1)P_k
 */
public static void setXPnprimeplusPnCoefs(final int n,final double [] s)
{
    s[n]=(2*n+1)-n;
    for(int k=n;--k>=0;)
    {
	s[k]=(2*k+1);
    }
}
/** Calculates xF'+F.
 * @param n Max index.
 * @param d Output, d[k]=(2*k+1)<xF'+F|Pk>.
 * @param Pcoefs,input Pcoefs[k]=(2*k+1)<F|Pk>.
 */
public static void setXPnprimeplusPn(final int n,final double [] d,final double [] Pcoefs)
{
    double s=0;
    for(int k=n;k>=0;--k)
    {
	final double x;
	d[k]=s*(2*k+1)+(k+1)*(x=Pcoefs[k]);
	s+=x;
    }
}

/** Calculate <F|Pn+xP'n> moments. Same as setXPnprimeplusPnMomentsFromPnMoments
 *  but uses extra (2k+1) normalizing factor.
 *  @param n Max index.
 *  @param d Output, d[k]=(2*k+1)<F|Pk+xP'k>, k=0..n.
 *  @param Pcoefs Input, Pcoefs[k]=(2*k+1)<F|Pk>, k=0..n.
 */
public static void setXPnprimeplusPnMoments(final int n,final double [] d,final double [] Pcoefs)
{
    double s=0;
    for(int k=0;k<=n;++k)
    {
	final double x;
	d[k]=s*(2*k+1)+(k+1)*(x=Pcoefs[k]);
	s+=x;
    }
}

/** Calculate <F|Pn+xP'n> moments.
 *  @param n Max index.
 *  @param d Output, d[k]=<F|Pk+xP'k>, k=0..n.
 *  @param Pcoefs Input, Pcoefs[k]=<F|Pk>, k=0..n.
 */
public static void setXPnprimeplusPnMomentsFromPnMoments(final int n,final double [] d,final double [] Pcoefs)
{
    double s=0;
    for(int k=0;k<=n;++k)
    {
	final double x;
	d[k]=s+(k+1)*(x=Pcoefs[k]);
	s+=x*(2*k+1);
    }
}


/** Calculate <F|(1-x)Pk> moments for k=0..n-1 from <F|Pk>, k=0..n
 *  @param n Max index.
 *  @param d Output, d[k]=<F|(1-x)Pk>, k=0..n-1;
 *  @param Pcoefs Input, Pcoefs[k]=<F|Pk>, k=0..n;
 */
public static void set1mXPnMomentsFromPnMoments(final int n,final double [] d,final double [] Pcoefs)
{
    double pk1=0,pk=Pcoefs[0];
    for(int k=0;k<n;++k)
    {
	final double f=2*(2*k+1);
	d[k]=-(k/f)*pk1+0.5*(pk1=pk)-((k+1)/f)*(pk=Pcoefs[k+1]);
    }
}


/** Calculate xF'(x) in Legendre shifted basis.
 * @param n Max index.
 * @param d Output, d[k]=(2*k+1)<xF'|Pk>.
 * @param Pcoefs,input Pcoefs[k]=(2*k+1)<F|Pk>.
 */
public static void setXDiffF(final int n,final double [] d,final double [] Pcoefs)
{
    double s=0;
    for(int k=n;k>=0;--k)
    {
	final double x;
	d[k]=s*(2*k+1)+k*(x=Pcoefs[k]);
	s+=x;
    }
}

/** The same as setQlQmExpansionCoefs,
 *  but uses 3 term recursion instead of hypergeometric function expansion.
 *  It is rather slow at large l,m. Used for unit test.
 */
public static void setQlQmExpansionCoefs_test(final int l,final int m,final double [] s)
{
    for(int k=l+m;k>=0;k--) s[k]=0;
    final int ll,mm;
    if(l<=m) {ll=l;mm=m;}
    else {ll=m;mm=l;}
    sumQlQmExpansionCoefs(ll,mm,s,1);
}
private static void sumQlQmExpansionCoefs(final int l,final int m,final double [] s,final double c)
{
    if(l>m) throw new IllegalArgumentException("l="+l+">m="+m);
    if(l<=0) {s[m]+=c; return;}
    final double ol=1.0/l;
    if(l>=2) sumQlQmExpansionCoefs(l-2,m,s,c*(-1+ol));
    final double ol2m1=c*(2-ol)/(2*m+1);
    sumQlQmExpansionCoefs(l-1,m-1,s,ol2m1*m);
    sumQlQmExpansionCoefs(l-1,m+1,s,ol2m1*(m+1));
}



/** Multiply p by x.
 * Calculations in Shifted Legendre basis.
 * Use Legendre recurrence, see
 * formulae 8.914.1, UVII99, page 1040, Gradshtein & Ryzhik 1963.
 * @param n The max dimension of p, elements=n+1.
 * @param res The result.
 * @param p Input polynomial.
 */
public static void setxP(final int n,final double [] res,final double [] p)
{
    BASIS.setaxbP(n,res,p,1,0);
}


/** Multiply p by (1-x).
 * Calculations in Shifted Legendre basis.
 * Use Legendre recurrence, see
 * formulae 8.914.1, UVII99, page 1040, Gradshtein & Ryzhik 1963.
 * @param n The max dimension of p, elements=n+1.
 * @param res The result.
 * @param p Input polynomial.
 */
public static void set1mxP(final int n,final double [] res,final double [] p)
{
    BASIS.setaxbP(n,res,p,-1,1);
}



/** A different way to calculate roots. Use Jacobi matrix.
 */
static double [] getRoots(final int n,final double x0,final boolean flag_radau)
{
    final double [] a=new double [n+1],b=new double [n];
    a[0]=1;
    for(int k=0;k<n;k++)
    {
	b[k]=0.5;
	a[k+1]=0.25*(k+1)/Math.sqrt((k+1)*(k+1)-0.25);
    }
    if(flag_radau)
    {
	b[n-1]+=a[n]*(BASIS.calculate(n,x0)/BASIS.calculate(n-1,x0))*Math.sqrt((2*n+1)/((double)(2*n-1)));
    }
    return EVSolver.getEV_3DCommon(n,b,0,a,1,false);
}
public static double [] getRoots(final int n)
{
    return getRoots(n,Double.NaN,false);
}
public static double [] getRadauRoots(final int n)
{
    return getRoots(n,1,true);
}
/** Uses Hildebrand, F. B. "Introduction to Numerical Analysis."
 *  New York: McGraw-Hill, pp. 323-325, 1956.
 */
public static double [] getWeights(final double [] xx)
{
    final int n=xx.length;
    final double [] w=new double [n];
    final double f=4.0/((n+1)*(n+1));
    for(int k=n;--k>=0;)
    {
	final double x=xx[k];
	w[k]=f*x*(1-x)/Math.pow(BASIS.calculate(n+1,x),2);
    }
    return w;
}
public static double [] getNodesPolynomial(final int n)
{
    final double [] nodes=new double [n+1];
    nodes[n]=1;
    return nodes;
}

/** Radau exact weights from
 * Walter Gautschi, "Gauss-Radau formulae for Jacobi and Laguerre weight functions".
 */
public static double [] getRadauWeights(final double [] xr)
{
    final int n=xr.length;
    final double [] w=new double [n];
    final double f=0.5/(n*n);
    for(int k=n;--k>=0;)
    {
	final double x=xr[k];
	// 1.8 . sign changed in (1-t) because our node is at the right
	w[k]=f*(1+(2*x-1))/Math.pow(BASIS.calculate(n-1,x),2);
    }
    return w;
}
public static double [] getRadauNodesPolynomial(final int n)
{
    final double [] nodes=new double [n+1];
    nodes[0]=1;
    setXPnprimeplusPnMoments(n-1,nodes,nodes); // get delta function by differentiating a step function
    set1mxP(n-1,nodes,nodes);
    return nodes;
}
/** Christoffel function. */
public static double [] getChristoffelKernel(final int n)
{
    final double [] c=new double [2*n+1],tmp=new double [2*n+1];
    for(int k=0;k<=n;k++)
    {
	BASIS.setQlQmExpansionCoefs(k,k,tmp);
	TMatr.sumP1(c,tmp,2*k+1,2*k+1);
    }
    return c;
}


/** A class for unit test. */
static class TestBasisLegendreShifted extends TestBasisPolynomials {

TestBasisLegendreShifted()
{
    this(new java.util.Random(-7591141),1e-5);
}

TestBasisLegendreShifted(final java.util.Random r,final double eps)
{
    super("LegendreShifted",BASIS,r,eps);
}

double [] getMonomialsMoments(final int n)
{
    final double [] pmom=new double [n];
    for(int k=0;k<pmom.length;k++) pmom[k]=1.0/(k+1); // monomials moments
    return pmom;
}
OrthogonalPolynomialsABasis getOrtPol()
{
    return new OrthogonalPolynomialsLegendreShiftedBasis();
}
double [][] getExact()
{
    return new double [][]{
	{1},
	{-1,2},
	{1,-6,6},
	{-1,12,-30,20},
	{1,-20,90,-140,70},
	{-1,30,-210,560,-630,252},
	{1,-42,420,-1680,3150,-2772,924},
	{-1,56,-756,4200,-11550,16632,-12012,3432},
	{1,-72,1260,-9240,34650,-72072,84084,-51480,12870},
	{-1,90,-1980,18480,-90090,252252,-420420,411840,-218790,48620}
    };
}


private void printAllQuadratures(final int n_max)
{
    final PolynomialRootsFinder rf=new PolynomialRootsFinder(){
	    public int findRoots(final int n,final double [] p,final double [] re_root,final double [] im_root){
		return new PolynomialRootJenkinsTraubORIG(n).findRoots(n,BASIS.convertToPolynomialFromBasis_Test(n,p),re_root,im_root);
	    }};
    for(int n=1;n<=n_max;n++)
    {
	printTestRoots(rf,n);
	printQuadratures(rf,n);
	printRadauQuadratures2(rf,n,r);
	printDeltaFunctionRoots(rf,n);
    }
}

void printQuadratures(final PolynomialRootsFinder rf,final int n)
{
    System.err.println("=========LegendreShifted: Gauss quadratures for n="+n);
    final double eps=1e-4;

    final double [] pol=new double [n+1];pol[n]=1;
    final double [] root_re=new double [n],root_im=new double [n];
    final int nroots=rf.findRoots(n,pol,root_re,root_im);
    if(nroots!=n) throw new RuntimeException("nroots="+nroots+" n="+n);
    SortArrays.sortArrayWithLoadDesc(nroots,root_re,root_im);
    final double [] xr=getRoots(n);
    SortArrays.sortArrayWithLoadDesc(xr.length,xr,null);
    UnitTests.checkDoubleArrayRel(xr,"xr",root_re,"root_re",eps);
    final double [] wh=getWeights(xr),ww=new double [xr.length];
    final double [] kk=getChristoffelKernel(n),wk=new double [ww.length];

    System.err.println("X\tW\tS");
    double s=0,S=0;
    for(int k=0;k<n;k++)
    {
	UnitTests.checkDoubleRel(root_im[k],"root_im[k]",0,"0",eps);
	final double w=2/(n*BASIS.calculate(n-1,root_re[k])*BASIS.calculateSumD1(n,pol,root_re[k]));
	s+=w;
	S-=w*Math.log(w);
	ww[k]=w;
	wk[k]=1.0/BASIS.calculateSumD0(kk.length-1,kk,root_re[k]);
	System.err.println(String.valueOf(root_re[k])+"\t"+w+"\t"+s);
    }
    UnitTests.checkDoubleRel(s,"s",1,"1",eps);
    UnitTests.checkDoubleArrayRel1(ww,"ww",wh,"wh",eps);
    UnitTests.checkDoubleArrayRel1(wk,"wk",wh,"wh",eps);
    System.err.println("s="+s+" S="+S);
    System.err.println();
}
void printTestRoots(final PolynomialRootsFinder rf,final int n)
{
    final double [] pol=new double [n+1];pol[n]=1;
    final double [] root_re=new double [n],root_im=new double [n];
    final int nroots=rf.findRoots(n,pol,root_re,root_im);
    if(nroots!=n) throw new RuntimeException("nroots="+nroots+" n="+n);
    SortArrays.sortArrayWithLoadDesc(nroots,root_re,root_im);
    final double [] xr=getRoots(n);
    SortArrays.sortArrayWithLoadDesc(xr.length,xr,null);
    UnitTests.checkDoubleArrayRel(xr,"xr",root_re,"root_re",eps);
    System.err.println("=========LegendreShifted: TestRoots for n="+n);
    System.err.println("Xm\tFm\tXf\tFf");
    for(int k=0;k<n;k++)
    {
	System.err.println(String.valueOf(xr[k])+"\t"+BASIS.calculate(n,xr[k])+"\t"+
			   root_re[k]+"\t"+BASIS.calculate(n,root_re[k]));
    }
    System.err.println();
}

void printRadauQuadratures2(final PolynomialRootsFinder rf,final int neq,final java.util.Random r)
{
    System.err.println("=========Legendre-Radau: Gauss quadratures for neq="+neq);
    final double eps=1e-4;

    final double [] mom=new double [2*neq+1];
    mom[0]=1;

    final double [] matr=new double [neq*neq],
	pol=new double [neq+1],e=new double [2*neq+1];
    for(int k=0;k<neq;k++)
    {
	BASIS.setQlQmExpansionCoefs(neq,k,e);
	set1mxP(neq+k,e,e);
	//System.err.println("e="+java.util.Arrays.toString(e));
	pol[k]=-TMatr.mulVV(neq+k+2,mom,e);
	for(int l=0;l<=k;l++)
	{
	    BASIS.setQlQmExpansionCoefs(l,k,e);
	    set1mxP(l+k,e,e);
	    matr[k+l*neq]=matr[l+k*neq]=TMatr.mulVV(l+k+2,e,mom);
	}
    }
    final int [] ipvt=new int [neq];
    final int pinfo=com.polytechnik.utils.Linsystems.dgefa_m1(matr,neq,neq,ipvt);
    if(pinfo!=0) throw new RuntimeException("pinfo="+pinfo);
    com.polytechnik.utils.Linsystems.dgesl_m1(matr,neq,neq,ipvt,pol,1);
    pol[neq]=1;
    //System.err.println("pol="+java.util.Arrays.toString(pol));
    final double [] xs=new double [neq+1],root_im=new double [neq+1];
    final int nroots=rf.findRoots(neq,pol,xs,root_im);
    if(nroots!=neq) throw new RuntimeException("nroots="+nroots+" neq="+neq);
    xs[neq]=1;root_im[neq]=0; // radau node
    final double [] rpol=new double [neq+2],tmp=new double [neq+1];
    set1mxP(neq,rpol,pol);
    if(true)
    {
	final double [] nrr=getRadauNodesPolynomial(rpol.length-1);
	for(int k=0;k<rpol.length;k++) nrr[k]*=rpol[rpol.length-1]/nrr[rpol.length-1];
	UnitTests.checkDoubleArrayRel(nrr,"nrr",rpol,"rpol",eps);
    }
    final double [] ws=new double [neq+1];
    for(int k=0;k<=neq;k++)
    {
	UnitTests.checkDoubleRel(root_im[k],"root_im[k]",0,"0",eps);
	final double x=xs[k];
	if(Double.compare(x,1.0)==0)
	{ // radau node
	    ws[k]=TMatr.mulVV(neq+1,pol,mom)/BASIS.calculateSumD0(neq,pol,x);
	}
	else
	{
	    BASIS.sdiv1(neq+1,tmp,rpol,x);
	    ws[k]=TMatr.mulVV(neq+1,tmp,mom)/BASIS.calculateSumD0(neq,tmp,x);
	}
    }
    SortArrays.sortArrayWithLoadDesc(neq+1,xs,ws);

    final double [] xr=getRadauRoots(neq+1);
    SortArrays.sortArrayWithLoadDesc(xr.length,xr,null);
    UnitTests.checkDoubleArrayRel(xr,"xr",xs,"xs",eps);

    System.err.println("X\tW\tS");
    double s=0,S=0;
    for(int k=0;k<=neq;k++)
    {
	final double x=xs[k],w=ws[k];
	s+=w;
	S-=w*Math.log(w);
	System.err.println(String.valueOf(x)+"\t"+w+"\t"+s);
    }
    UnitTests.checkDoubleRel(s,"s",1,"1",eps);
    System.err.println("s="+s+" S="+S);
    System.err.println();

    // test radau exact integral, polynomial of 2*neq power
    for(int t=0;t<1000;t++)
    {
    final double [] pt=new double [2*neq +1];
    for(int k=pt.length;--k>=0;) pt[k]=r.nextDouble();
    pt[0]+=0.6+r.nextDouble()+r.nextDouble();
    double i=0;
    for(int k=0;k<xs.length;k++)
    {
	i+=ws[k]*BASIS.calculateSumD0(pt.length-1,pt,xs[k]);
    }
    final double ia=TMatr.mulVV(pt.length,pt,mom);
    UnitTests.checkDoubleRel(i,"i",ia,"ia",eps);
    }

    // check Radau exact weights.
    if(neq<=12)
    checkRadauExactWeights(xs,ws,eps);
}


/** Check Radau exact weights from
 * Walter Gautschi, "Gauss-Radau formulae for Jacobi and Laguerre weight functions".
 */
void checkRadauExactWeights(final double [] xs,final double [] ws,final double eps)
{
    final int neq=xs.length-1;
    UnitTests.checkDoubleRel(ws[0],"ws[0]",1.0/((neq+1)*(neq+1)),"1.0/((neq+1)*(neq+1))",eps);  // 1.8
    UnitTests.checkDoubleArrayRel(ws,"ws",getRadauWeights(xs),"getRadauWeights(xs)",eps);

    System.err.println("checkRadauExactWeights: Radau exact weights OK, neq="+neq);
}

void printDeltaFunctionRoots(final PolynomialRootsFinder rf,final int neq)
{
    System.err.println("=========LegendreShifted: Quadratures on (1-x)delta(1-x) roots for neq="+neq);
    final double eps=1e-5;
    final double [] o=new double [neq+1];
    o[0]=1;
    setXPnprimeplusPnMoments(neq,o,o); // get delta function by differentiating a step function

    //System.err.println("o="+java.util.Arrays.toString(o));

    final double [] mom=new double [2*neq+1];
    mom[0]=1;

    final double [] wnodes=new double [neq+2],tmp=new double [neq+2];
    BASIS.setMult2Pol(wnodes,o,o.length-1,new double []{-0.5,0.5},1,tmp);
    final double [] xs=new double [neq+1],root_im=new double [neq+1];
    final int nroots=rf.findRoots(neq+1,wnodes,xs,root_im);
    if(nroots!=neq+1) throw new RuntimeException("nroots="+nroots+" neq+1="+(neq+1));
    final double [] ws=new double [neq+1];
    for(int k=0;k<=neq;k++)
    {
	UnitTests.checkDoubleRel(root_im[k],"root_im[k]",0,"0",eps);
	final double x=xs[k];
	BASIS.sdiv1(neq+1,tmp,wnodes,x);
	ws[k]=TMatr.mulVV(neq+1,tmp,mom)/BASIS.calculateSumD0(neq,tmp,x);
    }

    SortArrays.sortArrayWithLoadDesc(neq+1,xs,ws);
    System.err.println("X\tW\tS");
    double s=0;
    for(int k=0;k<=neq;k++)
    {
	final double x=xs[k],w=ws[k];
	s+=w;
	System.err.println(String.valueOf(x)+"\t"+w+"\t"+s);
    }
    UnitTests.checkDoubleRel(s,"s",1,"1",eps);
    System.err.println();
    if(neq<=12)
    checkRadauExactWeights(xs,ws,eps);
}



void test_setQlQmExpansionCoefs(final int n_tests)
{
    super.test_setQlQmExpansionCoefs(n_tests);

    for(int t=0;t<n_tests;t++)
    {

    for(int l=0;l<=nmax;l++)
    for(int m=0;m<=nmax;m++)
    {
	final double [] res=new double [l+m+1],resa=new double [l+m+1];
	for(int k=0;k<res.length;k++){res[k]=r.nextDouble();resa[k]=r.nextDouble();}
	BASIS.setQlQmExpansionCoefs(l,m,res);
	setQlQmExpansionCoefs_test(l,m,resa);
	UnitTests.checkDoubleArrayRel(res,"res",resa,"resa",eps);
    }
    }
    System.err.println("LegendreShifted:test_setQlQmExpansionCoefs(): EXTRA test OK with nmax="+nmax+" eps="+eps);
}


void test_getLegendreFromPByIntegration(final int n_tests)
{
    for(int t=0;t<n_tests;t++)
    {
	final int n=1+r.nextInt(6);
	final double [] p=new double [n+1];
	for(int k=p.length;--k>=0;) {p[k]=r.nextDouble();}
	//System.err.println("p="+java.util.Arrays.toString(p));
	final double [] res=new double [n+1];
	double [] ra=new double [n+1],rb=new double [n+1];
	for(int k=0;k<=n;k++)
	{
	    BASIS.setNext(k,rb,ra,rb);
	    final double [] x=rb;rb=ra;ra=x;
	    final double [] i=Polynomials.getPolynomialIntegrated(Polynomials.getMult2Pol(x,k+1,p,n+1));
	    //System.err.println("p x="+java.util.Arrays.toString(Polynomials.getMult2Pol(x,k+1,p,n+1)));

	    //System.err.println("i="+java.util.Arrays.toString(i));
	    //System.err.println("x["+k+"]="+java.util.Arrays.toString(x));
	    res[k]=(Polynomials.getD0NoScale(i,1)-Polynomials.getD0NoScale(i,0))*(2*k+1); // [0,1] interval, (2*k+1) normalizing coefficient.
	}
	//System.err.println("res="+java.util.Arrays.toString(res));
	UnitTests.checkDoubleArrayRel1(p,"p",BASIS.convertToPolynomialFromBasis_Test(n,res),"convertToPolynomialFromBasis_Test(res)",eps);
	UnitTests.checkDoubleArrayRel1(BASIS.convertToBasisFromPolynomial_Test(n,p),"convertToBasisFromPolynomial_Test",res,"res",eps);
    }
    System.err.println("LegendreShifted:test_getLegendreFromPByIntegration() norm (2n+1) test OK with n_tests="+n_tests+" eps="+eps);
}


void test_getLegendreAveraged(final int n_tests)
{
    for(int t=0;t<n_tests;t++)
    {
	final double [] p1=new double [1+r.nextInt(6)],p2=new double [p1.length];
	for(int k=0;k<p1.length;k++){p1[k]=r.nextDouble();p2[k]=r.nextDouble();}
	final double [] i=Polynomials.getPolynomialIntegrated(Polynomials.getMult2Pol(BASIS.convertToPolynomialFromBasis_Test(p1.length-1,p1),p1.length,BASIS.convertToPolynomialFromBasis_Test(p2.length-1,p2),p2.length));
	UnitTests.checkDoubleRel(get2LegendreAveraged(p1.length-1,p1,p2),"<p1 x p2>",Polynomials.getD0NoScale(i,1)-Polynomials.getD0NoScale(i,0),"int pols",eps);
    }
    System.err.println("LegendreShifted:test_getLegendreAveraged() test OK with n_tests="+n_tests+" eps="+eps);
}


void test_getXPnplusPn(final int n_tests)
{
    // testing setXPnprimeplusPn
    for(int n=1;n<16;n++)
    {
	final double [] pol=new double [n+1];
	pol[n]=1;
	final double [] xpplusp=new double [n+1];setXPnprimeplusPnCoefs(n,xpplusp);
	//System.err.println("P"+n+"="+java.util.Arrays.toString(POLS[n]));
	//System.err.println("xpplusp"+n+"="+java.util.Arrays.toString(xpplusp));

	final double [] xx=BASIS.convertToPolynomialFromBasis_Test(n,pol);
	for(int k=0;k<n;k++) {xx[k+1]+=(k+1)*xx[k+1];} // P+xP'
	UnitTests.checkDoubleArrayRel(BASIS.convertToBasisFromPolynomial_Test(n,xx),"xx",xpplusp,"xpplusp",eps);
	UnitTests.checkDoubleArrayRel(xx,"xx",BASIS.convertToPolynomialFromBasis_Test(n,xpplusp),"xpplusp",eps);
	//System.err.println("xx"+n+"="+java.util.Arrays.toString(xx));

	final double [] d=new double [n+1];
	setXPnprimeplusPn(n,d,pol);
	UnitTests.checkDoubleArrayRel(d,"d",xpplusp,"xpplusp",eps);
    }

    for(int t=0;t<n_tests;t++)
    {
	final double [] pol=new double [1+r.nextInt(7)];
	final int n=pol.length-1;
	for(int k=n;k>=0;--k){pol[k]=r.nextDouble();}
	if(true) // Test setXDiffF.
	{
	    final double [] p=BASIS.convertToPolynomialFromBasis_Test(n,pol),
		dp=Polynomials.getPolynomialDifferentiated(p),xdp=new double [n+1];
	    for(int k=0;k<n;k++){xdp[k+1]=dp[k];} // xF'

	    final double [] d=new double [n+1];
	    setXDiffF(n,d,pol);
	    UnitTests.checkDoubleArrayRel(d,"d",BASIS.convertToBasisFromPolynomial_Test(n,xdp),"x pol'",eps);
	    UnitTests.checkDoubleArrayRel(BASIS.convertToPolynomialFromBasis_Test(n,d),"d",xdp,"xdp",eps);
	}

	if(true) // Test setXPnprimeplusPn agains direct calculations.
	{
	    final double [] d=new double [n+1];
	    setXPnprimeplusPn(n,d,pol);
	    final double [] xx=BASIS.convertToPolynomialFromBasis_Test(n,pol);
	    for(int k=0;k<n;k++) {xx[k+1]+=(k+1)*xx[k+1];} // F+xF'
	    UnitTests.checkDoubleArrayRel(BASIS.convertToBasisFromPolynomial_Test(n,xx),"xx",d,"d",eps);
	    UnitTests.checkDoubleArrayRel(xx,"xx",BASIS.convertToPolynomialFromBasis_Test(n,d),"d",eps);
	}

	if(true) // Test setXPnprimeplusPnMoments.
	{
	    final double [] d=new double [n+1];
	    setXPnprimeplusPnMoments(n,d,pol);
	    if(true)
	    {
		final double [] pp=new double [n+1],dd=new double [n+1];
		for(int k=0;k<=n;k++) pp[k]=pol[k]/(2*k+1);
		setXPnprimeplusPnMomentsFromPnMoments(n,dd,pp);
		final double [] ddt=new double [n+1];
		for(int k=0;k<=n;k++) ddt[k]=dd[k]*(2*k+1);
		UnitTests.checkDoubleArrayRel(ddt,"ddt",d,"d",eps);

		for(int k=0;k<=n;k++)
		{
		    final double [] tmp=new double [k+1],rtmp=new double [k+1];
		    tmp[k]=1;
		    if(k>0) setxP(k-1,rtmp,BASIS.getDifferentiated(k,tmp));
		    TMatr.sumP1(rtmp,tmp,k+1,1);
		    ddt[k]=TMatr.mulVV(k+1,rtmp,pp);
		}
		UnitTests.checkDoubleArrayRel(ddt,"ddt",dd,"dd",eps);
		final double [] ddt2=pp.clone();
		setXPnprimeplusPnMomentsFromPnMoments(n,ddt2,ddt2);
		UnitTests.checkDoubleArrayRel(ddt2,"ddt2",dd,"dd",eps);
	    }


	    if(true) // Test against setXDiffF.
	    {
		final double f1=BASIS.calculateSumD0(n,pol,1);
		final double [] dd=new double [n+1],xdd=new double [n+1];
		for(int k=0;k<=n;k++){dd[k]=(2*k+1)*f1-d[k];}
		setXDiffF(n,xdd,pol);
		UnitTests.checkDoubleArrayRel(dd,"dd",xdd,"xdd",eps);
	    }

	    if(true) // Test against direct calculation.
	    {
		final double [] xd=new double [n+1];
		double [] ra=new double [n+1],rb=new double [n+1];
		for(int k=0;k<=n;k++)
		{
		    BASIS.setNext(k,rb,ra,rb);
		    final double [] l=rb;rb=ra;ra=l;
		    final double [] xx=new double [k+1];
		    xx[0]=l[0];for(int j=0;j<k;j++) {xx[j+1]=(j+1+1)*l[j+1];} // P+xP'
		    final double [] xxl=BASIS.convertToBasisFromPolynomial_Test(k,xx);
		    xd[k]=(2*k+1)*get2LegendreAveraged(k,xxl,pol);
		}
		UnitTests.checkDoubleArrayRel(d,"d",xd,"xd",eps);
	    }

	    if(true) // Test against setXPnprimeplusPnCoefs.
	    {
		final double [] sd=new double [n+1],temp=new double [n+1];
		for(int k=0;k<=n;k++)
		{
		    setXPnprimeplusPnCoefs(k,temp);
		    sd[k]=(2*k+1)*get2LegendreAveraged(k,temp,pol); // (2*k+1) to convert moments to coefficients
		}
		UnitTests.checkDoubleArrayRel(d,"d",sd,"sd",eps);
	    }
	}
	if(true)
	{
	    // save moments in pp.
	    final double [] pp=new double [n+1];
	    for(int k=0;k<=n;k++) pp[k]=pol[k]/(2*k+1);
	    final double [] Xpp=new double [n],_1mXpp=new double [n];
	    for(int k=0;k<n;k++)
	    {
		final double [] tmp=new double [k+1],rtmp=new double [k+2];
		tmp[k]=1;
		setxP(k,rtmp,tmp);
		Xpp[k]=TMatr.mulVV(k+2,rtmp,pp);
		set1mxP(k,rtmp,tmp);
		_1mXpp[k]=TMatr.mulVV(k+2,rtmp,pp);
	    }
	    {
	    final double [] sXpp=new double [n];
	    BASIS.setXQnMomentsFromQnMoments(n,sXpp,pp);
	    UnitTests.checkDoubleArrayRel(sXpp,"sXpp",Xpp,"Xpp",eps);
	    final double [] sXpp_same=pp.clone();
	    BASIS.setXQnMomentsFromQnMoments(n,sXpp_same,sXpp_same);
	    UnitTests.checkDoubleArrayRelFirstindex(n,sXpp_same,"sXpp_same",Xpp,"Xpp",eps);
	    }
	    {
	    final double [] s1mXpp=new double [n];
	    set1mXPnMomentsFromPnMoments(n,s1mXpp,pp);
	    UnitTests.checkDoubleArrayRel(s1mXpp,"s1mXpp",_1mXpp,"_1mXpp",eps);
	    final double [] s1mXpp_same=pp.clone();
	    set1mXPnMomentsFromPnMoments(n,s1mXpp_same,s1mXpp_same);
	    UnitTests.checkDoubleArrayRelFirstindex(n,s1mXpp_same,"s1mXpp_same",_1mXpp,"_1mXpp",eps);
	    }
	}
    }

    // Testing an expansion using setXPnprimeplusPnCoefs.
    for(int t=0;t<n_tests;t++)
    {
	final double [] pol=new double [1+r.nextInt(7)];
	final int n=pol.length-1;
	for(int k=n;k>=0;--k){pol[k]=r.nextDouble();}
	final double [] p=BASIS.convertToPolynomialFromBasis_Test(n,pol),
	    dp=Polynomials.getPolynomialDifferentiated(p),xdp=new double [n+1];
	for(int k=0;k<n;k++){xdp[k+1]=dp[k];} // xF'
	//System.err.println("p="+java.util.Arrays.toString(p));
	//System.err.println("xdp="+java.util.Arrays.toString(xdp));
	//System.err.println("pol="+java.util.Arrays.toString(pol));

	final double [] pk1=new double [n+1];
	final double [] pk2=new double [n+1],temp=new double [n+1];
	double [] ra=new double [n+1],rb=new double [n+1];
	for(int k=0;k<=n;k++)
	{
	    BASIS.setNext(k,rb,ra,rb);
	    final double [] pleg=rb;rb=ra;ra=pleg;
	    final double [] i=Polynomials.getPolynomialIntegrated(Polynomials.getMult2Pol(xdp,xdp.length,pleg,n+1));
	    //System.err.println("i"+k+"="+java.util.Arrays.toString(i));
	    pk1[k]=(Polynomials.getD0NoScale(i,1)-Polynomials.getD0NoScale(i,0)); // [0,1] interval.

	    setXPnprimeplusPnCoefs(k,temp);
	    //System.err.println("temp"+k+"="+java.util.Arrays.toString(temp));
	    double s=0;
	    for(int l=0;l<=k;l++){s+=temp[l]*pol[l]/(2*l+1);} // 1/(2*l+1) is Pl^2 integral value.
	    //System.err.println("legendreSum("+n+",pol,1)="+legendreSum(n,pol,1)+" s="+s);
	    pk2[k]=BASIS.calculateSumD0(pol.length-1,pol,1)-s;
	}

	UnitTests.checkDoubleArrayRel1(pk1,"pk1",pk2,"pk2",eps);

	final double [] x_dpol=new double [n+1],x_dpol_frompk=new double [n+1];
	setXDiffF(n,x_dpol,pol);
	for(int k=0;k<=n;k++)
	{
	    x_dpol_frompk[k]=pk1[k]*(2*k+1); // normalizing
	}

	UnitTests.checkDoubleArrayRel1(x_dpol,"x_dpol",x_dpol_frompk,"x_dpol_frompk",eps);
	UnitTests.checkDoubleArrayRel1(xdp,"xdp",BASIS.convertToPolynomialFromBasis_Test(n,x_dpol),"x_dpol",eps);

	//System.err.println("pk1="+java.util.Arrays.toString(pk1)+"\npk2="+java.util.Arrays.toString(pk2)+"\ndpol="+java.util.Arrays.toString(x_dpol)+"\n"+"DD2(xdp)="+java.util.Arrays.toString(convertToLegendreFromPolynomial_Test(n,xdp)));



    }
    System.err.println("LegendreShifted:test_getXPnplusPn() test OK with eps="+eps);
}

void printExtra(final int nm)
{
    printAllQuadratures(15);
}

void test_Extra(final int n_tests)
{
    test_getLegendreFromPByIntegration(n_tests);
    test_getLegendreAveraged(n_tests);
    test_getXPnplusPn(n_tests);
}

} // END OF class TestBasisLegendreShifted


public static void main(String [] args)
{
    new TestBasisLegendreShifted().runTests("$Id: LegendreShifted.java,v 1.161 2017/09/08 10:13:21 mal Exp $");
}


} // END of class LegendreShifted

