package com.polytechnik.utils;

/*
 * Taken from JDK 7 http://download.java.net/jdk7/
 * archive http://www.java.net/download/jdk7/jdk-7-ea-src-b34-jrl-28_aug_2008.jar
 * file hotspot/agent/src/share/classes/sun/jvm/hotspot/debugger/LongHashMap.java
 *
 * $Id: LongHashMap.java,v 1.53 2012/01/13 16:51:04 mal Exp $
 */

import java.util.*;

/**
 * This is a copy of java.util.HashMap which uses longs as keys
 * instead of Objects. It turns out that using this in the PageCache
 * implementation speeds up heap traversals by a factor of three.
 *
 * @author  Josh Bloch
 * @author Arthur van Hoff
 */

public class LongHashMap<V>
{
    private static class Entry<V> {
        final long   key;
        final int    hash;
        V value;
        Entry<V>  next;

        Entry(final int hash,final  long key,final V value,final  Entry<V> next) {
            this.hash  = hash;
            this.key   = key;
            this.value = value;
            this.next  = next;
        }
    }

    /**
     * The hash table data.
     */
    transient Entry<V> table[];

    /**
     * The total number of mappings in the hash table.
     */
    transient int size;

    /**
     * The table is rehashed when its size exceeds this threshold.  (The
     * value of this field is (int)(capacity * loadFactor).)
     *
     * @serial
     */
    int threshold;

    /**
     * The load factor for the hash table.
     *
     * @serial
     */
    final float loadFactor;

    /**
     * The number of times this HashMap has been structurally modified
     * Structural modifications are those that change the number of mappings in
     * the HashMap or otherwise modify its internal structure (e.g.,
     * rehash).  This field is used to make iterators on Collection-views of
     * the HashMap fail-fast.  (See ConcurrentModificationException).
     */
    transient int modCount = 0;

    /**
     * Constructs a new, empty map with the specified initial
     * capacity and the specified load factor.
     *
     * @param      initialCapacity   the initial capacity of the HashMap.
     * @param      loadFactor        the load factor of the HashMap
     * @throws     IllegalArgumentException  if the initial capacity is less
     *               than zero, or if the load factor is nonpositive.
     */
    public LongHashMap(int initialCapacity,final float loadFactor) {
        if (initialCapacity < 0)
            throw new IllegalArgumentException("Illegal Initial Capacity: "+
                                               initialCapacity);
        if (loadFactor <= 0 || Float.isNaN(loadFactor))
            throw new IllegalArgumentException("Illegal Load factor: "+
                                               loadFactor);
        if (initialCapacity==0)
            initialCapacity = 1;
        this.loadFactor = loadFactor;
	@SuppressWarnings("unchecked")
	final Entry<V> [] t=new Entry[initialCapacity];
        table = t;
        threshold = (int)(initialCapacity * loadFactor);
    }

    /**
     * Constructs a new, empty map with the specified initial capacity
     * and default load factor, which is <tt>0.75</tt>.
     *
     * @param   initialCapacity   the initial capacity of the HashMap.
     * @throws    IllegalArgumentException if the initial capacity is less
     *              than zero.
     */
    public LongHashMap(final int initialCapacity) {
        this(initialCapacity, 0.75f);
    }

    /**
     * Constructs a new, empty map with a default capacity and load
     * factor, which is <tt>0.75</tt>.
     */
    public LongHashMap() {
        this(11, 0.75f);
    }

    /**
     * Returns the number of key-value mappings in this map.
     *
     * @return the number of key-value mappings in this map.
     */
    public final int size() {
        return size;
    }

    /**
     * Returns <tt>true</tt> if this map contains no key-value mappings.
     *
     * @return <tt>true</tt> if this map contains no key-value mappings.
     */
    public final boolean isEmpty() {
        return size == 0;
    }

    /**
     * Returns the value to which this map maps the specified key.  Returns
     * <tt>null</tt> if the map contains no mapping for this key.  A return
     * value of <tt>null</tt> does not <i>necessarily</i> indicate that the
     * map contains no mapping for the key; it's also possible that the map
     * explicitly maps the key to <tt>null</tt>.  The <tt>containsKey</tt>
     * operation may be used to distinguish these two cases.
     *
     * @return the value to which this map maps the specified key.
     * @param key key whose associated value is to be returned.
     */
    public final V get(final long key) {
        final Entry<V> e = getEntry(key);
        return (e == null ? null : e.value);
    }

    /**
     * Returns <tt>true</tt> if this map contains a mapping for the specified
     * key.
     *
     * @return <tt>true</tt> if this map contains a mapping for the specified
     * key.
     * @param key key whose presence in this Map is to be tested.
     */
    public final boolean containsKey(final long key) {
        return getEntry(key) != null;
    }

    /**
     * Returns the entry associated with the specified key in the
     * HashMap.  Returns null if the HashMap contains no mapping
     * for this key.
     */
    final Entry<V> getEntry(final long key) {
	final int mc=modCount;
        final Entry<V> tab[] = table;
        final int hash = (int) key;
        final int index = (hash & 0x7FFFFFFF) % tab.length;

        for (Entry<V> e = tab[index]; e != null; e = e.next)
	{
            if (e.hash == hash && e.key ==key)
	    {
		if(mc!=modCount)
		{
		    throw new ConcurrentModificationException(
			"modCount mismatch:"+
			" old modCount="+mc+
			" new modCount="+modCount);
		}
                return e;
	    }
	}

	if(mc!=modCount)
	{
	    throw new ConcurrentModificationException(
		"modCount mismatch:"+
		" old modCount="+mc+
		" new modCount="+modCount);
	}
        return null;
    }

    /**
     * Associates the specified value with the specified key in this map.
     * If the map previously contained a mapping for this key, the old
     * value is replaced.
     *
     * @param key key with which the specified value is to be associated.
     * @param value value to be associated with the specified key.
     * @return previous value associated with specified key, or <tt>null</tt>
     *         if there was no mapping for key.  A <tt>null</tt> return can
     *         also indicate that the HashMap previously associated
     *         <tt>null</tt> with the specified key.
     */
    public V put(final long key,final V value) {
        Entry<V> tab[] = table;
        final int hash = (int) key;
        int index = (hash & 0x7FFFFFFF) % tab.length;

        // Look for entry in hash table
        for (Entry<V> e = tab[index] ; e != null ; e = e.next) {
            if (e.hash == hash && e.key == key) {
                final V oldValue = e.value;
                e.value = value;
                return oldValue;
            }
        }

        // It's not there; grow the hash table if necessary...
        modCount++;
        if (size >= threshold) {
            rehash();
            tab = table;
            index = (hash & 0x7FFFFFFF) % tab.length;
        }

        // ...and add the entry
        size++;
        tab[index] = new Entry<V>(hash, key, value, tab[index]);
        return null;
    }

    /**
     * Removes the mapping for this key from this map if present.
     *
     * @param key key whose mapping is to be removed from the map.
     * @return previous value associated with specified key, or <tt>null</tt>
     *         if there was no mapping for key.  A <tt>null</tt> return can
     *         also indicate that the map previously associated <tt>null</tt>
     *         with the specified key.
     */
    public V remove(final long key) {
        final Entry<V> e = removeEntryForKey(key);
        return (e == null ? null : e.value);
    }

    /**
     * Removes and returns the entry associated with the specified key
     * in the HashMap.  Returns null if the HashMap contains no mapping
     * for this key.
     */
    final Entry<V> removeEntryForKey(final long key) {
        final Entry<V> tab[] = table;
        final int hash = (int) key;
        final int index = (hash & 0x7FFFFFFF) % tab.length;

        for (Entry<V> e = tab[index], prev = null; e != null;
             prev = e, e = e.next) {
            if (e.hash == hash && e.key == key) {
                modCount++;
                if (prev != null)
                    prev.next = e.next;
                else
                    tab[index] = e.next;

                size--;
                return e;
            }
        }

        return null;
    }



    /**
     * Removes all mappings from this map.
     */
    public void clear() {
        final Entry<V> tab[] = table;
        modCount++;
        for (int index = tab.length; --index >= 0; )
            tab[index] = null;
        size = 0;
    }

    /**
     * Rehashes the contents of this map into a new <tt>HashMap</tt> instance
     * with a larger capacity. This method is called automatically when the
     * number of keys in this map exceeds its capacity and load factor.
     */
    void rehash() {
        final Entry<V> oldTable[] = table;
        final int oldCapacity = oldTable.length;
        final int newCapacity = oldCapacity * 2 + 1;
	@SuppressWarnings("unchecked")
        final Entry<V> newTable[] = new Entry[newCapacity];

        modCount++;
        threshold = (int)(newCapacity * loadFactor);
        table = newTable;

        for (int i = oldCapacity ; i-- > 0 ;) {
            for (Entry<V> old = oldTable[i] ; old != null ; ) {
                final Entry<V> e = old;
                old = old.next;

                final int index = (e.hash & 0x7FFFFFFF) % newCapacity;
                e.next = newTable[index];
                newTable[index] = e;
            }
        }
    }

    public int copyValues(final V[] valuesarr) {
	if(valuesarr.length<size)
	{
	    throw new IllegalArgumentException("The size of values array="+valuesarr.length+" is lower that the number of elements="+size);
	}
	final int mc=modCount;
	java.util.Arrays.fill(valuesarr,null);

	final Entry<V>[] tab = table;
	int n=0;
        for (int i = tab.length ; i-- > 0 ;) {
            for (Entry<V> e = tab[i] ; e != null ; ) {
		valuesarr[n++]=e.value;
		e=e.next;
            }
        }

	if(n!=size || mc!=modCount)
	{
	    throw new ConcurrentModificationException(
		"The number of entries="+n+
		" is not equal to size="+size+
		" old modCount="+mc+
		" new modCount="+modCount);
	}
	return n;
    }

    public int copyKeys(final long [] keylist)
    {
	if(keylist.length<size)
	{
	    throw new IllegalArgumentException("The size of keys array="+keylist.length+" is lower that the number of elements="+size);
	}
	final int mc=modCount;
	java.util.Arrays.fill(keylist,0L);

	final Entry<V>[] tab = table;
	int n=0;
        for (int i = tab.length ; i-- > 0 ;) {
            for (Entry<V> e = tab[i] ; e != null ; ) {
		keylist[n++]=e.key;
		e=e.next;
            }
        }

	if(n!=size || mc!=modCount)
	{
	    throw new ConcurrentModificationException(
		"The number of entries="+n+
		" is not equal to size="+size+
		" old modCount="+mc+
		" new modCount="+modCount);
	}
	return n;
    }



    public int copyKeysAndValues(final long [] keylist,final V[] valuesarr)
    {
	if(keylist.length<size || valuesarr.length<size)
	{
	    throw new IllegalArgumentException("The size of keys array="+keylist.length+
					       " or values array="+valuesarr.length+
					       " is lower that the number of elements="+size);
	}

	final int mc=modCount;
	java.util.Arrays.fill(keylist,0L);
	java.util.Arrays.fill(valuesarr,null);

	final Entry<V>[] tab = table;
	int n=0;
        for (int i = tab.length ; i-- > 0 ;) {
            for (Entry<V> e = tab[i] ; e != null ; ) {
		keylist[n]=e.key;
		valuesarr[n++]=e.value;
		e=e.next;
            }
        }

	if(n!=size || mc!=modCount)
	{
	    throw new ConcurrentModificationException(
		"The number of entries="+n+
		" is not equal to size="+size+
		" old modCount="+mc+
		" new modCount="+modCount);
	}
	return n;
    }


    /**
     * Executes procedure for each key/value entry in the map.
     * @param procedure - a LongObjectProcedure value.
     * @return  false if the loop over the entries terminated
     * because the procedure returned false for some entry.
     */
    public boolean forEachEntry(TLongObjectProcedure<V> procedure)
    {
	final int mc=modCount;
	final Entry<V>[] tab = table;
	int n=0;
        for (int i = tab.length ; i-- > 0 ;) {
            for (Entry<V> e = tab[i] ; e != null ; n++) {
		if(!procedure.execute(e.key,e.value))
		{
		    if(mc!=modCount)
		    {
			throw new ConcurrentModificationException(
			    "modCount mismatch:"+
			    " old modCount="+mc+
			    " new modCount="+modCount);
		    }
		    return false;
		}
		e=e.next;
            }
        }
	if(n!=size || mc!=modCount)
	{
	    throw new ConcurrentModificationException(
		"The number of entries="+n+
		" is not equal to size="+size+
		" old modCount="+mc+
		" new modCount="+modCount);
	}
	return true;
    }

    public static interface TLongObjectProcedure<V> {
	/** Executes this procedure.
	 * A false return value indicates that the
	 * application executing this procedure should
	 * not invoke this procedure again.
	 */
	boolean execute(final long a,final V b);
    }


    public boolean forEachEntry(TObjectProcedure<V> procedure)
    {
	final int mc=modCount;
	final Entry<V>[] tab = table;
	int n=0;
        for (int i = tab.length ; i-- > 0 ;) {
            for (Entry<V> e = tab[i] ; e != null ; n++) {
		if(!procedure.execute(e.value))
		{
		    if(mc!=modCount)
		    {
			throw new ConcurrentModificationException(
			    "modCount mismatch:"+
			    " old modCount="+mc+
			    " new modCount="+modCount);
		    }
		    return false;
		}
		e=e.next;
            }
        }
	if(n!=size || mc!=modCount)
	{
	    throw new ConcurrentModificationException(
		"The number of entries="+n+
		" is not equal to size="+size+
		" old modCount="+mc+
		" new modCount="+modCount);
	}
	return true;
    }



//========================== THE TEST ==================================
// to run the test use the command
// java -Xloggc:/tmp/glog -Xmx2500000K com.polytechnik.utils.LongHashMap


    private static final char [] SYMBOLS64=
        "0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ_.".toCharArray();
    private static final boolean DEBUG_TEST=false;

    private static byte [] randombytes(int n)
    {
	final byte [] b=new byte [n];
	for(int i=n;i-- >0;)
	{
	    b[i]=new Integer((int)(10000*Math.random())).byteValue();
	}
	return b;
    }
    private static String randomstring(int n)
    {
	final byte [] b=randombytes(n);
        final StringBuilder buf=new StringBuilder(1+n);
        for(int i=0;i<n;i++)
	{
	    buf.append(SYMBOLS64[(((int)b[i]) & 0x7f)>>>1]);
	}
        return buf.toString();

    }

    private static Long randomLongk(int i)
    {
	long l=(long)(2.3+1e7*Math.sin(i*3.8+2.0*Math.random()));
	final byte [] b=randombytes(8);
	for(int il=0;il<b.length;il++)
	{
	    l=l^(((long)b[il])<<8*il);
	}
	return new Long(l);
    }

    private static String randomStringk(int i)
    {
	return "VV"+String.valueOf(2000.8*Math.cos(i*1.8+12.0*Math.random()))+":"+
	    randomstring(30);
    }


    private static int comparisontest_hashmap()
    {
	final com.polytechnik.utils.LongHashMap<String> m=
	    new com.polytechnik.utils.LongHashMap<String>(3);
	final java.util.HashMap<Long,String> hm=new java.util.HashMap<Long,String>(3);

	int n_failed=0;

	for(int tnum=0;tnum<350;tnum++)
	{
	     if(tnum%170 == 0)
	     {
	        System.err.println("!!!!!!!!!CLEAR TABLE. ORIG SIZE="+hm.size()+" NEW SIZE=0");
	        m.clear();
		hm.clear();
	     }

	    for(int i=0;i<20000;i++)
	    {
		final Long l=randomLongk(i);
		final String s=randomStringk(i);
		m.put(l.longValue(),s);
		hm.put(l,s);
		if(DEBUG_TEST)
		{
		    System.err.println("L="+l+" S="+s);
		}
	    }
	    n_failed+=test_the_same("TEST="+tnum+" INIT",m,hm);

	    int n_removed=0;
	    // remove some elements
	    for(final java.util.Iterator<Long> e=new java.util.ArrayList<Long>(hm.keySet()).iterator();e.hasNext();)
	    {
		final Long l=e.next();
		// randomly remove elements
		if(Math.random()<0.5)
		{
		    n_removed++;
		    hm.remove(l);
		    m.remove(l.longValue());
		}
	    }
	    n_failed+=test_the_same("TEST="+tnum+" AFTER REMOVE of "+n_removed+" elements. new SIZE="+hm.size(),m,hm);
	    // add some elements in random
	    int n_added=0;
	    for(int i=hm.size()+300;i-- >0;)
	    {
		// randomly remove elements
		if(Math.random()<0.9)
		{
		    n_added++;
		    final Long l=randomLongk(i);
		    final String s=randomStringk(i);
		    m.put(l.longValue(),s);
		    hm.put(l,s);
		    if(DEBUG_TEST)
		    {
			System.err.println("L="+l+" S="+s);
		    }
		}
	    }
	    n_failed+=test_the_same("TEST="+tnum+" AFTER ADD of "+n_added+" new SIZE="+hm.size(),m,hm);

	    int n_replaced=0;
	    // replace some elements in random
	    for(final java.util.Iterator<Long> e=new java.util.ArrayList<Long>(hm.keySet()).iterator();e.hasNext();)
	    {
		final Long l=e.next();
		// randomly remove elements
		if(Math.random()<0.6)
		{
		    n_replaced++;
		    final String s=randomStringk(l.intValue());
		    hm.put(new Long(l.longValue()),s);
		    m.put(l.longValue(),s);
		}
	    }
	    n_failed+=test_the_same("TEST="+tnum+" AFTER REPLACE of "+n_replaced+" elements. new SIZE="+hm.size(),m,hm);
	}

	return n_failed;
    }

    private static int test_the_same(final String comment,
				     final com.polytechnik.utils.LongHashMap<String> m,
				     final java.util.HashMap<Long,String> hm)
    {
	System.err.println(comment+": TEST two tables: size1="+m.size()+" size2="+hm.size());
	if(m.size()!=hm.size())
	{
	    throw new RuntimeException("Size mismatch:"+m.size()+" and "+hm.size());
	}
	final long [] keys=new long [m.size()+200];
	final int ksize=m.size();
	m.copyKeys(keys);
	final String [] mvalues=new String [keys.length];
	m.copyValues(mvalues);
	final long [] keys2=new long [keys.length];
	final String [] mvalues2=new String [keys.length];
	m.copyKeysAndValues(keys2,mvalues2);
	int m1=0;
	for(int i=0;i<ksize;i++)
	{
	    if(!m.get(keys[i]).equals(hm.get(new Long(keys[i]))))
	    {
		throw new RuntimeException("Data mismatch kor key="+keys[i]+
					   " D1="+m.get(keys[i])+
					   " D2="+hm.get(new Long(keys[i])));
	    }
	    if(keys[i]!=keys2[i]||
	       mvalues[i]!=mvalues2[i]||
	       !m.get(keys[i]).equals(mvalues[i]))
	    {
		throw new RuntimeException("Data mismatch kor key="+keys[i]+
					   " key2="+keys2[i]+
					   " values="+mvalues[i]+
					   " values2="+mvalues2[i]+
					   " tablevalue="+m.get(keys[i]));
	    }
	    m1++;
	}
	int m2=0;
	for(final java.util.Iterator<Long> e=hm.keySet().iterator();e.hasNext();)
	{
	    final Long l=e.next();
	    if(!hm.get(l).equals(m.get(l.longValue())))
	    {
		throw new RuntimeException("Data mismatch kor key="+l+
					   " D1="+hm.get(l)+
					   " D2="+m.get(l.longValue()));
	    }
	    m2++;
	}
	System.err.println(comment+": TEST OK, "+ksize+" elements matched ("+m1+" and "+m2+" tries).");

	int nnull=0;
	for(int i=0;i<100000;i++)
	{
	   final Long l=randomLongk(i);
	   if(hm.get(l)==null&&
	      m.get(l.longValue())!=null)
	   {
	       throw new RuntimeException("Data mismatch kor key="+l+
					  " D1="+hm.get(l)+
					  " D2="+m.get(l.longValue()));
	   }
	   nnull++;
	}
	System.err.println(comment+": element not fount "+nnull+" tests OK");


	final int [] vr=new int []{0};
	m.forEachEntry(new LongHashMap.TLongObjectProcedure<String>(){
	   public boolean execute(long k,String v)
	   {
	     if(!hm.get(new Long(k)).equals(v))
	     {
	       throw new RuntimeException("MISMATCH FOR"+k+" -> "+v);
	     }
	     vr[0]++;
	     return true;
	   }
	});
	System.err.println(comment+": forEachEntry check "+vr[0]+" tests OK size="+m.size());


	return 0;
    }


    public static void main(String [] args)
    {
	comparisontest_hashmap();
    }

//========================== END OF THE TEST ==================================


}
