package com.polytechnik.utils;
/*
 * (C) Vladislav Malyshkin 2011
 * This file is under GPL version 3.
 *
 */

/** Monomials polynomials basis,
 * using  T_k = x*T_{k-1}.
 *  @version $Id: Monomials.java,v 1.8 2017/09/08 10:13:21 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
public class Monomials extends BasisPolynomials {


/** Calculates n-th Monomials polynomial
 * using  T_k = x*T_{k-1}
 * @param n T index.
 * @param res Output vector.
 * @param Tnm1 Input T_{n-1}
 * @param Tnm2 Input T_{n-2}
 */
public void setNext(final int nd,final double [] res,
		    final double [] Tnm1,final double [] Tnm2,
		    final boolean flag_type2)
{
    if(flag_type2) throw new IllegalArgumentException("flag_type2=true is not supported by Monomials");
    if(nd<0) throw new IllegalArgumentException();
    if(nd==0)
    {
	res[0]=1;
    }
    else
    {
	for(int k=nd;k>=0;--k)
	{
	    if(k>0)
		res[k]=Tnm1[k-1];
	    else
		res[k]=0;
	}
    }
}


/**
 * Calculation of the value of the Monomials polynomial.
 * @param n T index.
 * @param x Argument.
 * @return The value of the Monomials polynomial Tn at x.
 */
public double calculate(final int n,final double x)
{
    if(n<0) throw new RuntimeException("n="+n);
    double s=1;
    if(n==0) return s;
    for(int i=1;i<=n;++i)
    {
	s=s*x;
    }
    return s;
}

public double [] calculateMomentsFromSample(final int n,final double [] x,final double [] w,final int nelsinsample)
{
    final double [] m=new double [n+1];
    for(int k=0;k<nelsinsample;k++)
    {
	final double xx=x[k],ww=w!=null?w[k]:1;
	double s=1;
	for(int i=0;i<=n;++i)
	{
	    m[i]+=s*ww;
	    s=s*xx;
	}
    }
    return m;
}

/** Calculates sum_{k=0}^{k=n}T_k(x) Tcoefs_k.
 *  Two versions: original & the one copied from http://www.alglib.net/.
 * @param n T index.
 * @param Tcoefs Coefficients of T_n.
 * @param x Point to estimate the sum.
 * @return The sum.
 */
public double calculateSumD0(final int n,final double [] Tcoefs,final double x)
{
    if(n<0) throw new RuntimeException("n="+n);
    return Polynomials.getD0NoScale(n+1,Tcoefs,x);
}

/** Calculates d/dx(sum_{k=0}^{k=n}T_k(x) Tcoefs_k).
 * @param n T index, elements=n+1
 * @param Tcoefs Coefficients of T_n.
 * @param x Pont to estimate the sum.
 * @return The derivative.
 */
public double calculateSumD1(final int n,final double [] Tcoefs,final double x)
{
    if(n<0) throw new RuntimeException("n="+n);
    return Polynomials.getD1NoScale(n+1,Tcoefs,x);
}


/** Calculates d^2/dx^2(sum_{k=0}^{k=n}T_k(x) Tcoefs_k).
 * @param n T index, elements=n+1
 * @param Tcoefs Coefficients of T_n.
 * @param x Pont to estimate the sum.
 * @return The second derivative.
 */
public double calculateSumD2(final int n,final double [] Tcoefs,final double x)
{
    if(n<0) throw new RuntimeException("n="+n);
    return Polynomials.getD2NoScale(n+1,Tcoefs,x);
}




/** Differentiate a function in Monomials basis,
 * F(x) - a polynomial of n-th order.
 * @param n Max index.
 * @param Tcoefs Input: F(x)=T_k(x)*Tcoefs[k].
 * @return  dF/dx=T'_k(x)*Tcoefs[k].
 */
public double [] getDifferentiated(final int n,final double [] Tcoefs)
{
    final double [] pol=new double [n];
    Polynomials.setPolynomialDifferentiated(pol,Tcoefs,n+1);
    return pol;
}


/** Integrate a function in Monomials basis,
 * @param n Max index.
 * @param Tcoefs Input: F(x)=T_k(x)*Tcoefs[k].
 * @return  \int_0^x F(x)dx=\int_0^x T_k(x)dx*Tcoefs[k].
 */
public double [] getIntegrated(final int n,final double [] Tcoefs)
{
    if(n<0) throw new RuntimeException("n="+n);
    final double [] pol=new double [n+2];
    Polynomials.setPolynomialIntegrated(pol,Tcoefs,n+1);
    return pol;
}



/** Expand Tl(z)Tm(z)=sum_k=0^{l+m} s_k Tk(z),
 *  @param l l.
 *  @param m m.
 *  @param s coefs.
 */
public void setQlQmExpansionCoefs(final int l,final int m,final double [] s)
{
    java.util.Arrays.fill(s,0,l+m+1,0);
    s[l+m]=1;
}


/** Multiply p by ax+b.
 * Calculations in Monomials.
 * @param n The max dimension of p, elements=n+1.
 * @param res The result.
 * @param p Input polynomial.
 */
@Override
public void setaxbP(final int n,final double [] res,final double [] p,final double a,final double b)
{
    double r=p[n];
    res[n]=res[n+1]=0;
    for(int k=n;k>=0;k--)
    {
	res[k+1]+=a*r;
	res[k]+=b*r;
	if(k>0)
	{
	    final double t=p[k-1];
	    res[k-1]=0;
	    r=t;
	}
    }
}


/** Divide  p by (x-x0), returns the remainder.
 * Calculations in Monomials basis.
 * @param n The max dimension of p, n_elements=n+1.
 * @param res The result.
 * @param p Input polynomial.
 * @param x0 To divide on (x-x0)
 * @return The remainder.
 */
@Override
public double sdiv1(final int n,final double [] res,final double [] p,final double x0)
{
    return Polynomials.sdiv1(res,p,n+1,x0);
}


/** Calculates T_n(a(x+x0)) using 3 term recurrence.
 * Calculations in Monomials basis.
 * The b_k^{n} are defined as:
 * T_n(a(x+x0))=sum_{k=0}^{k=n}b_k^{n} T_k(x)
 * The method should be applied consequently to get b_k^{n} for n=0,1,2,3,...
 * @param n The max dimension of b.
 * @param b On input: b_k^{n-1}, on output b_k^{n}.
 * @param q On input: b_k^{n-2}, on output: b_k^{n-1}.
 */
public void setNewtonBinomialLikeCoefs(final int n,final double [] b,final double [] q,final double a,final double x0)
{
    if(n<=0)
    {
	b[n]=1;
	return;
    }
    System.arraycopy(b,0,q,0,n);
    setaxbP(n-1,b,q,a,a*x0);
}



/** Calculate <F|xTk> moments for k=0..n-1 from <F|Tk>, k=0..n
 *  @param n Max index.
 *  @param d Output, d[k]=<F|xTk>, k=0..n-1;
 *  @param Pcoefs Input, Pcoefs[k]=<F|Tk>, k=0..n;
 */
@Override
public void setXQnMomentsFromQnMoments(final int n,final double [] d,final double [] moments)
{
    for(int k=n;--k>=0;)
    {
	d[k]=moments[k+1];
    }
}


public double [] getConfederateMatrix(final double [] coefs)
{
    return Polynomials.getConfederateMatrix(coefs);
}


public static void main(String [] args)
{
    final double eps=1e-6;
    new TestBasisPolynomials("Monomials",new Monomials(),new java.util.Random(-79099),eps){
	double [] getMonomialsMoments(final int n){return null;}
	OrthogonalPolynomialsABasis getOrtPol(){return null;}
	double [][] getExact()
	{
	    return new double [][]{
		{1},
		{0,1},
		{0,0,1},
		{0,0,0,1},
		{0,0,0,0,1},
		{0,0,0,0,0,1},
		{0,0,0,0,0,0,1},
		{0,0,0,0,0,0,0,1},
		{0,0,0,0,0,0,0,0,1}
	    };
	}
    }.runTests("$Id: Monomials.java,v 1.8 2017/09/08 10:13:21 mal Exp $");
}

} // END of class Monomials
