package com.polytechnik.utils;
/*
 * (C) Vladislav Malyshkin 2010
 * This file is under GPL version 3.
 *
 */

/** A crude and simple Newton style root finder algorithm.
 *  Used for simple(non-degenerated) real roots finding only.
 *  Can work in an orthogonal polynomial basis.
 *  @version $Id: PolynomialSimpleRootsFinder.java,v 1.9 2012/04/11 13:17:15 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */

public abstract class PolynomialSimpleRootsFinder implements PolynomialRootsFinder  {
    final int nmax,maxiterperstep,miniter;
    final double x0,eps1;
    final double [] coefs;
    public PolynomialSimpleRootsFinder(final int nmax,
				       final int maxiterperstep,
				       final int miniter,
				       final double x0,
				       final double eps1)
    {
	this.nmax=nmax;
	this.maxiterperstep=maxiterperstep;
	this.miniter=miniter;
	this.x0=x0;
	this.eps1=eps1;
	this.coefs=new double [nmax+1];

    }
    public PolynomialSimpleRootsFinder(final int nmax)
    {
	this(nmax,100,17,1.1,1e-3);
    }
    public int findRoots(final int n,
			 final double [] polynomial_coefs,
			 final double [] re_root,
			 final double [] im_root)
    {
	System.arraycopy(polynomial_coefs,0,coefs,0,n+1);

	int r=0;
	double xinit=x0;
	for(int nn=n;nn>=1;nn--)
	{
	    double x=xinit;
	  rf:{
	     double fx=f(nn,x);
	     int s;
	     if(fx<0) s=-1;
	     else if(fx>0) s=1;
	     else break rf;

	     boolean ftest=true;
	     double xL=Double.NaN,xR=Double.NaN;
	     int sL=0,sR=0;
	     boolean bounded=false;
	     for(int k=0,ke=maxiterperstep+miniter+2,kx=0;k<ke;k++)
	     {
		 double dfx=df(nn,x);
		 if(!(Math.abs(dfx)>0)){dfx=-fx;}
		 final double dx1=fx/dfx;
		 double x1=x-dx1;
		 if(nn<=1){x=x1; break rf;}
		 double fx1=f(nn,x1);
		 int s1;
		 if(fx1>0) s1=1;
		 else if(fx1<0) s1=-1;
		 else {x=x1; /*System.err.println("Break FX1 fx1="+fx1+" x="+x1);*/ break rf;}
		 if(!bounded)
		 {
		     final double xx;
		     final int ss;
		     if(s1*s<0)
		     {
			 xx=x1;ss=s1;
		     }
		     else
		     {
			final double x1a=x1-dx1;
			final double fx1a=f(nn,x1a);
			final int s1a;
			if(fx1a>0) s1a=1;
			else if(fx1a<0) s1a=-1;
			else {x=x1a; break rf;}
			if(s1a*s<0){ss=s1a;xx=x1a;}
			else{ss=s1;xx=x1;}
		     }
		     if(ss*s<0)
		     {
			 if(x<xx){xL=x;xR=xx;sL=s;sR=ss;}
			 else{xL=xx;xR=x;sL=ss;sR=s;}
			 bounded=true;
		     }
		     //System.err.println("Not bounded nn="+nn+" k="+k+" ftest="+ftest+" kx="+kx+" miniter="+miniter);
		 }
		 else
		 {
		     final double xm=0.5*(xL+xR),fm=f(nn,xm);
		     final int sm;
		     if(fm>0) sm=1;
		     else if(fm<0) sm=-1;
		     else {x=xm; break rf;}
		     // adjust x1 if bissection is better
		     //System.err.println("Bounded nn="+nn+" fx1("+x1+")="+fx1+" fm("+xm+")="+fm);
		     //System.err.println("Bounded nn="+nn+" k="+k+" ftest="+ftest+" kx="+kx+" miniter="+miniter);
		     if(x1<xL || x1>xR || Math.abs(fx1)>Math.abs(fm))
		     {
			 x1=xm;
			 fx1=fm;
			 s1=sm;
			 //System.err.println("Bissection is better fx1("+x1+")="+fx1+" fm("+xm+")="+fm);
		     }
		     if(s1==sL){xL=x1;}else{xR=x1;}
		 }
		 final double xx=x;
		 x=x1;
		 fx=fx1;
		 s=s1;

		 if(Math.abs(xx-x)<=eps1)
		 {
		     ftest=false;
		 }
		 else kx=0;
		 if(!ftest)
		 {
		     //System.err.println(" EE: kx="+kx+" miniter="+miniter);
		     if(++kx>=miniter){break rf;}
		 }
	     }
	     //System.err.println("EXIT R");
	     // root not found, skip iterations
	     return r;
	  } // end rf
	  //System.err.println("nn="+nn+" x="+x);
	  re_root[r]=x;
	  im_root[r]=0;
	  doSdiv1AndScale(nn,coefs,coefs,x);
	  //xinit=x;
	  r++;
	}
	return r;
    }

    public abstract double f(final int n,final double x);
    public abstract double df(final int n,final double x);
    public abstract void doSdiv1AndScale(final int n,final double [] res,final double [] p,final double x0);

}
