package com.polytechnik.utils;
/*
 * (C) Vladislav Malyshkin 2011
 * This file is under GPL version 3.
 *
 */

/** Simple polynomials utils.
 *  @version $Id: Polynomials.java,v 1.47 2013/12/13 13:41:03 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
public class Polynomials {
/* Divide  p by d. Results: q- quotent, r-remainder.
 * Note: may be numerically unstable.
 */
public static void sdiv(final double [] q,final double [] r,final double [] p,final int pnels,final double [] d,final int dnels)
{
    final int nq=pnels-dnels,nd=dnels-1;
    final double a0=d[nd];
    for(int k=pnels;--k>=0;)
    {
	double s=0;
	if(k>=nd)
	{
	    for(int i=(k<=nq?0:k-nq),kmi=k-i;i<nd;++i,--kmi)
	    {
		s+=d[i]*q[kmi];
	    }
	    q[k-nd]=(p[k]-s)/a0;
	}
	else
	{
	    for(int i=0,ie=k<=nq?k:nq,kmi=k-i;i<=ie;++i,--kmi)
	    {
		s+=d[kmi]*q[i];
	    }
	    r[k]=p[k]-s;
	}
    }
}
/** Stable (more or less) sdiv. */
public static void sdiv_stable(final double [] q,final double [] r,final double [] p,final int pnels,final double [] d,final int dnels)
{
    final int nq=pnels-dnels,nd=dnels-1;
    final double [] matr=new double [pnels*pnels],rpart=new double [pnels];
    for(int k=pnels,o=pnels*(pnels-1);--k>=0;o-=pnels)
    {
	if(k>=nd)
	{
	    for(int i=0,ie=k<=nq?dnels:pnels-k,oi=o+k;i<ie;++i,++oi)
	    {
		matr[oi]=d[nd-i];
	    }
	}
	else
	{
	    for(int i=0,ie=k<=nq?k:nq,oi=o+nd;i<=ie;++i,++oi)
	    {
		matr[oi]=d[k-i];
	    }
	    matr[o+k]=1;
	}
	rpart[k]=p[k];
    }
    final int [] ipvt=new int [pnels];
    final int pinfo=com.polytechnik.utils.Linsystems.dgefa_m1(matr,pnels,pnels,ipvt);
    if(pinfo!=0)
    {
	java.util.Arrays.fill(q,0,nq+1,Double.NaN);
	java.util.Arrays.fill(r,0,nd,Double.NaN);
	return;
    }
    com.polytechnik.utils.Linsystems.dgesl_m1(matr,pnels,pnels,ipvt,rpart,1);
    System.arraycopy(rpart,nd,q,0,nq+1);
    System.arraycopy(rpart,0,r,0,nd);
}


/* Divide  p by (x-x0), returns the remainder.
 */
public static double sdiv1(final double [] res,final double [] p,final int nels,final double x0)
{
    final int n=nels-1;
    if(n==0) return p[0];
    double a=p[n-1],r=res[n-1]=p[n];
    for(int k=n-1;--k>=0;)
    {
	final double b=p[k];
	r=res[k]=a+r*x0;
	a=b;
    }
    return a+r*x0;
}


public static double [] getMult2Pol(final double [] p1,final double [] p2)
{
    return getMult2Pol(p1,p1.length,p2,p2.length);
}
public static double [] getMult2Pol(final double [] p1,final int n1,final double [] p2,final int n2)
{
    final double [] pol=new double [n1+n2-1];
    sumMult2Pol(pol,p1,n1,p2,n2);
    return pol;
}
public static void sumP2(final double [] res,final double [] pin,final int nels)
{
    sumMult2Pol(res,pin,nels,pin,nels);
}
public static void sumMult2Pol(final double [] res,final double [] p1,final int n1,final double [] p2,final int n2)
{
    if(n1<=0 || n2<=0) throw new IllegalArgumentException();

    final int n=n1+n2-1;
    for(int p=0;p<n1;p++)
    {
	double s=0;
	for(int j1=p,j2=0;j1>=0 && j2<n2;--j1,++j2)
	{
	    s+=p1[j1]*p2[j2];
	}
	res[p]+=s;
    }
    for(int p=n1;p<n;p++)
    {
	double s=0;
	for(int j1=n1,j2=p-n1+1;--j1>=0 && j2<n2;++j2)
	{
	    s+=p1[j1]*p2[j2];
	}
	res[p]+=s;
    }
}


/** Converts sum a_k x^k  to sum a_k (x*scale)^k.
 */
public static double [] getPolynomialScaled(final double [] coefs,final double scale)
{
    final double [] p=new double [coefs.length];
    double b=1;
    for(int k=0;k<p.length;k++)
    {
	p[k]=coefs[k]*b;
	b*=scale;
    }
    return p;
}


/** Converts sum a_k x^k  to sum a_k (x+shift)^k.
 */
public static double [] getPolynomialShifted(final double [] coefs,final double shift)
{
    final double [] p=new double [coefs.length];
    shiftPolynomial(p,coefs,shift);
    return p;
}


public static void shiftPolynomial(final double [] p,final double [] coefs,final double shift)
{
    final int n=coefs.length;
    long f=1;
    for(int j=0;j<n;++j,f=(f*(n-j))/j)
    {
	double s=0;
	long fs=f;
	for(int k=n;--k>=j;)
	{
	    s=s*shift+fs*coefs[k];
	    if(k>0) fs=(fs*(k-j))/k;
	}
	p[j]=s;
    }
}

/** Multiply p by x. */
public static void setxP(final double [] res,final double [] p,final int np)
{
    for(int k=np;k>0;k--)
    {
	res[k]=p[k-1];
    }
    res[0]=0;
}
/** Multiply p by (1-x). */
public static void set1mxP(final double [] res,final double [] p,final int np)
{
    for(int k=np;k>=0;k--)
    {
	if(k>=np) res[k]=-p[k-1];
	else if(k<=0) res[k]=p[k];
	else res[k]=p[k]-p[k-1];
    }
}

private static double [] getMult2PolTest(final double [] p1,final int n1,final double [] p2,final int n2)
{
  final double [] pol=new double [n1+n2-1];
  for(int i=0;i<n1;i++)
      for(int j=0;j<n2;j++)
	  pol[i+j]+=p1[i]*p2[j];
  return pol;
}


public static double getD0NoScale(final double [] coefs,final double x)
{
    return getD0NoScale(coefs.length,coefs,x);
}
public static double getD0NoScale(final int nels,final double [] coefs,final double x)
{
    double i=0;
    // i=sum_0^L coefs[k] p^k
    for(int k=nels;--k>=0;)
    {
	i=i*x+coefs[k];
    }
    return i;
}
public static double getD1NoScale(final double [] coefs,final double x)
{
    return getD1NoScale(coefs.length,coefs,x);
}
public static double getD1NoScale(final int nels,final double [] coefs,final double x)
{
    double i=0;
    // i=sum_1^L coefs[k] k p^{k-1}
    for(int k=nels;--k>=1;)
    {
	i=i*x+k*coefs[k];
    }
    return i;
}
public static double getD2NoScale(final double [] coefs,final double x)
{
    return getD2NoScale(coefs.length,coefs,x);
}
public static double getD2NoScale(final int nels,final double [] coefs,final double x)
{
    double i=0;
    // i=sum_2^L coefs[k] k(k-1) p^{k-2} / scale[k]
    for(int k=nels;--k>=2;)
    {
	i=i*x+(k*(k-1))*coefs[k];
    }
    return i;
}

/** \int_x^{infinity} f(x) exp(-x) dx = i(x) exp(-x).
 * Prudnikov, Brychkov, Marichev, volume 1, 2002,
 * formulae 1.3.2.9, page 113.
 */
public static double [] getPolynomialExpFIntegrated(final int n,final double [] f)
{
    final double [] res=new double [n];
    double [] temp=new double [n],temp1=new double [n];
    System.arraycopy(f,0,temp,0,n);
    for(int k=n;--k>=0;)
    {
	TMatr.sumP1(res,temp,k+1,1);
	setPolynomialDifferentiated(temp1,temp,k+1);
	final double [] x=temp;temp=temp1;temp1=x;
    }
    return res;
}


public static double [] getPolynomialIntegrated(final double [] pcoefs)
{
    final double [] pol=new double [1+pcoefs.length];
    setPolynomialIntegrated(pol,pcoefs,pcoefs.length);
    return pol;
}
public static void setPolynomialIntegrated(final double [] res,final double [] pcoefs,final int nels)
{
    for(int k=1;k<=nels;k++)
    {
	res[k]=pcoefs[k-1]/k;
    }
    res[0]=0;
}
public static double [] getPolynomialDifferentiated(final double [] pcoefs)
{
    return getPolynomialDifferentiated(pcoefs,pcoefs.length);
}
public static double [] getPolynomialDifferentiated(final double [] pcoefs,final int nels)
{
    final double [] pol=new double [nels-1];
    setPolynomialDifferentiated(pol,pcoefs,nels);
    return pol;
}
public static void setPolynomialDifferentiated(final double [] pol,final double [] pcoefs,final int nels)
{
    for(int k=1;k<nels;k++)
    {
	pol[k-1]=k*pcoefs[k];
    }
}

public static double getPolynomialRootMinValue(final double [] p,final PolynomialRootsFinder rf)
{
    final double [] dp=getPolynomialDifferentiated(p);
    final double [] root_re=new double [p.length-1],root_im=new double [p.length-1];
    final int nroots=rf.findRoots(dp.length-1,dp,root_re,root_im);
    double smin=Double.POSITIVE_INFINITY;int ind_smin=-1;
    for(int k=0;k<nroots;k++)
    {
	final double sk=getD0NoScale(p,root_re[k]);
	if(sk<=smin)
	{
	    smin=sk;
	    ind_smin=k;
	}
    }
    if(ind_smin>=0)
    {
	return root_re[ind_smin];
    }
    return Double.NaN;
}


/** Converts first kind orthogonal polynomial
 * to second kind orthogonal polynomial.
 * An operation some kind similar to derivative.
 * Calculations in monimials moments basis as an expansion of
 * <(P(x)-P(t))/(x-t)>_t/<1>_t.
 * @param p Orthogonal polynomial.
 * @param pmom Moments of <t^k>_t.
 * @return Second kind orthogonal polynomial.
 */
public static double [] getSecondKindOrthogonalPolynomialFromFirstKind(final double [] p,final double [] pmom)
{
    final int n=p.length-1;
    final double [] res=new double [n];
    double r=p[n];
    for(int k=n;--k>=0;)
    {
	for(int l=0;l<=k;l++)
	{
	    res[l]+=r*pmom[k-l];
	}
	r=p[k];
    }
    final double m0=pmom[0];
    for(int k=n;--k>=0;)
    {
	res[k]/=m0;
    }
    return res;
}


/** Calculates confederate matrix to solve
 * polynomial equations in Legendre basis directly.
 * Eigenvalues of confederate matrix are the roots.
 * See:
 * Tom Bella, Y.Eidelman, I.Gohberg, V.Olshevsky, E.Tyrtyshnikov,
 * "Fast inversion of Hessenberg-quasiseparable-Vandermonde matrices
 * and resulting recurrence relations and characterizations", March 2007.
 * http://www.math.uconn.edu/~olshevsky/papers/traubqs2_tb36.pdf
 * Here confederate matrix is equal to companion matrix.
 * Use the class PolynomialRootsConfederateMatrixABasis.ABASISMONOMIALS
 * to calculate roots.
 */
public static double [] getConfederateMatrix(final double [] coefs)
{
    final int n=coefs.length-1;
    final double [] matr=new double [n*n];
    for(int m=0;m<n;m++)
    {
	if(m<n-1) matr[(m+1)*n+m]=1;
	else
	{
	    final double f=coefs[n]; // p[n]
	    for(int j=0;j<n;j++) matr[j*n+m]-=coefs[j]/f;
	}
    }
    return matr;
}


/** Least squares approximation. Uses polynomial basis.
 */
public static double [] getLinsysSol(final int neq,
				     final double [] mom,final int mom_offset,
				     final double [] rpart,final int rpart_offset)
{
    final double [] r=new double [neq],matr=new double [neq*neq],norma=new double [neq];
    for(int k=0;k<neq;k++)
    {
	r[k]=rpart[k+rpart_offset]/(norma[k]=Math.sqrt(Math.abs(mom[mom_offset+2*k]))); // abs for negatively defined matrix
	for(int l=0;l<=k;l++)
	{
	    matr[k+neq*l]=matr[l+neq*k]=mom[mom_offset+k+l]/norma[l]/norma[k];
	}
    }
    final int [] ipvt=new int [neq];
    final int pinfo=com.polytechnik.utils.Linsystems.dgefa_m1(matr,neq,neq,ipvt);
    if(pinfo!=0)
    {
	java.util.Arrays.fill(r,Double.NaN);
    }
    else
    {
	com.polytechnik.utils.Linsystems.dgesl_m1(matr,neq,neq,ipvt,r,1);
	for(int k=0;k<neq;k++){r[k]/=norma[k];}
    }
    return r;
}

/** Least squares approximation. Uses polynomial basis. Several right parts.
 */
public static double [][] getLinsysSolEqs(final int neq,
					  final double [] mom,final int mom_offset,
					  final double [][] rparts)
{
    final double [] matr=new double [neq*neq],norma=new double [neq];
    final double [][] rr=new double [rparts.length][neq];
    for(int k=0;k<neq;k++)
    {
	norma[k]=Math.sqrt(mom[mom_offset+2*k]);
	for(int j=0;j<rr.length;j++){rr[j][k]=rparts[j][k]/norma[k];}
	matr[k+neq*k]=1;
	for(int l=0;l<k;l++)
	{
	    matr[k+neq*l]=matr[l+neq*k]=mom[mom_offset+k+l]/norma[l]/norma[k];
	}
    }
    final int [] ipvt=new int [neq];
    final int pinfo=com.polytechnik.utils.Linsystems.dgefa_m1(matr,neq,neq,ipvt);
    if(pinfo!=0)
    {
	for(int j=0;j<rr.length;j++){java.util.Arrays.fill(rr[j],Double.NaN);}
	return rr;
    }
    else
    {
	for(int j=0;j<rr.length;j++)
	{
	    final double [] r=rr[j];
	    com.polytechnik.utils.Linsystems.dgesl_m1(matr,neq,neq,ipvt,r,1);
	    for(int k=0;k<neq;k++){r[k]/=norma[k];}
	}
    }
    return rr;
}


/** Like getLinsysSol but uses orthogonal polynomials basis.
 *  The basis is presented as an array of regular polynomilas (monomial basis) in BASIS.
 */
public static double [] getLinsysSolPBasis(final int neq,
					    final double [] mom,final int mom_offset,final int mom_increment,
					    final double [] rpart,final int rpart_offset,final int rpart_increment,
					    final double [][] BASIS)
{
    final double [] r=new double [neq],matr=new double [neq*neq],mul=new double [2*neq-1];
    for(int k=0;k<neq;k++)
    {
	r[k]=rpart[rpart_offset+k*rpart_increment];
	for(int l=0;l<=k;l++)
	{
	    // When we cannot calculate conversion coefficients exactly (I just have no time for this),
	    // use synthetic division. May be numerically unstable.
	    Polynomials.sumMult2Pol(mul,BASIS[k],k+1,BASIS[l],l+1);
	    TMatr.expandOverLowerDiagBasis(mul,mul,l+k+1,BASIS);
	    double s=0;
	    for(int m=k+l;m>=0;m--)
	    {
		s+=mul[m]*mom[mom_offset+m*mom_increment];
		mul[m]=0; // reset for sumMult2Pol
	    }
	    matr[k+neq*l]=matr[l+neq*k]=s;
	    //System.err.println("matr["+k+"]["+l+"]="+s+" mom_offset="+mom_offset+" mom_increment="+mom_increment+" rpart_offset="+rpart_offset+" rpart_increment="+rpart_increment);
	}
    }
    final int [] ipvt=new int [neq];
    final int pinfo=com.polytechnik.utils.Linsystems.dgefa_m1(matr,neq,neq,ipvt);
    if(pinfo!=0)
    {
	java.util.Arrays.fill(r,Double.NaN);
    }
    else
    {
	com.polytechnik.utils.Linsystems.dgesl_m1(matr,neq,neq,ipvt,r,1);
    }
    return r;
}


    public static final long fact(final int n)
    {
	if(n<=1)
	{
	    return 1;
	}
	return n*fact(n-1);
    }
    public static final long fact2(final int n)
    {
	if(n<=1)
	{
	    return 1;
	}
	return n*fact2(n-2);
    }
    public static final long C(final int n,final int k)
    {
	if(k>n) throw new RuntimeException("k="+k+" n="+n);
	long c=1;
	for(int j=0;j<k;)
	{
	    c*=(n-j);
	    ++j;
	    c/=j;
	}
	return c;
    }
private static void test_C()
{
    final int nmax=10;
    for(int n=0;n<=nmax;n++)
    {
	for(int k=0;k<=n;k++)
	{
	    final long c=C(n,k),cf=fact(n)/(fact(n-k)*fact(k));
	    if(c!=cf)
		throw new RuntimeException("c="+c+" cf="+cf);
	}
    }
    System.err.println("Polynomials:test_C: OK with nmax="+nmax);
}


private static void matchmul(final double [] p,final double [] d,final double [] q,final double [] r,final double eps)
{
    final double [] mul=getMult2Pol(d,d.length,q,q.length);
    TMatr.sumP1(mul,r,r.length,1);
    UnitTests.checkDoubleArrayRel1(p,"p",mul,"q x d +r",eps);
}
private static void test_sdiv(final java.util.Random rn,final int n_tests,final double eps)
{
    final long t0=System.currentTimeMillis();
    for(int t=0;t<n_tests;t++)
    {
	final int n=1+rn.nextInt(21);
	final double [] p=new double [n];
	for(int i=p.length;--i>=0;) p[i]=rn.nextDouble();
	final int nd=1+(n<=1?0:rn.nextInt(n-1));
	final double [] d=new double [nd];
	for(int i=d.length;--i>=0;) d[i]=rn.nextDouble();
	if(false)
	{
	    System.err.println("d="+java.util.Arrays.toString(d));
	    System.err.println("p="+java.util.Arrays.toString(p));
	}
	// make the leading coefficient >1, to avoid instability in tests.
	d[d.length-1]=2+rn.nextDouble();

	final double [] q=new double [n-nd+1],r=new double [nd-1];
	for(int i=q.length;--i>=0;) q[i]=rn.nextDouble();
	for(int i=r.length;--i>=0;) r[i]=rn.nextDouble();

	Polynomials.sdiv_stable(q,r,p,p.length,d,d.length);
	matchmul(p,d,q,r,eps);

	for(int i=q.length;--i>=0;) q[i]=rn.nextDouble();
	for(int i=r.length;--i>=0;) r[i]=rn.nextDouble();
	Polynomials.sdiv(q,r,p,p.length,d,d.length);
	if(false)
	{
	    System.err.println("q="+java.util.Arrays.toString(q));
	    System.err.println("r="+java.util.Arrays.toString(r));
	}

	matchmul(p,d,q,r,eps);
    }
    final long t1=System.currentTimeMillis();
    System.err.println("Polynomials:test_sdiv: OK with n_tests="+n_tests+" eps="+eps+
		       " in "+(t1-t0)+" milliseconds.");
}
private static void test_sdiv1(final java.util.Random r,final int n_tests,final double eps)
{
    final long t0=System.currentTimeMillis();
    for(int t=0;t<n_tests;t++)
    {
	final int n=2+r.nextInt(21);
	final double [] p=new double [n];
	for(int i=p.length;--i>=0;) p[i]=r.nextDouble();
	if(true)
	{
	    final double x0=r.nextDouble();
	    final double [] res=new double [n-1];
	    for(int i=res.length;--i>=0;) res[i]=r.nextDouble();
	    final double remainder=Polynomials.sdiv1(res,p,p.length,x0);
	    matchmul(p,res,new double []{-x0,1},new double []{remainder},eps);

	    final double [] mul=getMult2Pol(res,res.length,new double []{-x0,1},2);
	    mul[0]+=remainder;
	    UnitTests.checkDoubleArrayRel1(p,"p",mul,"mul x d +r",eps);
	}

	if(true)
	{
	    // check with the same array
	    final double x0=r.nextDouble();
	    final double [] px=p.clone(),res=new double [p.length-1];
	    final double premainder=Polynomials.sdiv1(res,p,p.length,x0);
	    final double remainder=Polynomials.sdiv1(px,px,px.length,x0);
	    UnitTests.checkDoubleArrayRelFirstindex(res.length,res,"res",px,"px",eps*0);
	    UnitTests.checkDoubleRel(remainder,"remainder",premainder,"premainder",eps*0);
	    final double [] mul=getMult2Pol(px,px.length-1,new double []{-x0,1},2);
	    mul[0]+=remainder;
	    UnitTests.checkDoubleArrayRel1(p,"p",mul,"mul x d +r",eps);
	}
    }
    final long t1=System.currentTimeMillis();
    System.err.println("Polynomials:test_sdiv1: OK with n_tests="+n_tests+" eps="+eps+
		       " in "+(t1-t0)+" milliseconds.");
}
private static void test_getMult2Pol(final java.util.Random r,final int n_tests,final double eps)
{
    final long t0=System.currentTimeMillis();
    for(int t=0;t<n_tests;t++)
    {
	final int n1,n2;
	if(t<=n_tests/100) n1=n2=1+r.nextInt(21);
	else {n1=1+r.nextInt(21);n2=1+r.nextInt(21);}
	final double [] p1=new double [n1],p2=new double [n2];
	for(int i=p1.length;--i>=0;) p1[i]=r.nextDouble();
	for(int i=p2.length;--i>=0;) p2[i]=r.nextDouble();
	final double [] m1=getMult2Pol(p1,p1.length,p2,p2.length),
	    m2=getMult2PolTest(p1,p1.length,p2,p2.length);
	UnitTests.checkDoubleArrayRel(m1,"getMult2Pol",m2,"getMult2PolTest",eps);
    }
    final long t1=System.currentTimeMillis();
    System.err.println("Polynomials:test_getMult2Pol: OK with n_tests="+n_tests+" eps="+eps+
		       " in "+(t1-t0)+" milliseconds.");
}
private static void test_setxP(final java.util.Random r,final int n_tests,final double eps)
{
    final long t0=System.currentTimeMillis();
    for(int t=0;t<n_tests;t++)
    {
	final int n=1+r.nextInt(21);
	final double [] p=new double [n];
	for(int i=p.length;--i>=0;) p[i]=r.nextDouble();
	final double [] px=new double [n+1],p1mx=new double [n+1];
	setxP(px,p,n);
	matchmul(px,p,new double [] {0,1},new double [] {0},eps);
	UnitTests.checkDoubleArrayRel(px,"setxP",getMult2Pol(p,new double []{0,1}),"getMult2Pol",eps);
	set1mxP(p1mx,p,n);
	matchmul(p1mx,p,new double [] {1,-1},new double [] {0},eps);
	UnitTests.checkDoubleArrayRel(p1mx,"set1mxP",getMult2Pol(p,new double []{1,-1}),"getMult2Pol",eps);

	// try with the same array
	final double [] spx=new double [n+1],sp1mx=new double [n+1];
	System.arraycopy(p,0,spx,0,p.length);
	spx[n]=r.nextDouble();
	setxP(spx,spx,n);
	UnitTests.checkDoubleArrayRel(px,"px",spx,"spx",eps*0);
	System.arraycopy(p,0,sp1mx,0,p.length);
	sp1mx[n]=r.nextDouble();
	set1mxP(sp1mx,sp1mx,n);
	UnitTests.checkDoubleArrayRel(p1mx,"p1mx",sp1mx,"sp1mx",eps*0);
    }

    final long t1=System.currentTimeMillis();
    System.err.println("Polynomials:test_setxP: OK with n_tests="+n_tests+" eps="+eps+
		       " in "+(t1-t0)+" milliseconds.");
}

private static void printAllQuadratures(final int n_max,final java.util.Random r)
{
    for(int n=1;n<=n_max;n++)
    {
	printQuadratures(n);
    }
}
private static void printQuadratures(final int neq)
{
    System.err.println("=========Polynomials: Newton-Cotes quadratures for neq="+neq);
    final double eps=1e-4;

    final double [] xs=new double [neq+1],ws=new double [neq+1];
    double [] p=new double []{1};
    for(int k=0;k<=neq;k++)
    {
	xs[k]=(k/(double)neq);
	p=getMult2Pol(p,new double [] {-xs[k],1});
    }
    final double [] tmp=new double [neq+1];
    for(int k=0;k<=neq;k++)
    {
	sdiv1(tmp,p,p.length,xs[k]);
	final double [] i=getPolynomialIntegrated(tmp);
	ws[k]=(getD0NoScale(i,1)-getD0NoScale(i,0))/getD0NoScale(tmp,xs[k]);
    }

    System.err.println("X\tW\tS");
    double s=0;
    for(int k=0;k<=neq;k++)
    {
	final double x=xs[k],w=ws[k];
	s+=w;
	System.err.println(String.valueOf(x)+"\t"+w+"\t"+s);
    }
    UnitTests.checkDoubleRel(s,"s",1,"1",eps);
    System.err.println();

    if(neq==8)
    { // test first Newton Cotes rule with negative weights
      // (Abramowitz and Stegun 1972, p. 886, formulae 25.4.18)
      // see http://mathworld.wolfram.com/Newton-CotesFormulas.html
	final double [] w9=new double []{
	    989,
	    5888,
	    -928,
	    10496,
	    -4540,
	    10496,
	    -928,
	    5888,
	    989};
	final double n=14175;
	for(int k=w9.length;--k>=0;) w9[k]/=2*n;
	UnitTests.checkDoubleArrayRel(w9,"w9",ws,"ws",eps);
	System.err.println("printQuadratures: matched neq="+neq+" analytical match OK");
    }
}


private static void testPolynomialsSDIV()
{
    final java.util.Random r=new java.util.Random(-79091);
    final double eps=1e-10;
    printAllQuadratures(17,r);
    test_C();
    test_getMult2Pol(r,10000000,eps);
    test_setxP(r,10000000,eps);
    test_sdiv1(r,10000000,eps);
    test_sdiv(r,10000000,eps);
}


public static void main(String [] args)
{
    System.err.println("Polynomials::main(): starting ALL tests; ver=$Id: Polynomials.java,v 1.47 2013/12/13 13:41:03 mal Exp $");
    final long t0=System.currentTimeMillis();
    testPolynomialsSDIV();
    final long t1=System.currentTimeMillis();
    System.err.println("Polynomials::main(): ALL tests OK done in "+(t1-t0)+" milliseconds. ver=$Id: Polynomials.java,v 1.47 2013/12/13 13:41:03 mal Exp $");
}



}
