package com.polytechnik.utils;
/*
 * (C) Vladislav Malyshkin 2014
 * This file is under GPL version 3.
 *
 */

/** Probability Correlation of p and i.
 *  @version $Id: ProbabilityCorrelation.java,v 1.25 2018/05/14 21:27:19 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
public class ProbabilityCorrelation {
public final double pL,pH,p_aver,skewnessLike_p,iL,iH,i_aver,skewnessLike_i,
    W,w_pL_iL,w_pL_iH,w_pH_iL,w_pH_iH,corr;

/** Probaility Correlation (from monomials moments).
 */
public ProbabilityCorrelation(final double x0,final double x1,final double x2,
			      final double px0,final double px1,final double px2,
			      final double ix0,final double ix1,final double ix2)
{
    this(x0,x1,x2,px0,px1,px2,ix0,ix1,ix2,Skewness.MB);
}

/** Probaility Correlation (from the moments in arbitrary basis, the Q_0(x) is assumed to be a constant!).
 */
public ProbabilityCorrelation(final double x0,final double x1,final double x2,
			      final double px0,final double px1,final double px2,
			      final double ix0,final double ix1,final double ix2,
			      final OrthogonalPolynomialsABasis M)
{
    this(x0,x1,x2,px0,px1,px2,ix0,ix1,ix2,M,0);
}
/** From 2x2 matrices in arbitraty basis,
 * the Q_0(x) is assumed to be a constant, otherwise averages will be incorrect.
 * Unit selftest is disabled.
 */
public ProbabilityCorrelation(final double [] QQ,final double [] QQp,final double [] QQi,final OrthogonalPolynomialsABasis M)
{
    this(getW(QQ,M),QQ,QQp,QQi,M.EV,0);
}
/** Probaility Correlation (from the moments in arbitrary basis, the Q_0(x) is assumed to be a constant!).
 */
protected ProbabilityCorrelation(final double x0,final double x1,final double x2,
				 final double px0,final double px1,final double px2,
				 final double ix0,final double ix1,final double ix2,
				 final OrthogonalPolynomialsABasis M,final double eps_if_selftest_required)
{
    this(M.getQQMatr(2,new double []{x0,x1,x2}),
	 M.getQQMatr(2,new double []{px0,px1,px2}),
	 M.getQQMatr(2,new double []{ix0,ix1,ix2}),M,eps_if_selftest_required);
}
protected ProbabilityCorrelation(final double [] QQ,final double [] QQp,final double [] QQi,final OrthogonalPolynomialsABasis M,final double eps_if_selftest_required)
{
    this(getW(QQ,M),QQ,QQp,QQi,M.EV,eps_if_selftest_required);
}

static double getW(final double [] QQ,final OrthogonalPolynomialsABasis M)
{
    final double [] ONE=new double []{1};{ONE[0]/=M.fx(ONE,0);}
    return TMatr.getAMatrA(ONE,QQ,2);
}

/** From 2x2 matrices in arbitraty basis,
 *  if eps_if_selftest_required>0, then unit self-test with this eps is run.
 */
public ProbabilityCorrelation(final double W,final double [] QQ,final double [] QQp,final double [] QQi,
			      final EVSolver EV,final double eps_if_selftest_required)
{
    final int d=2;
    if(QQ.length!=d*d || QQp.length!=d*d || QQi.length!= d*d)
	throw new IllegalArgumentException("Matrix dimension is not 2");
    this.W=W;
    final double [] ev_MiM=EV.getGEV_Generalized(d,QQi,d,QQ,d,true),
	bQQp=new double [d*d],tmp=new double [d*d];
    iL=ev_MiM[0];
    iH=ev_MiM[d-1];
    i_aver=QQi[0]/QQ[0];
    // Transform basis to optimize calculations.
    convertMatrFromQToB_Basis(bQQp,d,QQp,ev_MiM,d,d,tmp);
    final double [] bev_MpMt=EV.getEV(d,bQQp,d,true);
    pL=bev_MpMt[0];
    pH=bev_MpMt[d-1];
    p_aver=QQp[0]/QQ[0];
    skewnessLike_p=(2*p_aver-pL-pH)/(pL-pH);
    skewnessLike_i=(2*i_aver-iL-iH)/(iL-iH);
    w_pL_iL=Math.pow(bev_MpMt[d+0*d+0],2);
    w_pL_iH=Math.pow(bev_MpMt[d+0*d+1],2);
    w_pH_iL=Math.pow(bev_MpMt[d+1*d+0],2);
    w_pH_iH=Math.pow(bev_MpMt[d+1*d+1],2);
    corr=((w_pH_iH+w_pL_iL)-(w_pL_iH+w_pH_iL))/d;
    if(eps_if_selftest_required>0)
    {
	final double eps=eps_if_selftest_required;
	// Test normalizing
	UnitTests.checkDoubleRel(w_pL_iL+w_pL_iH,"w_pL_iL+w_pL_iH",1,"1",eps);
	UnitTests.checkDoubleRel(w_pH_iL+w_pH_iH,"w_pH_iL+w_pH_iH",1,"1",eps);
	UnitTests.checkDoubleRel(w_pL_iL+w_pH_iL,"w_pL_iL+w_pH_iL",1,"1",eps);
	UnitTests.checkDoubleRel(w_pL_iH+w_pH_iH,"w_pL_iH+w_pH_iH",1,"1",eps);
	// Test calculations in original basis
	final double [] ev_MpM=EV.getGEV_Generalized(d,QQp,d,QQ,d,true);
	UnitTests.checkDoubleRel(w_pL_iL,"w_pL_iL",Math.pow(TMatr.getAMatrB(d,ev_MpM,d+0*d,ev_MiM,d+0*d,QQ,d),2),"w_pL_iL",eps);
	UnitTests.checkDoubleRel(w_pL_iH,"w_pL_iH",Math.pow(TMatr.getAMatrB(d,ev_MpM,d+0*d,ev_MiM,d+1*d,QQ,d),2),"w_pL_iH",eps);
	UnitTests.checkDoubleRel(w_pH_iL,"w_pH_iL",Math.pow(TMatr.getAMatrB(d,ev_MpM,d+1*d,ev_MiM,d+0*d,QQ,d),2),"w_pH_iL",eps);
	UnitTests.checkDoubleRel(w_pH_iH,"w_pH_iH",Math.pow(TMatr.getAMatrB(d,ev_MpM,d+1*d,ev_MiM,d+1*d,QQ,d),2),"w_pH_iH",eps);
	final double corr_test=(TMatr.getAMatrA(d,ev_MiM,d+0*d,QQp,d)-TMatr.getAMatrA(d,ev_MiM,d+1*d,QQp,d))
	    /(ev_MpM[0]-ev_MpM[1]);
	UnitTests.checkDoubleRel(corr,"corr",corr_test,"corr_test",eps);
	System.err.println("ProbabilityCorrelation:(<phi_IL|p|<phi_IL> - <phi_IH|p|<phi_IH>)/(p_min-p_max)="+corr_test+" corr="+corr);

	final double [] psiC={1/Math.sqrt(QQ[0]),0};
	final double _skewnessLike_i=Math.pow(TMatr.getAMatrB(d,psiC,0,ev_MiM,d+0*d,QQ,d),2)-
	    Math.pow(TMatr.getAMatrB(d,psiC,0,ev_MiM,d+1*d,QQ,d),2),
	 _skewnessLike_p=Math.pow(TMatr.getAMatrB(d,psiC,0,ev_MpM,d+0*d,QQ,d),2)-
	    Math.pow(TMatr.getAMatrB(d,psiC,0,ev_MpM,d+1*d,QQ,d),2);
	UnitTests.checkDoubleRel(_skewnessLike_i,"_skewnessLike_i",skewnessLike_i,"skewnessLike_i",eps);
	UnitTests.checkDoubleRel(_skewnessLike_p,"_skewnessLike_p",skewnessLike_p,"skewnessLike_p",eps);
	System.err.println("ProbabilityCorrelation: match OK eps="+eps);
    }
}


/** Copied from EVXData. */
static void convertMatrFromQToB_Basis(final double [] res,final int n,final double [] matrQ,final double [] basis,final int nd,final int basis_offset,final double [] tmp)
{
    // matrQ: n x n matrix in original Q basis.
    // basis: nd x n matrix original Q basis.
    // res: nd x nd matrix in new basis
    TMatr.setMA_x_MBTrans(tmp,n,matrQ,n,0,n,basis,nd,n,basis_offset+0*n); // tmp: n x nd
    TMatr.setMA_x_MB(res,nd,basis,n,basis_offset+0*n,n,tmp,nd,nd,0); // res: nd x nd
}
public static void matchProbabilityCorrelation(final ProbabilityCorrelation a,final ProbabilityCorrelation b,final double eps)
{
    matchProbabilityCorrelationWithOffset(a,0,b,0,eps);
}
public static void matchProbabilityCorrelationWithOffset(final ProbabilityCorrelation a,final double off_pa,final ProbabilityCorrelation b,final double off_pb,final double eps)
{
    final double poffset_shift=off_pb-off_pa;
    UnitTests.checkDoubleRel(a.p_aver,"a.p_aver",b.p_aver+poffset_shift,"b.p_aver+poffset_shift",eps);
    UnitTests.checkDoubleRel(a.pL,"a.pL",b.pL+poffset_shift,"b.pL+poffset_shift",eps);
    UnitTests.checkDoubleRel(a.pH,"a.pH",b.pH+poffset_shift,"b.pH+poffset_shift",eps);
    UnitTests.checkDouble(a.skewnessLike_p,"a.skewnessLike_p",b.skewnessLike_p,"b.skewnessLike_p",eps);
    UnitTests.checkDoubleRel(a.i_aver,"a.i_aver",b.i_aver,"b.i_aver",eps);
    UnitTests.checkDoubleRel(a.iL,"a.iL",b.iL,"b.iL",eps);
    UnitTests.checkDoubleRel(a.iH,"a.iH",b.iH,"b.iH",eps);
    UnitTests.checkDouble(a.skewnessLike_i,"a.skewnessLike_i",b.skewnessLike_i,"b.skewnessLike_i",eps);
    UnitTests.checkDoubleRel(a.w_pL_iL,"a.w_pL_iL",b.w_pL_iL,"b.w_pL_iL",eps);
    UnitTests.checkDoubleRel(a.w_pL_iH,"a.w_pL_iH",b.w_pL_iH,"b.w_pL_iH",eps);
    UnitTests.checkDoubleRel(a.w_pH_iL,"a.w_pH_iL",b.w_pH_iL,"b.w_pH_iL",eps);
    UnitTests.checkDoubleRel(a.w_pH_iH,"a.w_pH_iH",b.w_pH_iH,"b.w_pH_iH",eps);
    UnitTests.checkDoubleRel(a.W,"a.W",b.W,"b.W",eps);

    System.err.println("ProbabilityCorrelation:matchProbabilityCorrelation test OK eps="+eps);
}
}
