package com.polytechnik.utils;
/*
 * (C) Vladislav Malyshkin 2010
 * This file is under GPL version 3.
 *
 */

/** 5 points stable Cholesky elimination.
 *  Algorithm from:
 *  A.A.Samarskiy, E.S.Nikolaev, Grid Calculations Methods, 1978, p. 73.
 *  @version $Id: Progonka5.java,v 1.39 2012/04/19 22:24:10 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
public class Progonka5  {


    /** 5 non-zero points in a row.
     *  @param am diag-2 element.
     *  @param bm diag-1 element.
     *  @param cm diag element.
     *  @param dm diag_1 element.
     *  @param em diag element.
     *  @param fm right part.
     *  @param y solution.
     */
    public static void eliminate(final int n,
				 final double [] am,
				 final double [] bm,
				 final double [] cm,
				 final double [] dm,
				 final double [] em,
				 final double [] fm,
				 final double [] y,

				 final double [] alfa,
				 final double [] beta,
				 final int [] in1,
				 final int [] in2,
				 final int [] in3)
    {
	if(n<5)
	{
	    throw new IllegalArgumentException();
	}
	/*
	// for debugging
	java.util.Arrays.fill(in1,-1);
	java.util.Arrays.fill(in2,-1);
	java.util.Arrays.fill(in3,-1);
	java.util.Arrays.fill(alfa,Double.NaN);
	java.util.Arrays.fill(beta,Double.NaN);
	*/


	final int n2=n-2;
	double c=cm[0],d=-dm[0],f=fm[0],b=bm[1],q=cm[1],u=fm[1],s=am[2],t=-bm[2],g=fm[2],r=0,a=am[3],h=fm[3];
	int prev_in2=0,prev_in3=1;
	for(int i=0;i<n2;i++)
	{
	    final int i2=i+2,i3=i2+1,i4=i3+1;
	    final double em_i=em[i],dm_i1=dm[i+1],cm_i2=cm[i2],absc=Math.abs(c),absd=Math.abs(d),abs_em_i=Math.abs(em_i);
	    if(absc>=absd && absc>=abs_em_i)
	    {
		final double ali=alfa[i]=d/c,bei=beta[i]=em_i/c,gai=y[prev_in2]=f/c;
		in1[i]=prev_in2;
		prev_in2=in2[i]=prev_in3;
		prev_in3=in3[i]=i2;

		c=q+b*ali;
		d=b*bei-dm_i1;
		f=u-b*gai;
		b=s*ali-t;
		q=cm_i2-s*bei;
		u=g-s*gai;
		if(i3>=n)
		{
		    continue;
		}
		s=a+r*ali;
		t=r*bei-bm[i3];
		g=h-r*gai;
		if(i4>=n)
		{
		    continue;
		}
		r=0;
		a=am[i4];
		h=fm[i4];
	    }
	    else if(absd>=abs_em_i)
	    {
		final double ali=alfa[i]=c/d,bei=beta[i]=-em_i/d,gai=y[prev_in3]=-f/d;
		in1[i]=prev_in3;
		in2[i]=prev_in2;
		prev_in3=in3[i]=i2;

		c=q*ali+b;
		d=q*bei-dm_i1;
		f=u-q*gai;
		b=s-t*ali;
		q=t*bei+cm_i2;
		u=t*gai+g;
		if(i3>=n)
		{
		    continue;
		}
		s=a*ali+r;
		t=a*bei-bm[i3];
		g=h-a*gai;
		if(i4>=n)
		{
		    continue;
		}
		r=0;
		a=am[i4];
		h=fm[i4];
	    }
	    else
	    {
		final double ali=alfa[i]=d/em_i,bei=beta[i]=c/em_i,gai=y[i2]=f/em_i;
		in1[i]=i2;
		final int ps_prev_in2=prev_in2;
		prev_in2=in2[i]=prev_in3;
		prev_in3=in3[i]=ps_prev_in2;

		c=q+dm_i1*ali;
		d=dm_i1*bei-b;
		f=u-dm_i1*gai;
		b=cm_i2*ali-t;
		q=s-cm_i2*bei;
		u=g-cm_i2*gai;
		if(i3>=n)
		{
		    continue;
		}
		s=a+bm[i3]*ali;
		t=bm[i3]*bei-r;
		g=h-bm[i3]*gai;
		if(i4>=n)
		{
		    continue;
		}
		r=am[i4]*ali;
		a=-am[i4]*bei;
		h=fm[i4]-am[i4]*gai;
	    }
	}



	if(Math.abs(c)>=Math.abs(d))
	{
	    final double ali=d/c,gai=f/c,v=(u-b*gai)/(q+b*ali);
	    y[prev_in3]=v;
	    y[prev_in2]=ali*v+gai;
	}
	else
	{
	    final double ali=c/d,gai=f/d,v=(u+q*gai)/(b+q*ali);
	    y[prev_in2]=v;
	    y[prev_in3]=ali*v-gai;
	}

	for(int i=n2;i-->0;)
	{
	    // alfas & betas are saved with off by 1 index to allow dm & em usage as temporary storage for alfa & beta
	    y[in1[i]]+=alfa[i]*y[in2[i]]-beta[i]*y[in3[i]];
	}
    }



    /**
     * Test.
     * java  com/polytechnik/utils/Progonka5
     * checkLinSystem5: 100000 tests OK done in 144772 milliseconds. ver=$Id: Progonka5.java,v 1.39 2012/04/19 22:24:10 mal Exp $
     */
    private static void checkLinSystem5()
    {
	// a test with random matrix
	final java.util.Random r=new java.util.Random(-79093);
	final long t0=System.currentTimeMillis();
	final double eps=1e-5;
	int i=0;
	for(;i<100000;i++)
	{
	    final int n=i<1000&&false?(r.nextInt(8)):(5+r.nextInt(1+10000));
	    final double [] a=new double [n],b=new double [n],c=new double [n],
		d=new double [n],e=new double [n],f=new double [n],
		y=new double [n],al=new double [n],bt=new double [n];
	    final int [] i1=new int [n],i2=new int [n],i3=new int [n];
	    for(int j=0;j<n;j++)
	    {
		a[j]=r.nextDouble();
		b[j]=r.nextDouble();
		c[j]=r.nextDouble();
		d[j]=r.nextDouble();
		e[j]=r.nextDouble();
		f[j]=r.nextDouble();
	    }

	    eliminate(n,a,b,c,d,e,f,y,al,bt,i1,i2,i3);
	    double s=0;
	    if(n>0) s+=Math.abs(c[0]*y[0]+(n>1?d[0]*y[1]:0)+(n>2?e[0]*y[2]:0)-f[0]);
	    if(n>1) s+=Math.abs(b[1]*y[0]+c[1]*y[1]+(n>2?d[1]*y[2]:0)+(n>3?e[1]*y[3]:0)-f[1]);
	    if(n>3) s+=Math.abs(a[n-2]*y[n-4]+b[n-2]*y[n-3]+c[n-2]*y[n-2]+d[n-2]*y[n-1]-f[n-2]);
	    if(n>2) s+=Math.abs(a[n-1]*y[n-3]+b[n-1]*y[n-2]+c[n-1]*y[n-1]-f[n-1]);
	    for(int j=2;j<n-2;j++)
	    {
		s+=Math.abs(a[j]*y[j-2]+b[j]*y[j-1]+c[j]*y[j]+d[j]*y[j+1]+e[j]*y[j+2]-f[j]);
	    }
	    if(s>eps||Double.isNaN(s))
	    {
		throw new RuntimeException("s="+s+" n="+n);
	    }



	    // try with storage re-usage
	    final double [] es=e.clone(),ds=d.clone(),ys=f.clone();
	    eliminate(n,a,b,c,ds,es,ys,ys,ds,es,i1,i2,i3);
	    for(int j=0;j<n;j++)
	    {
		if(!(ys[j]==y[j]))
		{
		    throw new RuntimeException("ys["+j+"="+ys[j]+" y["+j+"="+y[j]+" n="+n);
		}
	    }
	}
	final long t1=System.currentTimeMillis();
	System.err.println("checkLinSystem5: "+i+" tests OK done in "+(t1-t0)+" milliseconds. ver=$Id: Progonka5.java,v 1.39 2012/04/19 22:24:10 mal Exp $");
    }




    public static void main(String [] args)
    {
	checkLinSystem5();
    }
}
