package com.polytechnik.utils;
/*
 * (C) Vladislav Malyshkin 2013
 * This file is under GPL version 3.
 *
 */

/** Rayleigh Quotient derivatives estimation.
 *  @version $Id: RayleighQuotient.java,v 1.15 2016/11/06 13:23:55 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
public class RayleighQuotient {


/** Determine variations <psi+dpsi|QQf|psi+dpsi>/<psi+dpsi|QQt|psi+dpsi>.
 *  Total variation is d0+<d1|dpsi>+(1/2)<dpsi|d2|dpsi>.
 */
public static class RQVariationInMatrixForm {
    public final double d0;
    public final double [] d1;
    public final double [] d2;
    public final int ndpsi;
    public RQVariationInMatrixForm(final double [] psi,final int ndpsi,final double [] QQf,final double [] QQt,final int ldQQ)
    {
	this.ndpsi=ndpsi;
	final double nt=TMatr.getAMatrA(psi.length,psi,QQt,ldQQ),f=TMatr.getAMatrA(psi.length,psi,QQf,ldQQ)/nt;
	this.d0=f;
	final double [] t=new double [ndpsi];
	this.d1=new double [ndpsi];
	this.d2=new double [ndpsi*ndpsi];
	for(int k=0;k<ndpsi;k++)
	{
	    t[k]=TMatr.mulVV(psi.length,QQt,k*ldQQ,psi,0)/nt;
	    this.d1[k]=2*(TMatr.mulVV(psi.length,QQf,k*ldQQ,psi,0)/nt-f*t[k]);
	    for(int l=0;l<=k;l++)
	    {
		this.d2[k*ndpsi+l]=this.d2[l*ndpsi+k]=2*(QQf[k*ldQQ+l]-f*QQt[k*ldQQ+l])/nt
		    -2*(t[k]*this.d1[l]+t[l]*this.d1[k]);
	    }
	}
    }
    public RQVariationInMatrixForm(final RQFact nom,final RQFact denom)
    {
	if(nom.n!=denom.n) throw new IllegalArgumentException();
	this.ndpsi=nom.n;
	final double nt=denom.f0,f=nom.f0/nt;
	this.d0=f;
	this.d1=new double [ndpsi];
	this.d2=new double [ndpsi*ndpsi];
	for(int k=0;k<ndpsi;k++)
	{
	    this.d1[k]=nom.f1[k]/nt-f*(denom.f1[k]/nt);
	    for(int l=0;l<=k;l++)
	    {
		this.d2[k*ndpsi+l]=this.d2[l*ndpsi+k]=
		    (nom.f2[k*ndpsi+l]-f*denom.f2[k*ndpsi+l])/nt
		    -(denom.f1[k]*this.d1[l]+denom.f1[l]*this.d1[k])/nt;
	    }
	}
    }
} // END OF class RQVariationInMatrixForm


/** Rayleigh--like (with linear term) nominator or denominator.
 * l2 is assumed to be symmetric.
 */
public static class RQFact {
    final int n;
    final double f0, f1[],f2[];
    public RQFact(final double [] psi,
		  final double l0,final double [] l1,final double [] l2,final int lDl2)
    {
	n=psi.length;
	f0=l0+TMatr.mulVV(psi,l1)+TMatr.getAMatrA(psi,l2,lDl2);
	f1=TMatr.getMatr_x_A(n,l2,lDl2,psi);
	for(int k=0;k<n;k++) f1[k]=l1[k]+2*f1[k];
	f2=new double [n*n];
	for(int k=0;k<n;k++) for(int l=0;l<=k;l++) f2[k*n+l]=f2[l*n+k]=2*l2[k*lDl2+l];
    }
}


/**
 *  Calculate d^m/ds^m <psi(x+s)|f|psi(x+s)>/<psi(x+s)|psi(x+s)> at s=0
 *  for m=0,1,2. Differentiate over s, set s to 0, then average over x.
 *  Typical use is for testing. For production use matrix version RQVariationInMatrixForm.
 *  @param psi psi.
 *  @param Dpsi d psi/dx.
 *  @param D2psi d^2 psi/dx^2.
 *  @param QQf The <Q_i f Q_j> matrix.
 *  @param QQt The <Q_i Q_j> matrix.
 *  @param ldQQ The leading dimension of QQ matrices.
 *  @return An array containing D0 D1 D2 derivatives.
 */
public static double [] getRQVariationValues(final double [] psi,final double [] Dpsi,final double [] D2psi,final double [] QQf,final double [] QQt,final int ldQQ)
{
    //System.err.println("ldQQ="+ldQQ+" psi.length="+psi.length);
    final double dn=TMatr.getAMatrA(psi.length,psi,QQt,ldQQ),
	f=TMatr.getAMatrA(psi.length,psi,QQf,ldQQ)/dn,
	df=2*TMatr.getAMatrB(psi.length,psi,0,Dpsi.length,Dpsi,0,QQf,ldQQ)/dn,
	dg=2*TMatr.getAMatrB(psi.length,psi,0,Dpsi.length,Dpsi,0,QQt,ldQQ)/dn,
	ddf=2*(TMatr.getAMatrA(Dpsi.length,Dpsi,QQf,ldQQ)+(D2psi==null?0:TMatr.getAMatrB(psi.length,psi,0,D2psi.length,D2psi,0,QQf,ldQQ)))/dn,
	ddg=2*(TMatr.getAMatrA(Dpsi.length,Dpsi,QQt,ldQQ)+(D2psi==null?0:TMatr.getAMatrB(psi.length,psi,0,D2psi.length,D2psi,0,QQt,ldQQ)))/dn;
    final double D0=f,D1=(df-f*dg),D2cut=(ddf-f*ddg),D2=D2cut-2*dg*D1;
    return new double []{D0,D1,D2,D2cut};
}

/**
 *  Calculate d^m/ds^m <psi(x+s)|f|phi(x+s)>/<psi(x+s)|phi(x+s)> at s=0
 *  for m=0,1,2. Differentiate over s, set s to 0, then average over x.
 *  Typical use is for testing. For production use matrix version RQVariationInMatrixForm.
 *  @param psi psi.
 *  @param Dpsi d psi/dx.
 *  @param D2psi d^2 psi/dx^2.
 *  @param phi phi.
 *  @param Dphi d phi/dx.
 *  @param D2phi d^2 phi/dx^2.
 *  @param QQf The <Q_i f Q_j> matrix.
 *  @param QQt The <Q_i Q_j> matrix.
 *  @param ldQQ The leading dimension of QQ matrices.
 *  @return An array containing D0 D1 D2 derivatives.
 */
public static double [] getRQVariationValues(final double [] psi,final double [] Dpsi,final double [] D2psi,
					     final double [] phi,final double [] Dphi,final double [] D2phi,
					     final double [] QQf,final double [] QQt,final int ldQQ)
{
    //System.err.println("ldQQ="+ldQQ+" psi.length="+psi.length);
    final double dn=TMatr.getAMatrB(psi,phi,QQt,ldQQ),
	f=TMatr.getAMatrB(psi,phi,QQf,ldQQ)/dn,
	df=(TMatr.getAMatrB(psi,Dphi,QQf,ldQQ)+TMatr.getAMatrB(Dpsi,phi,QQf,ldQQ))/dn,
	dg=(TMatr.getAMatrB(psi,Dphi,QQt,ldQQ)+TMatr.getAMatrB(Dpsi,phi,QQt,ldQQ))/dn,
	ddf=(TMatr.getAMatrB(Dpsi,Dphi,QQf,ldQQ)+TMatr.getAMatrB(Dphi,Dpsi,QQf,ldQQ)+TMatr.getAMatrB(psi,D2phi,QQf,ldQQ)+TMatr.getAMatrB(D2psi,phi,QQf,ldQQ))/dn,
	ddg=(TMatr.getAMatrB(Dpsi,Dphi,QQt,ldQQ)+TMatr.getAMatrB(Dphi,Dpsi,QQt,ldQQ)+TMatr.getAMatrB(psi,D2phi,QQt,ldQQ)+TMatr.getAMatrB(D2psi,phi,QQt,ldQQ))/dn;
    final double D0=f,D1=(df-f*dg),D2=(ddf-f*ddg)-2*dg*D1;
    return new double []{D0,D1,D2};
}


static class TestRayleighQuotient {
    final TestMonomials TM=new OrthogonalPolynomialsChebyshevBasis.TestChebyshevInMonomials();
    final OrthogonalPolynomialsABasis M=TM.O;
    final double eps=1e-7;
    final String name;
    final java.util.Random r;
    TestRayleighQuotient(final String text,final long random)
    {
	this.name=text;
	this.r=new java.util.Random(random);
    }
    void test_MatrixAndValues(final int n)
    {
	final int ntests=700,size=12001,psisize=89,dd=5;
	for(int t=0;t<ntests;t++)
	{
	    final double [] momf=TM.getBasisRandomMoments(2*(n+dd)-1,TM.x_min,TM.x_max,size,this.r),
		momt=TM.getBasisRandomMoments(2*(n+dd)-1,TM.x_min,TM.x_max,size,this.r);
	    final double [] psi=TM.getBasisRandomMoments(n+r.nextInt(dd),TM.x_min,TM.x_max,psisize,this.r),
		dpsi=TM.getBasisRandomMoments(n+r.nextInt(dd),TM.x_min,TM.x_max,psisize,this.r);
	    final double [] mv;
	    {
	    final int d=Math.max(psi.length,dpsi.length),ldQQ=d+r.nextInt(dd-(d-n)+1);
	    mv=getRQVariationValues(psi,dpsi,null,M.getQQMatr(ldQQ,momf),M.getQQMatr(ldQQ,momt),ldQQ);
	    }
	    final RQVariationInMatrixForm mm;
	    {
	    final int d=Math.max(psi.length,dpsi.length),ldQQ=d+r.nextInt(dd-(d-n)+1);
	    mm=new RQVariationInMatrixForm(psi,dpsi.length,M.getQQMatr(ldQQ,momf),M.getQQMatr(ldQQ,momt),ldQQ);
	    }
	    UnitTests.checkDoubleRel(mv[0],"mv[0]",mm.d0,"mm.d0",eps);
	    UnitTests.checkDoubleRel(mv[1],"mv[1]",TMatr.mulVV(dpsi.length,dpsi,mm.d1),"<dpsi|d1>",eps);
	    UnitTests.checkDoubleRel(mv[2],"mv[2]",TMatr.getAMatrA(dpsi.length,dpsi,mm.d2,mm.ndpsi),"<dpsi|d2|dpsi>",eps);
	    test_RQMatrixAndValues(psi.length,psi,M.getQQMatr(psi.length,momf),M.getQQMatr(psi.length,momt));
	}
	System.err.println(name+":test_MatrixAndValues(): test OK"+" n="+n+" ntests="+ntests+" eps="+eps);
    }

    void test_RQMatrixAndValues(final int n,final double [] psi,final double [] QQf,final double [] QQt)
    {
	final RQVariationInMatrixForm mm=new RQVariationInMatrixForm(psi,n,QQf,QQt,n);
	final int nd=n-1;
	for(int ind=0;ind<psi.length;ind++)
	{
	    final double [][] nom=getMatrSplitted(ind,psi,QQf,n),
		denom=getMatrSplitted(ind,psi,QQt,n);
	    final RQVariationInMatrixForm mmind=
		new RQVariationInMatrixForm(new RQFact(nom[3],nom[0][0],nom[1],nom[2],nd),
					    new RQFact(denom[3],denom[0][0],denom[1],denom[2],nd));
	     UnitTests.checkDoubleRel(mm.d0,"mm.d0",mmind.d0,"mmind.d0",eps);
	     for(int l=0,o=0,oo=0;l<n;l++)
	     {
		 if(l==ind) continue;
		 UnitTests.checkDoubleRel(mm.d1[l],"mm.d1[l]",mmind.d1[o++],"mmind.d1[o++]",eps);
		 for(int j=0;j<n;j++)
		 {
		     if(j==ind) continue;
		     UnitTests.checkDoubleRel(mm.d2[l*n+j],"mm.d2[l*n+j]",mmind.d2[oo++],"mmind.d2[oo++]",eps);
		 }
	     }
	}
	//System.err.println(name+":test_RQMatrixAndValues(): test OK"+" n="+n+" eps="+eps);
    }
    static double [][] getMatrSplitted(final int ind,final double [] psi,final double [] QQ,final int n)
    {
	final int nd=n-1;
	double f0=psi[ind]*psi[ind]*QQ[ind+n*ind];
	final double [] f1=new double [nd],f2=new double [nd*nd],ppsi=new double [nd];
	for(int k=0,ok=0,ook=0;k<n;k++)
	{
	    if(k!=ind) {
		ppsi[ook]=psi[k];
		f1[ook++]=2*psi[ind]*QQ[ind+n*k];
	    }
	    for(int l=0;l<n;l++)
	    {
		if(k!=ind&&l!=ind) f2[ok++]=QQ[l+n*k];
	    }
	}
	return new double [][]{new double []{f0},f1,f2,ppsi};
    }

    void test_Values(final int n)
    {
	final int ntests=701,size=12001,psisize=89,dd=5;
	for(int t=0;t<ntests;t++)
	{
	    final double [] momf=TM.getBasisRandomMoments(2*(n+dd)-1,TM.x_min,TM.x_max,size,this.r),
		momt=TM.getBasisRandomMoments(2*(n+dd)-1,TM.x_min,TM.x_max,size,this.r);
	    final double [] psi=TM.getBasisRandomMoments(n+r.nextInt(dd),TM.x_min,TM.x_max,psisize,this.r);
	    final int ldQQ=psi.length;

	    final double [] dpsi=M.B.getDifferentiated(psi.length-1,psi),
		d2psi=dpsi.length<=1?null:M.B.getDifferentiated(dpsi.length-1,dpsi);
	    final double [] rq=getRQVariationValues(psi,dpsi,d2psi,M.getQQMatr(ldQQ,momf),M.getQQMatr(ldQQ,momt),ldQQ);

	    final double [] psi2=new double [2*psi.length-1];
	    M.B.setMult2Pol(psi2,psi,psi.length-1,psi,psi.length-1,new double [psi2.length]);
	    final double [] dpsi2=M.B.getDifferentiated(psi2.length-1,psi2),
		d2psi2=dpsi2.length<=1?new double []{}:M.B.getDifferentiated(dpsi2.length-1,dpsi2);
	    final double [] nom=new double []{
		TMatr.mulVV(psi2.length,psi2,momf),TMatr.mulVV(dpsi2.length,dpsi2,momf),0.5*TMatr.mulVV(d2psi2.length,d2psi2,momf)},
		denom=new double []{
		    TMatr.mulVV(psi2.length,psi2,momt),TMatr.mulVV(dpsi2.length,dpsi2,momt),0.5*TMatr.mulVV(d2psi2.length,d2psi2,momt)
		};
	     final double nt=Polynomials.getD0NoScale(denom,0),f=Polynomials.getD0NoScale(nom,0)/nt;
	     UnitTests.checkDoubleRel(rq[0],"rq[0]",f,"f",eps);
	     final double df=Polynomials.getD1NoScale(nom,0)/nt-f*Polynomials.getD1NoScale(denom,0)/nt;
	     UnitTests.checkDoubleRel(rq[1],"rq[1]",df,"df",eps);
	     final double d2f=Polynomials.getD2NoScale(nom,0)/nt-f*Polynomials.getD2NoScale(denom,0)/nt
		 -2*(Polynomials.getD1NoScale(denom,0)/nt)*df;
	     UnitTests.checkDoubleRel(rq[2],"rq[2]",d2f,"d2f",eps);
	}

	System.err.println(name+":test_Values(): test OK"+" n="+n+" ntests="+ntests+" eps="+eps);
    }
    void test_Values2(final int n)
    {
	final int ntests=701,size=12001,psisize=89,dd=5;
	for(int t=0;t<ntests;t++)
	{
	    final double [] momf=TM.getBasisRandomMoments(2*(n+dd)-1,TM.x_min,TM.x_max,size,this.r),
		momt=TM.getBasisRandomMoments(2*(n+dd)-1,TM.x_min,TM.x_max,size,this.r);
	    final double [] psi=TM.getBasisRandomMoments(n+r.nextInt(dd),TM.x_min,TM.x_max,psisize,this.r),
		phi=TM.getBasisRandomMoments(n+r.nextInt(dd),TM.x_min,TM.x_max,psisize,this.r);
	    final int ldQQ=Math.max(psi.length,phi.length)+r.nextInt(2);

	    final double [] dpsi=M.B.getDifferentiated(psi.length-1,psi),
		d2psi=(dpsi.length<=1?new double []{}:M.B.getDifferentiated(dpsi.length-1,dpsi)),
		dphi=M.B.getDifferentiated(phi.length-1,phi),
		d2phi=(dphi.length<=1?new double []{}:M.B.getDifferentiated(dphi.length-1,dphi));



	    final double [] rq=getRQVariationValues(psi,dpsi,d2psi,phi,dphi,d2phi,
						    M.getQQMatr(ldQQ,momf),M.getQQMatr(ldQQ,momt),ldQQ);

	    final double [] psi_phi=new double [psi.length+phi.length-1];
	    M.B.setMult2Pol(psi_phi,psi,psi.length-1,phi,phi.length-1,new double [psi_phi.length]);
	    final double [] d_psi_phi=M.B.getDifferentiated(psi_phi.length-1,psi_phi),
		d2_psi_phi=d_psi_phi.length<=1?new double []{}:M.B.getDifferentiated(d_psi_phi.length-1,d_psi_phi);
	    final double [] nom=new double []{
		TMatr.mulVV(psi_phi,momf),TMatr.mulVV(d_psi_phi,momf),0.5*TMatr.mulVV(d2_psi_phi,momf)},
		denom=new double []{
		    TMatr.mulVV(psi_phi,momt),TMatr.mulVV(d_psi_phi,momt),0.5*TMatr.mulVV(d2_psi_phi,momt)
		};
	     final double nt=Polynomials.getD0NoScale(denom,0),f=Polynomials.getD0NoScale(nom,0)/nt;
	     UnitTests.checkDoubleRel(rq[0],"rq[0]",f,"f",eps);
	     final double df=Polynomials.getD1NoScale(nom,0)/nt-f*Polynomials.getD1NoScale(denom,0)/nt;
	     UnitTests.checkDoubleRel(rq[1],"rq[1]",df,"df",eps);
	     final double d2f=Polynomials.getD2NoScale(nom,0)/nt-f*Polynomials.getD2NoScale(denom,0)/nt
		 -2*(Polynomials.getD1NoScale(denom,0)/nt)*df;
	     UnitTests.checkDoubleRel(rq[2],"rq[2]",d2f,"d2f",eps);
	}

	System.err.println(name+":test_Values2(): test OK"+" n="+n+" ntests="+ntests+" eps="+eps);
    }

    void runTests(final int neq_max)
    {
	System.err.println(name+":TestRayleighQuotient:starting tests; ver=$Id: RayleighQuotient.java,v 1.15 2016/11/06 13:23:55 mal Exp $");
	final long t0=System.currentTimeMillis();
	for(int n=1;n<=neq_max;n++)
	{
	    test_MatrixAndValues(n);
	    test_Values(n);
	    test_Values2(n);
	}
	final long t1=System.currentTimeMillis();
	System.err.println(name+":TestRayleighQuotient: tests OK done in "+(t1-t0)+" milliseconds. neq_max="+neq_max+" ver=$Id: RayleighQuotient.java,v 1.15 2016/11/06 13:23:55 mal Exp $");
    }
} // END OF class TestRayleighQuotient

public static void main(final String [] args)
{
    new TestRayleighQuotient("RayleighQuotient",178L).runTests(8);
}


}
