package com.polytechnik.utils;
/*
 * (C) Vladislav Malyshkin 2010
 * This file is under GPL version 3.
 *
 */

/** A crude and simple Newton style root finder algorithm.
 *  Used for simple(non-degenerated) real roots finding only.
 *  @version $Id: SimpleRootsFinder.java,v 1.8 2017/03/15 21:16:34 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */

public class SimpleRootsFinder  {

private static final int miniter=17;

public static double findRoots(final DerivativesAtCalculateable fun,final double x0,final int maxiterperstep,final double epsx)
{
    return findRoots(fun,x0,maxiterperstep,false,Double.NaN,Double.NaN,epsx);
}

public static double findRoots(final DerivativesAtCalculateable fun,final double x0,final int maxiterperstep,
			       boolean bounded,double xL,double xR,final double epsx)
{
    double x=x0,prevx=x0;
    int prevs=0,sL=0,sR=0;
    if(bounded)
    {
	if(xL>xR) return Double.NaN;
	fun.calculateAt(xL);
	sL=(int)Math.signum(fun.getF());
	fun.calculateAt(xR);
	sR=(int)Math.signum(fun.getF());
	if(sL*sR>0) return Double.NaN;
	if(!(x>=xL) || !(x<=xR)) {x=0.5*(xL+xR);}
	if(!(x>=xL) || !(x<=xR)) return Double.NaN;
    }
    else
    {
	xL=xR=Double.NaN;
    }
rf:{
    for(int k=0,ke=maxiterperstep+miniter+2,kx=0;k<ke;k++)
    {
	fun.calculateAt(x);
	final double fx=fun.getF(),dfx=fun.getDF();
	final int s;
	if(fx<0) s=-1;
	else if(fx>0) s=1;
	else break rf;
	double deltax=fx/dfx,x1=x-deltax; // newton iteration
	//System.err.println("k="+k+" at x="+x+" s="+s+" fx="+fx);
	if(bounded)
	{
	    if(true && (x<xL || x>xR)) {
		throw new RuntimeException("x="+x+" xL="+xL+" xR="+xR+" out of interval. This should never happen");
	    }
	    if(s==sL){xL=x;}else{xR=x;}
	    if(!(x1>=xL) || !(x1<=xR)) {
		//System.err.print("Bissection: x1="+x1+" xL="+xL+" xR="+xR);
		x1=0.5*(xL+xR);
		//System.err.println(" set x1="+x1);
	    }
	}
	else
	{
	    if(prevs*s<0)
	    {
		bounded=true;
		if(x<prevx){xL=x;xR=prevx;sL=s;sR=prevs;}
		else{xL=prevx;xR=x;sL=prevs;sR=s;}
		if(!(x1>=xL) || !(x1<=xR)) {
		    //System.err.print("Bissection: x1="+x1+" xL="+xL+" xR="+xR);
		    x1=0.5*(xL+xR);
		    //System.err.println(" set x1="+x1);
		}
	    }
	}
	if(!(Math.abs(deltax)<=epsx)){kx=0;}
	prevs=s;
	prevx=x;
	x=x1;
	//System.err.println("findRoots: k="+k+" x="+x+" prevx="+prevx+" s="+s+" prevs="+prevs+" bounded="+bounded+" xL="+xL+" xR="+xR+" sL="+sL+" sR="+sR+" kx="+kx+" deltax="+deltax+" epsx="+epsx+" deltax<epsx="+(Math.abs(deltax)<=epsx));
	//System.err.println(" EE: kx="+kx+" miniter="+miniter);
	if(++kx>miniter){break rf;}
    }
    // root calculation failed.
    x=Double.NaN;
    } // END rf
    fun.calculateAt(x);
    //System.err.println("Result x="+x+" getF()="+fun.getF());
    return x;
}


}
