package com.polytechnik.utils;
/*
 * (C) Vladislav Malyshkin 2014
 * This file is under GPL version 3.
 *
 */

/** Skewness estimator.
 *  @version $Id: Skewness.java,v 1.19 2018/01/05 19:55:04 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
public class Skewness {
    public final double xa,xm,sk,x1,w1,x2,w2;
    public Skewness(final double xa,
	     final double xm,
	     final double sk,
	     final double x1,final double w1,final double x2,final double w2)
    {
	this.xa=xa;
	this.xm=xm;
	this.sk=sk;
	this.x1=x1;
	this.w1=w1;
	this.x2=x2;
	this.w2=w2;
	//System.err.println("Skewness: sk="+sk+" xa="+xa+" xm="+xm+" x1="+x1+" w1="+w1+" x2="+x2+" w2="+w2);
    }

    /** Calculate modified skewness given 0..3 moments in arbitrary basis.
     */
    public static Skewness getSkewness2(final double m0,final double m1,final double m2,final double m3,final OrthogonalPolynomialsABasis M)
    {
	final double [] moms=new double []{m0,m1,m2,m3};
	final double [] P1=M.getSingleOrthogonalPolynomial(1,moms);
	final double xa=-M.fx(P1,0)/M.Dfx(P1,0);
	final double [] p=M.getSingleOrthogonalPolynomial(2,moms),
	    re=M.getXNodes(p),ws=M.getWeightsByIntegration(p,re,moms);
	final double xm=0.5*(re[0]+re[1]),sk=2*(xa-xm)/(re[0]-re[1]);
	//if(true) throw new RuntimeException();
	return new Skewness(xa,xm,sk,re[0],ws[0],re[1],ws[1]);
    }

    /** Same as getSkewness2, but different algorithm.
     */
    public static Skewness getSkewness(final double m0,final double m1,final double m2,final double m3,final OrthogonalPolynomialsABasis M)
    {
	final double [] moms=new double []{m0,m1,m2,m3};
	final double [][] Q=M.getQuadraturesFromMoments_EV(2,moms);
	final double [] x=Q[0],w=Q[1];
	final double nr=w[0]+w[1],xa=(x[0]*w[0]+x[1]*w[1])/nr,
	    xm=0.5*(x[0]+x[1]),sk=(w[0]-w[1])/nr;
	return new Skewness(xa,xm,sk,x[0],w[0],x[1],w[1]);
    }

/** Generalized skewness from monomial moments.
 */
public static Skewness getGSkewnessFromMonomialMoments(final double m0,final double m1,final double m2,
						       final double m0x,final double m1x,final double m2x)
{
    return getGSkewness(m0,m1,m2,m0x,m1x,m2x,MB);
}
/** Generalized skewness from the moments in arbitraty basis.
 */
public static Skewness getGSkewness(final double m0,final double m1,final double m2,
				    final double m0x,final double m1x,final double m2x,
				    final OrthogonalPolynomialsABasis M)
{
    return getGSkewness(M.getQQMatr(2,new double []{m0,m1,m2}),
			M.getQQMatr(2,new double []{m0x,m1x,m2x}),M);
}
/** Generalized skewness.
 */
public static Skewness getGSkewness(final double [] QQ,final double [] QQx,final OrthogonalPolynomialsABasis M)
{
    final int d=2;
    if(QQ.length!=d*d || QQx.length!=d*d)
	throw new IllegalArgumentException("Matrix dimension is not 2");
    final double [] ev=M.EV.getGEV_Generalized(d,QQx,d,QQ,d,false);
    final double [] ONE=new double []{1};{ONE[0]/=M.fx(ONE,0);}
    final double xa=QQx[0]/QQ[0],x1=ev[0],x2=ev[d-1],dif=x1-x2,
	W=TMatr.getAMatrA(ONE,QQ,d),w1=W*((xa-x2)/dif),w2=W*((x1-xa)/dif);
    return new Skewness(xa,(x1+x2)/2,(2*xa-x1-x2)/dif,x1,w1,x2,w2);
}

/** Generalized skewness, but uses <psi0|x|psi0> instead of average x.
 * calls getG0Skewness. basis_arg0 is the argument Q(basis_arg0) defining the |psi0>.
 */
public static Skewness getG0SkewnessFromMonomialMoments(final double m0,final double m1,final double m2,
							final double m0x,final double m1x,final double m2x,
							final double basis_arg0)
{
    final OrthogonalPolynomialsMonomialsBasis M=MB;
    return getG0Skewness(M.getQQMatr(2,new double []{m0,m1,m2}),
			 M.getQQMatr(2,new double []{m0x,m1x,m2x}),M,basis_arg0);
}

/** Generalized skewness, but uses <psi0|x|psi0> instead of average x.
 *  The basis_arg0 is taken from M as Radau root.
 */
public static Skewness getG0Skewness(final double [] QQ,final double [] QQx,final OrthogonalPolynomialsABasis M)
{
    return getG0Skewness(QQ,QQx,M,M.getRadauRoot());
}

/** Generalized skewness, but uses <psi0|x|psi0> instead of average x.
 *  The basis_arg0 is the argument Q(basis_arg0) defining the |psi0>.
 * (Also uses different normalizing w1+w2=1 instead of w1+w2=mu).
 */
public static Skewness getG0Skewness(final double [] QQ,final double [] QQx,final OrthogonalPolynomialsABasis M,final double basis_arg0)
{
    final int d=2;
    if(QQ.length!=d*d || QQx.length!=d*d)
	throw new IllegalArgumentException("Matrix dimension is not 2");
    final double [] ev=M.EV.getGEV_Generalized(d,QQx,d,QQ,d,true);
    final double [] psik_squared_at_x0=new double [d],tmp=new double [d];
    for(int k=0;k<d;k++)
    {
	System.arraycopy(ev,d+k*d,tmp,0,d);
	psik_squared_at_x0[k]=Math.pow(M.fx(tmp,basis_arg0),2);
    }
    final double o_o=TMatr.sumElements(psik_squared_at_x0),w1=psik_squared_at_x0[0]/o_o,w2=psik_squared_at_x0[d-1]/o_o,
	x1=ev[0],x2=ev[d-1];
    // For G0 we use the normalizing w1+w2=1, not w1+w2=mu as for G.
    return new Skewness(x1*w1+x2*w2,(x1+x2)/2,w1-w2,x1,w1,x2,w2);
}

/** Calculate modified skewness given 0..3 moments in monomials basis.
 */
public static Skewness getSkewnessFromMonomialMoments(final double m0,final double m1,final double m2,final double m3)
{
    return getSkewness(m0,m1,m2,m3,MB);
}
public static Skewness getSkewnessFromMonomialMoments2(final double m0,final double m1,final double m2,final double m3)
{
    return getSkewness2(m0,m1,m2,m3,MB);
}
static final OrthogonalPolynomialsMonomialsBasis MB=new OrthogonalPolynomialsMonomialsBasis();


public static void matchSkewness(final Skewness a,final Skewness b,final double eps)
{
    matchSkewnessWithOffset(a,0,b,0,eps);
}
public static void matchSkewnessWithOffset(final Skewness a,final double off_a,final Skewness b,final double off_b,final double eps)
{
    final double offset_shift=off_b-off_a;
    UnitTests.checkDoubleRel(a.xa,"a.xa",b.xa+offset_shift,"b.xa+offset_shift",eps);
    UnitTests.checkDoubleRel(a.xm,"a.xm",b.xm+offset_shift,"b.xm+offset_shift",eps);
    UnitTests.checkDoubleRel(a.sk,"a.sk",b.sk,"b.sk",eps);
    UnitTests.checkDoubleRel(a.x1,"a.x1",b.x1+offset_shift,"b.x1+offset_shift",eps);
    UnitTests.checkDoubleRel(a.w1,"a.w1",b.w1,"b.w1",eps);
    UnitTests.checkDoubleRel(a.x2,"a.x2",b.x2+offset_shift,"b.x2+offset_shift",eps);
    UnitTests.checkDoubleRel(a.w2,"a.w2",b.w2,"b.w2",eps);

    System.err.println("Skewness:matchSkewness test OK eps="+eps);
}

}
