package com.polytechnik.utils;
/*
 * (C) Vladislav Malyshkin 2011
 * This file is under GPL version 3.
 *
 */

/** Simple (and inefficient) matrix multiplication and linear systems.
 *  Use for testing only.
 *  @version $Id: TMatr.java,v 1.62 2018/01/27 18:22:42 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
public class TMatr {

public static double sumElements(final double [] pin)
{
    return sumElements(pin.length,pin);
}
public static double sumElements(final int n,final double [] pin)
{
    double s=0;
    for(int k=n;--k>=0;) s+=pin[k];
    return s;
}
public static double [] remove0Element(final double [] m)
{
    if(m.length==0) return new double []{};
    final double [] r=new double [m.length-1];
    System.arraycopy(m,1,r,0,m.length-1);
    return r;
}
public static double [] squareArrayElements(final double [] m)
{
    return getVAB(m.length,m,m);
}

public static void expandOverLowerDiagBasis(final double [] res,final double [] p,final int nels,final double [][] basis)
{
    System.arraycopy(p,0,res,0,nels);
    for(int k=nels; --k>=0;)
    {
	final double [] b=basis[k];
	final double e=res[k]/b[k];
	sumP1(res,b,k,-e);
	res[k]=e;
    }
}


public static void sumP1(final double [] res,final double [] pin,final double w)
{
    sumP1(res,pin,pin.length,w);
}
public static void sumP1(final double [] res,final double [] pin,final int nels,final double w)
{
    sumP1(res,pin,0,nels,w);
}
public static void sumP1(final double [] res,final double [] pin,final int pin_offset,final int nels,final double w)
{
    for(int k=0;k<nels;k++)
    {
	res[k]+=w*pin[k+pin_offset];
    }
}
public static double [] sumAB(final int n,final double [] A,final double [] B)
{
    final double [] r=new double [n];
    for(int k=0;k<n;k++)
    {
	r[k]=A[k]+B[k];
    }
    return r;

}
public static double [] sumAcB(final double [] A,final double c,final double [] B)
{
    if(A.length!=B.length) throw new IllegalArgumentException();
    return sumAcB(A.length,A,c,B);
}
public static double [] sumAcB(final int n,final double [] A,final double c,final double [] B)
{
    final double [] r=new double [n];
    for(int k=0;k<n;k++)
    {
	r[k]=A[k]+c*B[k];
    }
    return r;
}
public static double [] sumaAbB(final double a,final double [] A,final double b,final double [] B)
{
    if(A.length!=B.length) throw new IllegalArgumentException();
    return sumaAbB(A.length,a,A,b,B);
}
public static double [] sumaAbB(final int n,final double a,final double [] A,final double b,final double [] B)
{
    final double [] r=new double [n];
    for(int k=0;k<n;k++)
    {
	r[k]=a*A[k]+b*B[k];
    }
    return r;
}

/** Scale vector and return it.
 */
public static double [] scaleVector(final int n,final double [] v,final double w)
{
    for(int k=n;--k>=0;){v[k]*=w;}
    return v;
}
public static double [] scaleVector(final double [] v,final double w)
{
    return scaleVector(v.length,v,w);
}

public static double [] getVectorScaled(final double [] pin,final double w)
{
    final double [] r=new double [pin.length];
    sumP1(r,pin,pin.length,w);
    return r;
}

public static double [] mulMV(final double [][] m,final double [] a)
{
    final double [] b=new double [m.length];
    for(int k=0;k<b.length;k++)
    {
	b[k]=mulVV(m[k],a);
    }
    return b;
}
/** Same as mulVV(n,a,b) but takes a.length as n.
 */
public static double mulVV(final double [] a,final double [] b)
{
    //if(a.length!=b.length) throw new IllegalArgumentException();
    return mulVV(a.length,a,b);
}
public static double mulVV(final int n,final double [] a,final double [] b)
{
    return mulVV(n,a,0,b,0);
}

public static double [] getRatio(final int n,final double [] a,final double [] b)
{
    final double [] r=new double [n];
    for(int k=0;k<n;k++) r[k]=a[k]/b[k];
    return r;
}

/** Calculate m x n^T, where m - matrix, n - transpose matrix. */
public static double [][] mulM_x_NT(final double [][] m,final double [][] nT)
{
    final int d=nT.length;
    final double [][] r=new double [m.length][];
    for(int k=0;k<r.length;k++)
    {
	r[k]=mulMV(nT,m[k]);
    }
    return r;
}
public static double [] mulVMT(final double [] a,final double [][] mT)
{
    if(a.length!=mT.length) throw new IllegalArgumentException();
    final double [] b=new double [mT[0].length];
    for(int l=0;l<a.length;l++)
    {
	if(b.length!=mT[l].length) throw new IllegalArgumentException();
	for(int k=0;k<b.length;k++)
	{
	    b[k]+=a[l]*mT[l][k];
	}
    }
    return b;
}
public static double [][] mulM_x_N(final double [][] m,final double [][] n)
{
    final int d=n.length;
    final double [][] r=new double [m.length][];
    for(int k=0;k<r.length;k++)
    {
	r[k]=mulVMT(m[k],n);
    }
    return r;
}
public static double getA_M_A(final int n,final double [] A,final double [][] matr)
{
    return getA_M_B(n,A,matr,A);
}
public static double getA_M_B(final int n,final double [] A,final double [][] matr,final double [] B)
{
    double s=0;
    for(int l=0;l<n;l++)
    {
	double sm=0;
	for(int m=0;m<n;m++) sm+=matr[l][m]*B[m];
	s+=A[l]*sm;
    }
    return s;
}



/** Calculates spur of matrix.
 */
public static double getMatrSpur(final int n,final double [] matr,final int nm)
{
    double s=0;
    for(int l=0;l<n;l++)
    {
	s+=matr[l*nm+l];
    }
    return s;
}
/** Calculates sum(a[k]*matr[k,k]).
 */
public static double getMatrWSpur(final double [] a,final double [] matr,final int nm)
{
    final int n=a.length;
    double s=0;
    for(int l=0;l<n;l++)
    {
	s+=a[l]*matr[l*nm+l];
    }
    return s;
}
/** Calculates spur of A*B.
 */
public static double getSpurAB(final int n,final double [] A,final int na,final double [] B,final int nb)
{
    double s=0;
    for(int k=0;k<n;k++)
    {
	double ss=0;
	for(int l=0;l<n;l++){ss+=A[k*na+l]*B[l*nb+k];}
	s+=ss;
    }
    return s;
}
public static double getAMatrB(final double [] A,final double [] B,final double [] matr,final int nm)
{
    return getAMatrB(A.length,A,0,B.length,B,0,matr,nm);
}
public static double getAMatrB(final int n,final double [] A,final double [] B,final double [] matr,final int nm)
{
    return getAMatrB(n,A,0,B,0,matr,nm);
}
public static double getAMatrB(final int n,final double [] A,final int A_offset,final double [] B,final int B_offset,final double [] matr,final int nm)
{
    return getAMatrB(n,A,A_offset,n,B,B_offset,matr,nm,0);
}
public static double getAMatrB(final int na,final double [] A,final int A_offset,final int nb,final double [] B,final int B_offset,final double [] matr,final int nm)
{
    return getAMatrB(na,A,A_offset,nb,B,B_offset,matr,nm,0) ;
}

/** Calculates A_i matr_ij B_j
 */
public static double getAMatrB(final int na,final double [] A,final int A_offset,final int nb,final double [] B,final int B_offset,final double [] matr,final int nm,final int off_matr)
{
    double s=0;
    for(int l=0;l<na;l++)
    {
	s+=A[l+A_offset]*mulVV(nb,B,B_offset,matr,off_matr+l*nm);
    }
    return s;
}
public static double getAMatrA(final double [] A,final double [] matr,final int nm)
{
    return getAMatrA(A.length,A,matr,nm);
}
/** Calculates A_i matr_ij A_j
 */
public static double getAMatrA(final int n,final double [] A,final double [] matr,final int nm)
{
    return getAMatrB(n,A,A,matr,nm);
}
public static double getAMatrA(final int n,final double [] A,final int A_offset,final double [] matr,final int nm)
{
    return getAMatrB(n,A,A_offset,n,A,A_offset,matr,nm);
}
/** Calculates matr_ij A_j
 */
public static double [] getMatr_x_A(final int n,final double [] matr,final int nm,final double [] A)
{
    final double [] r=new double [n];
    set_getMatr_x_A(n,r,matr,nm,A);
    return r;
}
/** Calculates matr_ij A_j
 */
public static void set_getMatr_x_A(final int n,final double [] r,final double [] matr,final int nm,final double [] A)
{
    set_getMatr_x_A(n,r,matr,nm,0,A,0);
}
/** Calculates matr_ij A_j
 */
public static void set_getMatr_x_A(final int n,final double [] r,final double [] matr,final int nm,final int matr_offset,final double [] A,final int A_offset)
{
    for(int l=0;l<n;l++)
    {
	r[l]=mulVV(n,A,A_offset,matr,matr_offset+l*nm);
    }
}
/** Calculates matrA_im matrB_mj
 */
public static void set_getMatr_x_Matr(final int n,final double [] r,final double [] matrA,final int nmA,final double [] matrB,final int nmB)
{
    set_getMatr_x_Matr(n,r,matrA,nmA,0,matrB,nmB,0);
}
public static double mulVV(final int n,final double [] a,final int off_a,final double [] b,final int off_b)
{
    double s=0;
    for(int i=n;--i>=0;)
    {
	s+=a[i+off_a]*b[i+off_b];
    }
    return s;
}

public static double mulVVV(final double [] a,final double [] b,final double [] c)
{
    return mulVVV(a.length,a,0,b,0,c,0);
}
public static double mulVVV(final int n,final double [] a,final int off_a,final double [] b,final int off_b,final double [] c,final int off_c)
{
    double s=0;
    for(int i=n;--i>=0;)
    {
	s+=a[i+off_a]*b[i+off_b]*c[i+off_c];
    }
    return s;
}
public static double mulV2V(final double [] a,final double [] b)
{
    return mulV2V(a.length,a,0,b,0);
}
public static double mulV2V(final int n,final double [] a,final int off_a,final double [] b,final int off_b)
{
    return mulVVV(n,a,off_a,a,off_a,b,off_b);
}
/** Calculates matrA_im matrB_mj
 */
public static void set_getMatr_x_Matr(final int n,final double [] r,final double [] matrA,final int nmA,final int off_matrA,final double [] matrB,final int nmB,final int off_matrB)
{
    for(int k=0;k<n;k++)
    {
	for(int l=0;l<n;l++)
	{
	    double s=0;
	    for(int m=0,o=off_matrA+k*nmA;m<n;m++,o++) s+=matrA[o]*matrB[m*nmB+off_matrB+l];
	    r[k*n+l]=s;
	}
    }
}
/** Calculates M_ij=matrA_im (matrB_mj)^T=matrA_im matrB_jm
 */
public static void set_getMatr_x_MatrTrans(final int n,final double [] r,final double [] matrA,final int nmA,final int off_matrA,final double [] matrB,final int nmB,final int off_matrB)
{
    for(int k=0;k<n;k++)
    {
	for(int l=0;l<n;l++)
	{
	    double s=0;
	    for(int m=0,o=off_matrA+k*nmA;m<n;m++,o++) s+=matrA[o]*matrB[l*nmB+off_matrB+m];
	    r[k*n+l]=s;
	}
    }
}
/** Calculates M_ij=(matrA_im)^T matrB_mj=matrA_mi matrB_mj
 */
public static void set_getMatrTrans_x_Matr(final int n,final double [] r,final double [] matrA,final int nmA,final int off_matrA,final double [] matrB,final int nmB,final int off_matrB)
{
    for(int k=0;k<n;k++)
    {
	for(int l=0;l<n;l++)
	{
	    double s=0;
	    for(int m=0;m<n;m++) s+=matrA[m*nmA+(k+off_matrA)]*matrB[m*nmB+(l+off_matrB)];
	    r[k*n+l]=s;
	}
    }
}
/** Calculates M_ij=(matrA_im)^T (matrB_mj)^T=matrA_mi matrB_jm
 */
public static void set_getMatrTrans_x_MatrTrans(final int n,final double [] r,final double [] matrA,final int nmA,final int off_matrA,final double [] matrB,final int nmB,final int off_matrB)
{
    for(int k=0;k<n;k++)
    {
	for(int l=0;l<n;l++)
	{
	    double s=0;
	    for(int m=0;m<n;m++) s+=matrA[m*nmA+(k+off_matrA)]*matrB[(l*nmB+off_matrB)+m];
	    r[k*n+l]=s;
	}
    }
}

public static double [] getMatr_x_Matr(final int n,final double [] matrA,final int nmA,final double [] matrB,final int nmB)
{
    final double [] r=new double [n*n];
    set_getMatr_x_Matr(n,r,matrA,nmA,matrB,nmB);
    return r;
}

public static double [] getMatrTrans(final int n,final double [] matr,final int nmA,final int off)
{
    final double [] m=new double [n*n];
    for(int k=0;k<n;k++)
    {
	for(int l=0;l<n;l++)
	{
	    m[k*n+l]=matr[l*nmA+k+off];
	}
    }
    return m;
}
public static double [] getMatrNoTrans(final int n,final double [] matr,final int nmA,final int off)
{
    final double [] m=new double [n*n];
    for(int k=0;k<n;k++)
    {
	for(int l=0;l<n;l++)
	{
	    m[k*n+l]=matr[k*nmA+l+off];
	}
    }
    return m;
}

/** Multiply several matrices res=matrs[0] x matrs[1] x matrs[2]  ...
 *  Used for unit tests.
 */
public static double [] mulMatrices(final int n,final double [][] matrs)
{
    if(matrs.length<=0) return null;
    if(matrs.length==1) return matrs[0].clone();
    if(matrs.length==2) return getMatr_x_Matr(n,matrs[0],n,matrs[1],n);
    double [] tmp1=new double [n*n],tmp2=new double [n*n],xx=null;
    for(int k=1;k<matrs.length;k++)
    {
	set_getMatr_x_Matr(n,tmp2,k>1?tmp1:matrs[0],n,0,matrs[k],n,0);
	xx=tmp2;
	tmp2=tmp1;
	tmp1=xx;
    }
    return xx;
}


/** r[i*sB+j]=\sum_{k=0}^{k=s-1} A[i*ldA+k]*B[k*ldB+j] ; i=[0..sA-1], j=[0..sB-1].
    A: sA x s matrix, B: s x sB matrix, r: sA x sB matrix, ld(r)=sB.
 */
public static void setMA_x_MB(final double [] r,final int sA,final double [] A,final int ldA,final int off_A,final int s,final double [] B,final int sB,final int ldB,final int off_B)
{
    if(s>ldA||sB>ldB) throw new IllegalArgumentException();
    for(int i=0;i<sA;i++)
    {
	for(int j=0;j<sB;j++)
	{
	    double sum=0;
	    for(int k=0;k<s;k++)
	    {
		sum+=A[off_A+i*ldA+k]*B[off_B+k*ldB+j];
	    }
	    r[i*sB+j]=sum;
	}
    }
}

/** r[i*sB+j]=\sum_{k=0}^{k=s-1} A[i*ldA+k]*B[j*ldB+k] ; i=[0..sA-1], j=[0..sB-1].
    A: sA x s matrix, B: sB x s matrix, r: sA x sB matrix, ld(r)=sB.
 */
public static void setMA_x_MBTrans(final double [] r,final int sA,final double [] A,final int ldA,final int off_A,final int s,final double [] B,final int sB,final int ldB,final int off_B)
{
    if(s>ldA||s>ldB) throw new IllegalArgumentException();
    for(int i=0;i<sA;i++)
    {
	for(int j=0;j<sB;j++)
	{
	    double sum=0;
	    for(int k=0;k<s;k++)
	    {
		sum+=A[off_A+i*ldA+k]*B[off_B+j*ldB+k];
	    }
	    r[i*sB+j]=sum;
	}
    }
}

/** r[i*sB+j]=\sum_{k=0}^{k=s-1} A[k*ldA+i]*B[k*ldB+j] ; i=[0..sA-1], j=[0..sB-1].
    A: s x sA matrix, B: s x sB matrix, r: sA x sB matrix, ld(r)=sB.
 */
public static void setMATrans_x_MB(final double [] r,final int sA,final double [] A,final int ldA,final int off_A,final int s,final double [] B,final int sB,final int ldB,final int off_B)
{
    if(sA>ldA||sB>ldB) throw new IllegalArgumentException();
    for(int i=0;i<sA;i++)
    {
	for(int j=0;j<sB;j++)
	{
	    double sum=0;
	    for(int k=0;k<s;k++)
	    {
		sum+=A[off_A+k*ldA+i]*B[off_B+k*ldB+j];
	    }
	    r[i*sB+j]=sum;
	}
    }
}


/** Calculates QQ+QQ^T QQ-QQ^T
    depending on flag_antiHermitian.
*/
 public static void selectHermitianPart(final int n,final double [] QQ,final int ldQQ,final int offQQ,final boolean flag_antiHermitian)
{
    if(n>ldQQ) throw new IllegalArgumentException();
    for(int k=0;k<n;k++)
    {
	for(int l=0;l<=k;l++)
	{
	    if(flag_antiHermitian)
	    {
		final double d=QQ[offQQ+k*ldQQ+l]-QQ[offQQ+k+l*ldQQ];
		QQ[offQQ+k+l*ldQQ]=-d;
		QQ[offQQ+k*ldQQ+l]=d;
	    }
	    else
	    {
		final double d=QQ[offQQ+k*ldQQ+l]+QQ[offQQ+k+l*ldQQ];
		QQ[offQQ+k+l*ldQQ]=QQ[offQQ+k*ldQQ+l]=d;
	    }
	}
    }
}

/** Calculates a[i]*a[j] matrix.
 */
public static double [] getQQaa(final double [] a)
{
    return scaleVector(getQQaiaj(a),0.5);
}

/** Calculates 2*a[i]*a[j] matrix.
 */
public static double [] getQQaiaj(final double [] a)
{
    return getQQab(a.length,a,a);
}

/** Calculates a[i]*b[j]+b[i]*a[j] matrix.
 */
public static double [] getQQab(final int n,final double [] a,final double [] b)
{
    final double [] QQ=new double [n*n];
    for(int i=0;i<n;i++){
	final double ai=a[i],bi=b[i];
	for(int j=0,in=i*n;j<n;j++) QQ[in+j]=ai*b[j]+bi*a[j];
    }
    return QQ;
}

/** Calculates c[i]=a[i]*b[i].
 */
public static double [] getVAB(final int n,final double [] a,final double [] b)
{
    final double [] v=new double [n];
    for(int i=0;i<n;i++){
	v[i]=a[i]*b[i];
    }
    return v;
}
public static double [] getVAB(final double [] a,final double [] b)
{
    return getVAB(a.length,a,b);
}
public static double [] getUnitMatrix(final int n)
{
    final double [] O=new double [n*n];
    for(int k=0;k<n;k++) O[k+n*k]=1;
    return O;
}
public static double [] getDiagMatrix(final double [] v)
{
    return getDiagMatrix(v.length,v);
}
public static double [] getDiagMatrix(final int n,final double [] v)
{
    final double [] O=new double [n*n];
    for(int k=0;k<n;k++) O[k+n*k]=v[k];
    return O;
}

static double [] getRandomMatrix(final int nels,final java.util.Random r)
{
    final double [] q=new double [nels];
    for(int k=0;k<q.length;k++) q[k]=1-2*r.nextDouble();
    return q;
}

static void test_mulMatrices(final java.util.Random r,final double eps)
{
    final int n_tests=5000;
    for(int t=0;t<n_tests;t++)
    {
    for(int n=1;n<10;n++)
    {
	for(int d=1;d<6;d++)
	{
	    final double [][] m=new double [d][];
	    for(int kk=0;kk<m.length;kk++)
	    {
		final double [] qq=m[kk]=getRandomMatrix(n*n,r);
	    }

	    final double [] mmul=mulMatrices(n,m);
	    double [] mt=m[0];
	    for(int k=1;k<m.length;k++)
	    {
		mt=getMatr_x_Matr(n,mt,n,m[k],n);
	    }
	    UnitTests.checkDoubleArrayRel(mmul,"mmul",mt,"mt",eps);
	}
    }
    }
    System.err.println("TMatr:test_mulMatrices: OK with n_tests="+n_tests+" eps="+eps);
}
static void test_Mul(final java.util.Random r,final double eps)
{
    final int n_tests=5000;
    for(int t=0;t<n_tests;t++)
    {
	for(int n=1;n<10;n++)
	{
	    for(int ld1=n;ld1<=n+3;ld1++)
	    {
		final int o1=r.nextInt(3),o2=r.nextInt(30),ld2=n+r.nextInt(ld1-n+1);
		final double [] m1=getRandomMatrix(o1+n*ld1,r),m2=getRandomMatrix(o2+n*ld2,r);
		final double [] res=new double [n*n];
		set_getMatr_x_Matr(n,res,m1,ld1,o1,m2,ld2,o2);
		UnitTests.checkDoubleArrayRel(res,"set_getMatr_x_Matr",mulMatrices(n,new double [][]{getMatrNoTrans(n,m1,ld1,o1),getMatrNoTrans(n,m2,ld2,o2)}),"mul",eps);

		set_getMatr_x_MatrTrans(n,res,m1,ld1,o1,m2,ld2,o2);
		UnitTests.checkDoubleArrayRel(res,"set_getMatr_x_MatrTrans",mulMatrices(n,new double [][]{getMatrNoTrans(n,m1,ld1,o1),getMatrTrans(n,m2,ld2,o2)}),"mul",eps);

		set_getMatrTrans_x_Matr(n,res,m1,ld1,o1,m2,ld2,o2);
		UnitTests.checkDoubleArrayRel(res,"set_getMatrTrans_x_Matr",mulMatrices(n,new double [][]{getMatrTrans(n,m1,ld1,o1),getMatrNoTrans(n,m2,ld2,o2)}),"mul",eps);

		set_getMatrTrans_x_MatrTrans(n,res,m1,ld1,o1,m2,ld2,o2);
		UnitTests.checkDoubleArrayRel(res,"set_getMatrTrans_x_MatrTrans",mulMatrices(n,new double [][]{getMatrTrans(n,m1,ld1,o1),getMatrTrans(n,m2,ld2,o2)}),"mul",eps);
	    }
	}
    }
    System.err.println("TMatr:test_Mul: OK with n_tests="+n_tests+" eps="+eps);
}



public static void main(String [] args)
{
    final java.util.Random r=new java.util.Random(662);
    final double eps=1e-7;

    System.err.println("TMatr::main(): starting ALL tests; ver=$Id: TMatr.java,v 1.62 2018/01/27 18:22:42 mal Exp $");
    final long t0=System.currentTimeMillis();
    test_mulMatrices(r,eps);
    test_Mul(r,eps);
    final long t1=System.currentTimeMillis();
    System.err.println("TMatr::main(): ALL tests OK done in "+(t1-t0)+" milliseconds. ver=$Id: TMatr.java,v 1.62 2018/01/27 18:22:42 mal Exp $");
}


}
