package com.polytechnik.utils;
/*
 * (C) Vladislav Malyshkin 2010
 * This file is under GPL version 3.
 *
 */

/** N points Vandermonde function interpolation.
 *  @version $Id: VandermondeInterpolation.java,v 1.22 2011/11/05 19:31:52 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */

public class VandermondeInterpolation  {
    final int order;
    private final double [][] matrix;
    final double [] rightpart;
    private final int [] ipvt;
    public VandermondeInterpolation(final int order)
    {
	this.order=order;
	this.rightpart=new double [this.order];
	this.matrix=getMatrix(this.order);
	// prepare matrix for linear equaltion solving
	this.ipvt=new int [this.order];
	final int info=Linsystems.dgefa(this.matrix,this.order,this.ipvt);
	if(info!=0)
	{
	    throw new RuntimeException("INFO="+info);
	}
    }
    protected double [][] getMatrix(final int o)
    {
	final double [][] m=new double [o][];
	for(int i=0;i<o;i++)
	{
	    final double [] l=m[i]=new double [o];
	    final double x=((double)(i+1))/(1+o); // equal distance
	    l[0]=1;
	    for(int j=1;j<o;j++)
	    {
		l[j]=l[j-1]*x;
	    }
	}
	return m;
    }

    public void setFMoments(final double [] fmoments)
    {
	if(fmoments.length!=1+this.order)
	{
	    throw new RuntimeException("fmoments.length="+fmoments.length+
				       " should be 1+this.order="+(1+this.order));
	}
	// the fmoments[0] is the norm, copy just 1..this.order elements
	System.arraycopy(fmoments,1,this.rightpart,0,this.order);
	Linsystems.dgesl(this.matrix,this.order,this.ipvt,this.rightpart,1);
    }

    public double getF(final double x)
    {
	double s=0;
	for(int i=this.order;i-->0;)
	{
	    s=x*s+this.rightpart[i];
	}
	return s;
    }
    public double getDF(final double x)
    {
	double s=0;
	for(int i=this.order;i-->1;)
	{
	    s=x*s+this.rightpart[i]*i;
	}
	return s;
    }
    public double getD2F(final double x)
    {
	double s=0;
	for(int i=this.order;i-->2;)
	{
	    s=x*s+this.rightpart[i]*(i*(i-1));
	}
	return s;
    }
    public double getD3F(final double x)
    {
	double s=0;
	for(int i=this.order;i-->3;)
	{
	    s=x*s+this.rightpart[i]*(i*(i-1)*(i-2));
	}
	return s;
    }


    private static void checkSymD3()
    {
	final VandermondeInterpolation v=new VandermondeInterpolation(3);
	final double [] fm=new double [v.order+1];
	final java.util.Random r=new java.util.Random(-79091);
	final double eps=1e-9;
	int n=0;
	final double d=1.0/(1+v.order);
	for(;n<100000;n++)
	{
	    for(int j=1;j<fm.length;j++)
	    {
		fm[j]=r.nextDouble();
	    }
	    v.setFMoments(fm);

	    {
	    final double f=v.getDF(0.5);
	    final double fe=(-1*fm[1]+fm[3])/(2*d);
	    UnitTests.checkDouble(f,"f",fe,"fe",eps);
	    }

	    {
	    final double f=v.getDF(d);
	    final double fe=(-3*fm[1]+4*fm[2]-fm[3])/(2*d);
	    UnitTests.checkDouble(f,"f",fe,"fe",eps);
	    }

	    {
	    final double f=v.getDF(1-d);
	    final double fe=(fm[1]-4*fm[2]+3*fm[3])/(2*d);
	    UnitTests.checkDouble(f,"f",fe,"fe",eps);
	    }

	    {
	    final double f2=v.getD2F(0.5);
	    final double fe2=(fm[1]-2*fm[2]+fm[3])/(d*d);
	    UnitTests.checkDouble(f2,"f2",fe2,"fe2",eps);
	    }


	}
	System.err.println("checkSymD3() "+n+" test OK matched to explicit 3-points formulae.");
    }


    private static void checkSymD5()
    {
	final VandermondeInterpolation v=new VandermondeInterpolation(5);
	final double [] fm=new double [v.order+1];
	final java.util.Random r=new java.util.Random(-79091);
	final double eps=1e-9;
	int n=0;
	final double d=1.0/(1+v.order);
	for(;n<100000;n++)
	{
	    for(int j=1;j<fm.length;j++)
	    {
		fm[j]=r.nextDouble();
	    }
	    v.setFMoments(fm);
	    {
	    final double f=v.getDF(0.5);
	    final double fe=(fm[1]-8*fm[2]+8*fm[4]-fm[5])/(12*d);
	    UnitTests.checkDouble(f,"f",fe,"fe",eps);
	    }

	    {
	    final double ff=v.getDF(d);
	    final double fef=(-25*fm[1]+48*fm[2]-36*fm[3]+16*fm[4]-3*fm[5])/(12*d);
	    UnitTests.checkDouble(ff,"ff",fef,"fef",eps);
	    }

	    {
	    final double ff1=v.getDF(2*d);
	    final double fef1=(-3*fm[1]-10*fm[2]+18*fm[3]-6*fm[4]+fm[5])/(12*d);
	    UnitTests.checkDouble(ff1,"ff1",fef1,"fef1",eps);
	    }

	    {
	    final double f2=v.getD2F(0.5);
	    final double fe2=(-fm[1]+16*fm[2]-30*fm[3]+16*fm[4]-fm[5])/(12*Math.pow(d,2));
	    UnitTests.checkDouble(f2,"f2",fe2,"fe2",eps);
	    }

	    {
	    final double f2f=v.getD2F(d);
	    final double fe2f=(35*fm[1]-104*fm[2]+114*fm[3]-56*fm[4]+11*fm[5])/(12*Math.pow(d,2));
	    UnitTests.checkDouble(f2f,"f2f",fe2f,"fe2f",eps);
	    }

	    {
	    final double f2f1=v.getD2F(2*d);
	    final double fe2f1=(11*fm[1]-20*fm[2]+6*fm[3]+4*fm[4]-fm[5])/(12*Math.pow(d,2));
	    UnitTests.checkDouble(f2f1,"f2f1",fe2f1,"fe2f1",eps);
	    }

	    {
	    final double f3=v.getD3F(0.5);
	    final double fe3=(-fm[1]+2*fm[2]-2*fm[4]+fm[5])/(2*d*d*d);
	    UnitTests.checkDouble(f3,"f3",fe3,"fe3",eps);
	    }

	    {
	    final double f3=v.getD3F(d);
	    final double fe3=(-5*fm[1]+18*fm[2]-24*fm[3]+14*fm[4]-3*fm[5])/(2*d*d*d);
	    UnitTests.checkDouble(f3,"f3",fe3,"fe3",eps);
	    }

	    {
	    // backward approximation
	    final double f3fb=v.getD3F(1-d);
	    final double fe3fb=(3*fm[1]-14*fm[2]+24*fm[3]-18*fm[4]+5*fm[5])/(2*d*d*d);
	    UnitTests.checkDouble(f3fb,"f3fb",fe3fb,"fe3fb",eps);
	    final double f2fb=v.getD2F(1-d);
	    final double fe2fb=(11*fm[1]-56*fm[2]+114*fm[3]-104*fm[4]+35*fm[5])/(12*d*d);
	    UnitTests.checkDouble(f2fb,"f2fb",fe2fb,"fe2fb",eps);
	    final double ffb=v.getDF(1-d);
	    final double fefb=(3*fm[1]-16*fm[2]+36*fm[3]-48*fm[4]+25*fm[5])/(12*d);
	    UnitTests.checkDouble(ffb,"ffb",fefb,"fefb",eps);
	    }
	}
	System.err.println("checkSymD5() "+n+" test OK matched to explicit 5-points formulae.");
    }


    static void checkValues(final VandermondeInterpolation v)
    {
	final double [] fm=new double [v.order+1];
	final java.util.Random r=new java.util.Random(-79091);
	final double eps=1e-6;
	int n=0;
	for(;n<100000;n++)
	{
	    for(int j=1;j<fm.length;j++)
	    {
		fm[j]=r.nextDouble();
	    }
	    v.setFMoments(fm);
	    for(int j=1;j<fm.length;j++)
	    {
		final double fi=v.getF(((double)j)/(1+v.order));
		// match values
		UnitTests.checkDouble(fi,"fi",fm[j],"fm["+j+"]",eps);
	    }
	}
	System.err.println("checkValues() "+n+" test OK values matched order="+v.order+".");
    }

    private static void checkExactPolinomial()
    {
	final VandermondeInterpolation v=new VandermondeInterpolation(7);
	final double [] fm=new double [v.order+1],coefs=new double [v.order];
	final java.util.Random r=new java.util.Random(-79091);
	final double eps=1e-7;
	int n=0;
	for(;n<100000;n++)
	{
	    // random polinomian
	    for(int j=0;j<coefs.length;j++)
	    {
		coefs[j]=r.nextDouble();
	    }
	    for(int j=1;j<fm.length;j++)
	    {
		final double x=((double)j)/(v.order+1);
		double s=0;
		for(int k=v.order;k-->0;)
		{
		    s=s*x+coefs[k];
		}
		fm[j]=s;
	    }
	    v.setFMoments(fm);

	    // match exact values of v.order-1 polynomial
	    for(double x=-3;x<3;x+=1e-2)
	    {
		final double fi=v.getF(x),dfi=v.getDF(x);
		double s=0,ds=0;
		for(int k=v.order;k-->0;)
		{
		    s=s*x+coefs[k];
		    if(k>0)
		    {
			ds=ds*x+coefs[k]*k;
		    }
		}
		// match values
		UnitTests.checkDouble(s,"s",fi,"fi",eps);
		UnitTests.checkDouble(ds,"ds",dfi,"dfi",eps);
	    }
	}
	System.err.println("checkExactPolinomial() "+n+" test OK values matched.");
    }

    private static void printExact5PointsCoeficients()
    {
	final VandermondeInterpolation v=new VandermondeInterpolation(5);
	final double [] fm=new double [v.order+1];
	final double h=1.0/(1+v.order);
	System.err.println("printExact5PointsCoeficients(): exact coefficients");



	System.err.println("F");
	System.err.println("n\tback\tmid\tforward\tahead\tahead2");
	for(int i=1;i<=v.order;i++)
	{
	    final double n=1.0;
	    java.util.Arrays.fill(fm,0);
	    fm[i]=1;
	    v.setFMoments(fm);
	    System.err.println(String.valueOf(i)+":\t"+(v.getF(h)/n)+"\t"+(v.getF(0.5)/n)+"\t"+(v.getF(1-h)/n)+"\t"+(v.getF(1)/n)+"\t"+(v.getF(1+h)/n));
	}


	System.err.println("DF, norm=1/12h");
	System.err.println("n\tback\tmid\tforward\tahead\tahead2");
	for(int i=1;i<=v.order;i++)
	{
	    final double n=1.0/(12*h);
	    java.util.Arrays.fill(fm,0);
	    fm[i]=1;
	    v.setFMoments(fm);
	    System.err.println(String.valueOf(i)+":\t"+(v.getDF(h)/n)+"\t"+(v.getDF(0.5)/n)+"\t"+(v.getDF(1-h)/n)+"\t"+(v.getDF(1)/n)+"\t"+(v.getDF(1+h)/n));
	}

	System.err.println("D2F, norm=1/12h^2");
	System.err.println("n\tback\tmid\tforward\tahead\tahead2");
	for(int i=1;i<=v.order;i++)
	{
	    final double n=1.0/(12*h*h);
	    java.util.Arrays.fill(fm,0);
	    fm[i]=1;
	    v.setFMoments(fm);
	    System.err.println(String.valueOf(i)+":\t"+(v.getD2F(h)/n)+"\t"+(v.getD2F(0.5)/n)+"\t"+(v.getD2F(1-h)/n)+"\t"+(v.getD2F(1)/n)+"\t"+(v.getD2F(1+h)/n));
	}


	System.err.println("D3F, norm=1/2h^3");
	System.err.println("n\tback\tmid\tforward\tahead\tahead2");
	for(int i=1;i<=v.order;i++)
	{
	    final double n=1.0/(2*h*h*h);
	    java.util.Arrays.fill(fm,0);
	    fm[i]=1;
	    v.setFMoments(fm);
	    System.err.println(String.valueOf(i)+":\t"+(v.getD3F(h)/n)+"\t"+(v.getD3F(0.5)/n)+"\t"+(v.getD3F(1-h)/n)+"\t"+(v.getD3F(1)/n)+"\t"+(v.getD3F(1+h)/n));
	}


	System.err.println("D1(ff), norm=1/12h");
	System.err.println("n\tD1(ff)");
	for(int i=1;i<=v.order;i++)
	{
	    final double n=1.0/(12*h);
	    java.util.Arrays.fill(fm,0);
	    fm[i]=1;
	    v.setFMoments(fm);
	    System.err.println(String.valueOf(i)+":\t"+(((v.getF(1+h)-v.getF(1))/h)/n));
	}

	System.err.println("D1S, norm=1/12h");
	for(int i=0;i<=v.order;i++)
	{
	    if(i<=0)
	    {

		System.err.print("n");
		for(int j=1;j<=v.order;j++)
		{
		    System.err.print("\tback_"+(v.order-j));
		}
		System.err.println();
	    }
	    else
	    {
		System.err.print(String.valueOf(i)+":");
		final double n=1.0/(12*h);
		java.util.Arrays.fill(fm,0);
		fm[i]=1;
		v.setFMoments(fm);
		for(int j=1;j<=v.order;j++)
		{
		    System.err.print("\t"+(v.getDF(j*h)/n));
		}
		System.err.println();
	    }
	}
    }

    private static void printFExact7PointsCoeficients()
    {
	final VandermondeInterpolation v=new VandermondeInterpolation(7);
	final double [] fm=new double [v.order+1];
	final double h=1.0/(1+v.order);
	System.err.println("printFExact7PointsCoeficients(): exact coefficients");



	System.err.println("F");
	System.err.println("n\tback\tmid\tforward\tahead\tahead2");
	for(int i=1;i<=v.order;i++)
	{
	    final double n=1.0;
	    java.util.Arrays.fill(fm,0);
	    fm[i]=1;
	    v.setFMoments(fm);
	    System.err.println(String.valueOf(i)+":\t"+(v.getF(h)/n)+"\t"+(v.getF(0.5)/n)+"\t"+(v.getF(1-h)/n)+"\t"+(v.getF(1)/n)+"\t"+(v.getF(1+h)/n));
	}

    }


    public static void main(String [] s)
    {
	checkSymD5();
	checkSymD3();
	checkValues(new VandermondeInterpolation(7));
	checkExactPolinomial();
	printExact5PointsCoeficients();
	printFExact7PointsCoeficients();
    }
}
